/******************************************************************************
 **
 ** $Id: main.C 176 2009-04-17 21:16:12Z AndrewLahiff $
 **
 ******************************************************************************
 ** 
 ** Copyright (C) 2007-2009 Andrew Lahiff.
 ** 
 ** This file is part of plotcf.
 **
 ** plotcf is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 ** 
 ** plotcf is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 ** 
 ** You should have received a copy of the GNU General Public License
 ** along with plotcf.  If not, see <http://www.gnu.org/licenses/>.
 ** 
 *******************************************************************************/

#include <iostream>
#include <cstdlib>

#include "readfile.h"

#include "misc.h"

/*******************************************************************************
 **
 ** convert time in HH:MM:SS format to seconds
 **
 *******************************************************************************/

double time_to_tis(char timein[])
{

   QRegExp timeCheck("\\d\\d:\\d\\d:\\d\\d");
   QString timeQStr = timein;
   if (timeCheck.indexIn(timeQStr) == -1)
   {
      return(-1);
   }

   double result, d_hr, d_mn, d_sc;
   std::string timeStr = timein;

   d_hr = atof(((timeStr.substr(0, 2)).c_str()));
   d_mn = atof(((timeStr.substr(3, 2)).c_str()));
   d_sc = atof(((timeStr.substr(6, 6)).c_str()));

   result = d_sc + d_mn*60.0 + d_hr*3600.0;

   return(result);
}

/*******************************************************************************
 **
 ** convert a date in QString YYYY-MM-DD format to QDate
 **
 *******************************************************************************/

QDate toQDate(QString in)
{
    int beginYr = in.mid(0, 4).toInt();
    int beginMn = in.mid(5, 2).toInt();
    int beginDy = in.mid(8, 2).toInt();
    return(QDate(beginYr, beginMn, beginDy));
}

/*******************************************************************************
 **
 ** add seconds to a time stamp
 **
 *******************************************************************************/

QString timeBounds(QString timeStamp, double offset)
{
   QDateTime dateTimeIn = QDateTime::fromString(timeStamp, "yyyy-MM-ddTHH:mm:ss.zzzZ");
   return((dateTimeIn.addMSecs(offset*1000.0)).toString("yyyy-MM-ddTHH:mm:ss.zzzZ"));   
}

/*******************************************************************************
 **
 ** returns 'true' if DATASET_ID corresponds to something that should be
 ** plotted as a spectrogram. Most instruments haven't bothered to use
 ** DISPLAYTYPE unfortunately.
 **
 *******************************************************************************/

bool isSpectro(QString in)
{
   if (in.contains("RAP_E2DD_R")) return (true);
   else if (in.contains("CP_RAP_E2DD")) return (true);
   else if (in.contains("CP_CIS-HIA_HS_MAG_IONS") && !in.contains("MOMENTS")) return (true);
   else if (in.contains("CP_CIS-CODIF_HS_O1") && !in.contains("MOMENTS")) return (true);
   else if (in.contains("CP_CIS-CODIF_HS_H1") && !in.contains("MOMENTS")) return (true);
   else if (in.contains("CP_CIS-CODIF_HS_He2") && !in.contains("MOMENTS")) return (true);
   else if (in.contains("CP_RAP_E3DD")) return(true);
   else if (in.contains("CP_RAP_E3DD_R")) return(true);
   else if (in.contains("CP_CIS-HIA_HS_SW_IONS")) return (true);
   else if (in.contains("CP_WHI_ACTIVE")) return(true);
   else if (in.contains("CP_WHI_NATURAL")) return(true);
   else if (in.contains("CP_RAP_ESPCT")) return(true);
   else if (in.contains("CP_RAP_HSPCT")) return(true);
   else if (in.contains("CP_RAP_ISPCT")) return(true);
   else return(false);
}

/*******************************************************************************
 **
 ** return the maximum value in an array of a specified size
 **
 *******************************************************************************/

int maxValueOf(int array[], int size)
{
   int max = -1;
   for (int i=0; i<size; i++)
      if (array[i] > max) max = array[i];
   return(max);
}

/*******************************************************************************
 **
 ** is a floating point number a fill value?
 **
 *******************************************************************************/

// this should really be replaced with something better. Here b is the fill value,
// an a is the number we want to check if it's a fill value or not.

bool fl_compare(double a, double b)
{
   if (fabs(b) > 1.0e+37)
   {
      if (fabs(a) > 1.0e+37) return(true);
      else return(false);
   }
   else
   {
      if (a != b) return(false);
      else return(true);
   }
}

/*******************************************************************************
 **
 ** print help information
 **
 *******************************************************************************/

using namespace std;

void printHelp()
{
   cout << "" << endl;
   cout << "Usage: plotcf [mode] [options] [files]" << endl;
   cout << "" << endl;
   cout << "plotcf has four modes: (1) generation of time-series plots and spectrograms," << endl;
   cout << "(2) generation of wheel plots, (3) generation of Sauvaud energy-angle-time" << endl;
   cout << "spectrograms, and (4) list all data variables in the specified CDF and CEF" << endl;
   cout << "files. By default plotcf will make time-series plots and spectrograms." << endl;
   cout << "" << endl;
   cout << "In addition, a number of short cuts are provided that enable users to easily" << endl;
   cout << "make commonly-used plots with a minimum of effort." << endl;
   cout << "" << endl;
   cout << "Mode:" << endl;
   cout << "  -wheel        generate wheel plots" << endl;
   cout << "  -sauvaud      generate Sauvaud plots" << endl;
   cout << "  -list         list all data variables" << endl;
   cout << "  -list all     list all data variables, including SUPPORT_DATA variables" << endl;
   cout << "" << endl;
   cout << "Options:" << endl;
   cout << "  -st           specify a start time in the format HH:MM:SS" << endl;
   cout << "  -sp           specify an end time in the format HH:MM:SS" << endl;
   cout << "  -title        specify the title of the plot" << endl;
   cout << "  -nodisplay    don't show the plot on the screen" << endl;
   cout << "  -o            specify an output file. The file format is determined by the" << endl;
   cout << "                extension given in the filename." << endl;
   cout << "  -w            plot width in pixels" << endl;
   cout << "  -h            plot height in pixels" << endl;
   cout << "" << endl;  
   cout << "  -ymin         set the minimum value of all y-axes" << endl;
   cout << "  -ymax         set the maximum value of all y-axes" << endl;
   cout << "  -cmin         set the minimum value of all colour bars" << endl;
   cout << "  -cmax         set the maximum value of all colour bars" << endl;
   cout << "" << endl;
   cout << "  -ctable       set the colour table used for spectrograms. The possiblities are" << endl;
   cout << "                0: default colour table, 1: colour table similar to CAA/CSDS;" << endl;
   cout << "                2: similar to IDL's red-blue colour table" << endl;
   cout << "" << endl;
   cout << "  -yline        draw a horizontal line at a particular y-axis value" << endl;
   cout << "  -grid         draw a grid using dotted lines" << endl;
   cout << "" << endl;
   cout << "  -smooth       apply box-car averaging to all time-series curves, using" << endl;
   cout << "                the time specified in seconds" << endl;
   cout << "" << endl;
   cout << "  -noplotlabel  don't show the plot label in the bottom RH corner" << endl;
   cout << "  -plotlabel    specify a plot label to be shown in the bottom RH corner" << endl;
   cout << "" << endl;
   cout << "Short cuts:" << endl;
   cout << "  -n            plot all variables with PROPERTY = \"Number_Density\"" << endl;
   cout << "  -v            plot all variables with PROPERTY = \"Velocity\"" << endl;
   cout << "  -t            plot all variables with PROPERTY = \"Temperature\"" << endl;
   cout << "  -nt           plot all density and temperature variables" << endl;
   cout << "  -nv           plot all density and velocity variables" << endl;
   cout << "  -nvt          plot all density, velocity and temperature variables" << endl;
   cout << "" << endl;
   cout << "Additional features of the above short cuts:" << endl;
   cout << "  * if a PEACE 2-D or 3-D data file is specified, a spectrogram will be" << endl;
   cout << "    included in the top panel." << endl;
   cout << "  * if both a PEACE 2-D or 3-D data file and an EFW spacecraft" << endl;
   cout << "    potential file are specified, the potential will be overlaid on" << endl;
   cout << "    the spectrogram." << endl;
   cout << "  * if an FGM B-field file is specified, a panel will be included showing" << endl;
   cout << "    the magnetic field vector." << endl;
   cout << "" << endl;
   cout << "More short cuts:" << endl;
   cout << "  -ppa          plot parallel, perpendicular and anti-parallel pitch angle" << endl;
   cout << "                spectrograms. If an EFW spacecraft potential file is specified," << endl;
   cout << "                by default the potential will be overlaid on the spectrograms." << endl;
   cout << "  -app          plot anti-parallel, perpendicular and parallel pitch angle" << endl;
   cout << "                spectrograms. If an EFW spacecraft potential file is specified," << endl;
   cout << "                by default the potential will be overlaid on the spectrograms." << endl;
   cout << "  -prs          plot a combined PEACE and RAPID spectrogram. PEACE and RAPID" << endl;
   cout << "                files must be specified. The energy units in the spectrogram" << endl;
   cout << "                y-axis will be eV." << endl;
   cout << "" << endl;
   cout << "The following options are only applicable to wheel plots:" << endl;
   cout << "  -wheel ga     plot a mirror image of the LHS on the RHS of the wheel" << endl;
   cout << "  -wheel bl     plot background level data on the RHS of the wheel" << endl;
   cout << "  -time         specify time of data record to plot. The time can be in one" << endl;
   cout << "                of the following formats: HH:MM:SS, HH:MM:SS.SSS, or" << endl;
   cout << "                YYYY-MM-DDTHH:MM:SS.SSSZ. The record with the specified time" << endl;
   cout << "                or nearest time after the specified time will be plotted." << endl;
   cout << " " << endl;
   cout << "The following options are only applicable to Sauvaud plots:" << endl;
   cout << "  - status      includes a small panel containing the PADSelectionQuality status" << endl;
   cout << "                variable" << endl;
   cout << " " << endl;
   cout << "Spacecraft potential options:" << endl;
   cout << "  -scp-show     overlay the EFW spacecraft potential on all PEACE spectrograms." << endl;
   cout << "                If a plot contains PEACE spectrograms from multiple spacecraft," << endl;
   cout << "                EFW spacecraft potential files from multiple spacecraft should" << endl;
   cout << "                be specified. The correct spacecraft potential will be overlaid" << endl;
   cout << "                on each spectrogram." << endl;
   cout << "  -scp-corr     apply spacecraft potential correction to all PEACE spectrograms." << endl;
   cout << "                If a plot contains PEACE spectrograms from multiple spacecraft," << endl;
   cout << "                EFW spacecraft potential files from multiple spacecraft should" << endl;
   cout << "                be specified. The correct spacecraft potential will be used to" << endl;
   cout << "                correct each spectrogram. If a floating point number is supplied" << endl;
   cout << "                as an additional argument, this will be used as an offset to the" << endl;
   cout << "                EFW potential. A positive number will increase the magnitude of" << endl;
   cout << "                the potential, while a negative number will decrease the magnitude" << endl;
   cout << "                of the potential." << endl;
   cout << " " << endl;
   cout << "The following options are only applicable when the PEACE PITCH_FULL or" << endl;
   cout << "PEACE PITCH_FULL_LAR products are used, and can be used for spectrograms, wheel" << endl;
   cout << "plots or Sauvaud plots:" << endl;
   cout << "  -sensor l     plot the LEEA sensor only" << endl;
   cout << "  -sensor h     plot the HEEA sensor only" << endl;
   cout << "  -sensor lh    plot the LEEA & HEEA sensors only" << endl;
   cout << "  -halfspin n   use data from half-spin n, where n = 1 or 2" << endl;
   cout << " " << endl;

}

/*******************************************************************************
 **
 ** find appropriate S/C potential value
 **
 *******************************************************************************/

void getPotentialValue(float azBegin, float azEnd, int *count, int max,
                       double *xx, double *yy, float *potential)
{

// azimuth centre time

   float centreTime = (azBegin + azEnd)/2.0;

// we now need to find the nearest potential values just before and just after the azimuth
// centre time. We will then interpolate to determine the potential at the azimuth centre
// time. However, if there is a single potential value during the azimuth, we will use this
// instead.

   int posBefore =- 1, posAfter = -1, posDuring =-1;
   int found = -1;

   while (found == -1 && *count < max)
   {
      if (xx[*count] < azBegin)
      {
         posBefore = *count;
      }
      if (xx[*count] > azEnd)
      {
         posAfter = *count;
      }
      if (posBefore >= 0 && posAfter >= 0)
      {
         found = 1;
         *count = *count - 10;
         if (*count < 0) *count = 0;
      }
      if (xx[*count] >= azBegin && xx[*count] <= azEnd)
      {
         posDuring = *count;
         found = 2;
         *count = *count - 10;
         if (*count < 0) *count = 0;
      }
      *count = *count + 1;
   }

   if (found == 2)
   {
      *potential = yy[posDuring];
   }
   else if (found == 1)
   {
      *potential = yy[posBefore] + (yy[posAfter] - yy[posBefore])*(centreTime - xx[posBefore])/
                                                                  (xx[posAfter] - xx[posBefore]);                                       
   }
   else *potential = 0; // if we can't determine the potential, just set it to zero
   *potential = -*potential; // EFW potential has negative sign
}

/*******************************************************************************
 **   
 ** get name of variable containing Energy table
 **      
 *******************************************************************************/

QString getVarNameEnergy(ReadFile& file, int varNum)
{
    QString name = "";
    if (file.getVariableAttr(varNum, "DEPEND_1").contains("Energy"))
        name = file.getVariableAttr(varNum, "DEPEND_1");
    if (file.getVariableAttr(varNum, "DEPEND_2").contains("Energy"))
        name = file.getVariableAttr(varNum, "DEPEND_2");
    if (file.getVariableAttr(varNum, "DEPEND_3").contains("Energy"))
        name = file.getVariableAttr(varNum, "DEPEND_3");
    return name;
}

/*******************************************************************************
 **
 ** get name of variable containing PA table
 **
 *******************************************************************************/

QString getVarNamePolar(ReadFile& file, int varNum)
{
    QString name = "";
    if (file.getVariableAttr(varNum, "DEPEND_1").contains("PitchAngle"))
        name = file.getVariableAttr(varNum, "DEPEND_1");
    if (file.getVariableAttr(varNum, "DEPEND_2").contains("PitchAngle"))
        name = file.getVariableAttr(varNum, "DEPEND_2");
    if (file.getVariableAttr(varNum, "DEPEND_3").contains("PitchAngle"))
        name = file.getVariableAttr(varNum, "DEPEND_3");
    return name;
}

/*******************************************************************************
 **
 ** get name of variable containing Energy table
 **
 *******************************************************************************/

QString getVarNameAzimuth(ReadFile& file, int varNum)
{
    QString name = "";
    if (file.getVariableAttr(varNum, "DEPEND_1").contains("Azimuth"))
        name = file.getVariableAttr(varNum, "DEPEND_1");
    if (file.getVariableAttr(varNum, "DEPEND_2").contains("Azimuth"))
        name = file.getVariableAttr(varNum, "DEPEND_2");
    if (file.getVariableAttr(varNum, "DEPEND_3").contains("Azimuth"))
        name = file.getVariableAttr(varNum, "DEPEND_3");
    return name;
}

/*******************************************************************************
 **
 ** find the variable number for a given variable name
 **
 *******************************************************************************/

int varNumFromName(ReadFile& file, QString name)
{
    for (int i=0; i<file.getNumVariables(); ++i)
    {
        if (file.getVariableAttr(i, "NAME") == name)
        {
            return i;
        }
    }
    return -1;
}
