/******************************************************************************
 **
 ** $Id: wheel.cpp 183 2009-04-24 12:33:15Z AndrewLahiff $
 **
 ******************************************************************************
 **
 ** Copyright (C) 2007-2009 Andrew Lahiff.
 **
 ** This file is part of plotcf.
 **
 ** plotcf is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** plotcf is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with plotcf.  If not, see <http://www.gnu.org/licenses/>.
 **
 *******************************************************************************/

#include <iostream> // for debugging only

#include "readfile.h"
#include "misc.h"
#include "QWheelPlot.h"

#include "wheel.h"

int getVarAzi(ReadFile& file, int var)
{
   return varNumFromName(file, getVarNameAzimuth(file, var));
}

Wheel::Wheel()
{
    num_en = 0;
    num_pa = 0;
    num_az = 0;
}

void Wheel::includeReflection(bool refl)
{
    reflection_ = refl;
}


bool Wheel::getData(int varNum, bool getAz, ReadFile& file)
{
    // get variable numbers of the required energy, polar & azimuth tables

    int vn_en = varNumFromName(file, getVarNameEnergy(file, varNum));
    if (vn_en == -1) return false;

    int vn_pa = varNumFromName(file, getVarNamePolar(file, varNum));
    if (vn_pa == -1) return false;

    int vn_az = -1;
    if (getAz)
    {
        vn_az = varNumFromName(file, getVarNameAzimuth(file, varNum));
        if (vn_az == -1) return false;
    }

    // determine the sizes of the energy, polar & azimuth tables

    num_en = file.getVariableSize(vn_en, 0);
    if (num_en < 1) return false;

    en_cen.resize(num_en);
    en_low.resize(num_en);
    en_upp.resize(num_en);

    num_pa = file.getVariableSize(vn_pa, 0);
    if (num_pa < 1) return false;

    pa_cen.resize(num_pa);
    pa_low.resize(num_pa);
    pa_upp.resize(num_pa);

    if (getAz)
    {
        num_az = file.getVariableSize(vn_az, 0);
        if (num_az < 1) return false;

        az_cen.resize(num_az);
        az_low.resize(num_az);
        az_upp.resize(num_az);
   
        data.resize(num_en*num_pa*num_az);
    }
    else
    {
        data.resize(num_en*num_pa);
    }

    // get the required energy, polar & azimuth tables

    file.getDataValue(vn_en, &en_cen[0]);
    file.getDataValueDelta(vn_en, _PLUS, &en_upp[0]);
    file.getDataValueDelta(vn_en, _MINUS, &en_low[0]);

    file.getDataValue(vn_pa, &pa_cen[0]);
    file.getDataValueDelta(vn_pa, _PLUS, &pa_upp[0]);
    file.getDataValueDelta(vn_pa, _MINUS, &pa_low[0]);

    if (getAz)
    {
        file.getDataValue(vn_az, &az_cen[0]);
        file.getDataValueDelta(vn_az, _PLUS, &az_upp[0]);
        file.getDataValueDelta(vn_az, _MINUS, &az_low[0]);
    }

    // get the data

    file.getDataValue(varNum, &data[0]);

    // get the fill values

    fillEnergy = file.getVariableAttr(vn_en, "FILLVAL").toDouble();
    fillPolar = file.getVariableAttr(vn_pa, "FILLVAL").toDouble();
    if (getAz)
    {
        fillAzimuth = file.getVariableAttr(vn_az, "FILLVAL").toDouble();
    }
    fillData = file.getVariableAttr(varNum, "FILLVAL").toDouble();
}

void Wheel::plot(QWheelPlot& plot, bool side, double ymin, double ymax, bool seten)
{
    plot.setInvisibleValue(fillData);
    plot.z = &data[0];
    plot.numR = num_en;
    plot.numT = num_pa;

    std::vector<float> ebins_low;
    ebins_low.resize(num_en);

    std::vector<float> ebins_upp;
    ebins_upp.resize(num_en);

    std::vector<float> pbins_low;
    pbins_low.resize(num_pa);
   
    std::vector<float> pbins_upp;
    pbins_upp.resize(num_pa);

    // generate required energy arrays; calculate min/max energies

    double yMin_ = 3.0e+38, yMax_ = -3.0e+38;

    for (int i=0; i<num_en; ++i)
    {
        ebins_low.at(i) = en_cen.at(i) - en_low.at(i);
        ebins_upp.at(i) = en_cen.at(i) + en_upp.at(i);
        if (yMin_ > ebins_low.at(i) && ebins_low.at(i) != fillEnergy && ebins_low.at(i) > 0.0)
            yMin_ = ebins_low.at(i);
        if (yMax_ < ebins_upp.at(i) && ebins_upp.at(i) != fillEnergy)
            yMax_ = ebins_upp.at(i);
    }

    // generate required pa arrays
 
    for (int i=0; i<num_pa; ++i)
    {
        pbins_low.at(i) = 0.0 - (pa_cen.at(i) - pa_low.at(i) + 90.0); // LHS of plot
        pbins_upp.at(i) = 0.0 - (pa_cen.at(i) + pa_upp.at(i) + 90.0); 
        if (side) pbins_low.at(num_pa - i - 1) = 180.0 - (pa_cen.at(i) - pa_low.at(i) + 90.0);
        if (side) pbins_upp.at(num_pa - i - 1) = 180.0 - (pa_cen.at(i) + pa_upp.at(i) + 90.0);
    }

    // set energy & pa arrays for plotting

    plot.rlo = &ebins_low[0];
    plot.rhi = &ebins_upp[0];
    plot.thlo = &pbins_low[0];
    plot.thhi = &pbins_upp[0];

    // set the data array

    plot.z = &data[0];

    // set auto energy limits if necessary - this should only be done ONCE, only before
    // any other plots are drawn

    if (seten)
    {
        if (ymin < 0)
        {
            plot.yAxisMin = yMin_;
        }
        if (ymax < 0)
        {
            plot.yAxisMax = yMax_;
        }
    }

    // draw the plot

    plot.drawPlot();

    if (num_az == 2)
    {

        // generate required pa arrays

        for (int i=0; i<num_pa; ++i)
        {
            pbins_low.at(num_pa - i - 1) = 180.0 - (pa_cen.at(i) - pa_low.at(i) + 90.0);
            pbins_upp.at(num_pa - i - 1) = 180.0 - (pa_cen.at(i) + pa_upp.at(i) + 90.0);
        }

        plot.z = &data[num_en*num_pa];
        plot.drawPlot();
    }

    // draw reflection if needed

    if (reflection_)
    {
        // generate required pa arrays

        for (int i=0; i<num_pa; ++i)
        {
            pbins_low.at(num_pa - i - 1) = 180.0 - (pa_cen.at(i) - pa_low.at(i) + 90.0);
            pbins_upp.at(num_pa - i - 1) = 180.0 - (pa_cen.at(i) + pa_upp.at(i) + 90.0);
        }

        // set pa arrays for plotting

        plot.thlo = &pbins_low[0];
        plot.thhi = &pbins_upp[0];

        // make plot

        plot.drawPlot();
    }
}



