/*******************************************************************************
 **
 ** $Id: QWheelPlot.C 183 2009-04-24 12:33:15Z AndrewLahiff $
 **
 ****************************************************************************** 
 ** 
 ** Copyright (C) 2007-2009 Andrew Lahiff.
 ** 
 ** This file is part of plotcf.
 **
 ** plotcf is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 ** 
 ** plotcf is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 ** 
 ** You should have received a copy of the GNU General Public License
 ** along with plotcf.  If not, see <http://www.gnu.org/licenses/>.
 ** 
 *******************************************************************************/

#include <iostream>
#include "QWheelPlot.h"

QWheelPlot::QWheelPlot(int uPw, int uPh, int nR, int nT,
                       double yMin, double yMax, double cMin, double cMax,
                       int defaultFontSize, bool gui)
{
   noShowLabel = false;
   plotWidth = uPw;
   plotHeight = uPh;
   numR = nR;
   numT = nT;
   yAxisMin = yMin;
   yAxisMax = yMax;
   cAxisMin = cMin;
   cAxisMax = cMax;

   plot = new QImage(plotWidth, plotHeight, QImage::Format_RGB32);
   plot->fill(QColor(255, 255, 255).rgb());

   xMarginL = 90., xMarginR = 120.; 
   yMargin = 60.;

   gap = 20., colourBarWidth = 18.;
   
   gapBetweenAxisAndTick = 5.;

   iPlotTickMajor = 8;
   iPlotTickMinor = 4;
  
   cornerX = xMarginL;
   cornerY = yMargin;

   myPlotWidth = plotWidth - xMarginL - xMarginR;
   myPlotHeight = myPlotWidth;

   imageWidthPlot = plotWidth;
   imageHeightPlot = plotHeight;
   
   QString defaultFontFamily = "Lucida";

   fontLabels.setFamily(defaultFontFamily);
   fontLabels.setPointSize(defaultFontSize);
   fontTitles.setFamily(defaultFontFamily);
   fontTitles.setPointSize(defaultFontSize+1);
   fontTitles.setBold(true);
   fontSubTitles.setFamily(defaultFontFamily);
   fontSubTitles.setPointSize(defaultFontSize+1);   
   fontTicks.setFamily(defaultFontFamily);
   fontTicks.setPointSize(defaultFontSize+2);
   fontMarkers.setFamily(defaultFontFamily);
   fontMarkers.setPointSize(defaultFontSize+3);
   fontPL.setFamily(defaultFontFamily);
   fontPL.setPointSize(defaultFontSize-4);
 
   useGUI = gui;

}

//-------------------------------------------------------------------------------------------------
// 
// destructor
// 
//-------------------------------------------------------------------------------------------------

QWheelPlot::~QWheelPlot()
{
   delete plot;
}

//-------------------------------------------------------------------------------------------------
//
// draw title on RHS or LHS
//
//-------------------------------------------------------------------------------------------------

void QWheelPlot::drawSideTitle(int which, QString text)
{
   QPainter paint(plot);
   paint.setFont(fontSubTitles);
   double xv = cornerX + 25.;
   double yv = cornerY + 25.;
   if (which == RHS) xv = xv + myPlotWidth/2.;
   paint.translate(QPointF(xv, yv));
   drawRichText(text, fontSubTitles, ALIGN_LEFT, &paint);
}

//-------------------------------------------------------------------------------------------------
//
// draw frame
//
//-------------------------------------------------------------------------------------------------

void QWheelPlot::drawFrame(QString xLabel, QString yLabel, QString cLabel, QString title, QString subTitle)
{
   QPainter paint(plot);

   paint.setRenderHint(QPainter::TextAntialiasing);

   paint.setPen(Qt::black);
   paint.drawRect(QRectF(cornerX, cornerY, myPlotWidth, myPlotHeight));
   paint.drawLine(QLineF(cornerX, cornerY + myPlotHeight/2., cornerX + myPlotWidth, cornerY + myPlotHeight/2.));
   paint.drawLine(QLineF(cornerX + myPlotWidth/2., cornerY, cornerX + myPlotWidth/2., cornerY + myPlotHeight));

   int nmax = (int)myPlotHeight; // handle different plot sizes
   if (nmax < 100) nmax = 100;

   for (int i=0; i<nmax; i++)
   {
      double yLoc = (double)i*(myPlotHeight/(double)nmax);
      double yPos = cornerY + myPlotHeight - yLoc;
      double hue = ((double)(myPlotHeight - yLoc))/((double)myPlotHeight);
      QColor myColor = colourBarDefinition(hue);
      paint.setPen(myColor);
      paint.drawLine(QLineF(cornerX + myPlotWidth + gap, yPos, 
                            cornerX + myPlotWidth + gap + colourBarWidth, yPos));
   }

   paint.setPen(Qt::black);
   paint.drawRect(QRectF(cornerX + myPlotWidth + gap, cornerY, colourBarWidth, myPlotHeight));   
   paint.setRenderHint(QPainter::TextAntialiasing);

// draw the title

   paint.setPen(Qt::black);
   paint.translate(myPlotWidth/2. + cornerX, cornerY - 35.);
   QSizeF titlePos = drawRichText(title, fontTitles, ALIGN_CENTRE, &paint);
   paint.resetMatrix();
   
// draw the subtitle

   paint.setPen(Qt::black);
   paint.translate(myPlotWidth/2. + cornerX, cornerY - 35. + 17.);
   drawRichText(subTitle, fontSubTitles, ALIGN_CENTRE, &paint);
   paint.resetMatrix();   

// draw the x-axis label

   paint.setPen(Qt::black);
   paint.translate(myPlotWidth/2. + cornerX, plotHeight - 15.);
   drawRichText(xLabel, fontLabels, ALIGN_CENTRE, &paint);
   paint.resetMatrix();

// draw the y-axis label

   paint.setPen(Qt::black);
   paint.translate(40, myPlotHeight/2.0 + cornerY);
   paint.rotate(270);
   drawRichText(yLabel, fontLabels, ALIGN_CENTRE, &paint);
   paint.resetMatrix();

// draw the z-axis label

   paint.setPen(Qt::black);
   paint.translate(myPlotWidth + cornerX + gap/2., myPlotHeight/2. + cornerY);
   paint.rotate(270);
   drawRichText(cLabel, fontLabels, ALIGN_CENTRE, &paint);
   paint.resetMatrix();

// draw colour box ticks

   double ccAxisMin = SafeLog(cAxisMin);
   double ccAxisMax = SafeLog(cAxisMax);
   double scaleFactorC = myPlotHeight/(ccAxisMax - ccAxisMin);

   double posX1 = cornerX + myPlotWidth + gap + colourBarWidth;
   double posX2 = cornerX + myPlotWidth + gap + colourBarWidth - iPlotTickMajor;
   double posX3 = cornerX + myPlotWidth + gap + colourBarWidth - iPlotTickMinor;

   int iNumberWritten = 0;
   for (int i=-20; i<20; i++)
   {
      double yValue = (double)i;
      double posY = myPlotHeight - scaleFactorC*(yValue - ccAxisMin) + cornerY;

// major ticks and labels

      if (yValue >= ccAxisMin && yValue <= ccAxisMax)
      {
         paint.drawLine(QLineF(posX1, posY, posX2, posY));
         paint.setFont(fontTicks);
         double realValue = pow(10.0, (double)i);
         QString text = QString("%1").arg(realValue, 0, 'e', 0);
         iNumberWritten++;
         paint.translate(cornerX + myPlotWidth + gap + colourBarWidth + gapBetweenAxisAndTick, posY);
         drawRichText(text, fontTicks, ALIGN_LEFT, &paint, 1);
         paint.resetMatrix();
      }
   }

// minor ticks

   for (int i=-20; i<20; i++)
   {
      double yValue = (double)i;
      double posY = myPlotHeight - scaleFactorC*(yValue - ccAxisMin) + cornerY;

      int wroteOne = 0;
      for (int j=2; j<10; j++)
      {
         yValue = SafeLog((double)j*pow(10.0, (double)i));
         posY = myPlotHeight - scaleFactorC*(yValue - ccAxisMin) + cornerY;
         if (yValue > ccAxisMin && yValue < ccAxisMax)
         {
            paint.drawLine(QLineF(posX1, posY, posX3, posY));
            if (iNumberWritten == 0 || (iNumberWritten < 2 && wroteOne == 0))
            {
               double realValue = (double)j*pow(10.0, (double)i);
               QString text = QString("%1").arg(realValue, 0, 'e', 3);
               paint.translate(cornerX-43.0, posY);
               drawRichText(text, fontTicks, ALIGN_RIGHT, &paint, 1);
               paint.resetMatrix();
               wroteOne++;
            }
         }
      }
   }

// setup pen for circles

   QVector<qreal> myDots;
   myDots << 4 << 8;
   QPen penCircles;
   penCircles.setDashPattern(myDots);
   penCircles.setColor(QColor("#ffffff"));

// major & minor y-axis ticks

   ccAxisMin = SafeLog(yAxisMin);
   ccAxisMax = SafeLog(yAxisMax);
   scaleFactorC = myPlotHeight/(ccAxisMax - ccAxisMin)/2.;

   posX1 = cornerX + myPlotWidth;
   posX2 = cornerX + myPlotWidth - iPlotTickMajor;
   posX3 = cornerX + myPlotWidth - iPlotTickMinor;
   double posX4 = cornerX;
   double posX5 = cornerX + iPlotTickMajor;
   double posX6 = cornerX + iPlotTickMinor;
   
   QSizeF tickSize;

   iNumberWritten = 0;
   for (int i=-20; i<20; i++)
   {  
      double yValue = (double)i;
      double posY;
         
      if (yValue > SafeLog(yAxisMin) && yValue < SafeLog(yAxisMax))
      {
         posY = myPlotHeight/2. - scaleFactorC*(yValue - ccAxisMin) + cornerY;
         paint.drawLine(QLineF(posX1, posY, posX2, posY));
	 paint.drawLine(QLineF(posX4, posY, posX5, posY));
         paint.drawLine(QLineF(posX1 - myPlotWidth/2., posY, posX2 - myPlotWidth/2., posY));
         paint.drawLine(QLineF(posX4 + myPlotWidth/2., posY, posX5 + myPlotWidth/2., posY));

         posY = myPlotHeight/2. + scaleFactorC*(yValue - ccAxisMin) + cornerY;
         paint.drawLine(QLineF(posX1, posY, posX2, posY));
	 paint.drawLine(QLineF(posX4, posY, posX5, posY));
         paint.drawLine(QLineF(posX1 - myPlotWidth/2., posY, posX2 - myPlotWidth/2., posY));
         paint.drawLine(QLineF(posX4 + myPlotWidth/2., posY, posX5 + myPlotWidth/2., posY));

	 double realValue = pow(10.0, (double)i);
         QString text = QString("%1").arg(realValue, 0, 'e', 0);
	 posY = myPlotHeight/2. - scaleFactorC*(yValue - ccAxisMin) + cornerY;
         paint.translate(cornerX-gapBetweenAxisAndTick, posY);
         tickSize = drawRichText(text, fontTicks, ALIGN_RIGHT, &paint, 1);
         paint.resetMatrix();

	 posY = myPlotHeight/2. + scaleFactorC*(yValue - ccAxisMin) + cornerY;
         paint.translate(cornerX-gapBetweenAxisAndTick, posY);
         drawRichText(text, fontTicks, ALIGN_RIGHT, &paint, 1);
         paint.resetMatrix();

         double pos_Y1 = myPlotHeight/2. - scaleFactorC*(yValue - ccAxisMin) + cornerY;
         double pos_X1 = cornerX + myPlotWidth/2. - scaleFactorC*(yValue - ccAxisMin);

         paint.setPen(penCircles);
         paint.drawEllipse(QRectF(pos_X1, pos_Y1,
			          2.*(cornerX + myPlotWidth/2. - pos_X1), 2.*(cornerY + myPlotHeight/2. - pos_Y1)));
	 paint.setPen(Qt::SolidLine);

      }

      for (int j=2; j<10; j++)
      {
         yValue = SafeLog((double)j*pow(10.0, (double)i));
         if (yValue > SafeLog(yAxisMin) && yValue < SafeLog(yAxisMax))
         {
            posY = myPlotHeight/2. - scaleFactorC*(yValue - ccAxisMin) + cornerY;
            paint.drawLine(QLineF(posX1, posY, posX3, posY));
	    paint.drawLine(QLineF(posX4, posY, posX6, posY));
            paint.drawLine(QLineF(posX1 - myPlotWidth/2., posY, posX3 - myPlotWidth/2., posY));
            paint.drawLine(QLineF(posX4 + myPlotWidth/2., posY, posX6 + myPlotWidth/2., posY));

            posY = myPlotHeight/2. + scaleFactorC*(yValue - ccAxisMin) + cornerY;
            paint.drawLine(QLineF(posX1, posY, posX3, posY));
	    paint.drawLine(QLineF(posX4, posY, posX6, posY));
            paint.drawLine(QLineF(posX1 - myPlotWidth/2., posY, posX3 - myPlotWidth/2., posY));
            paint.drawLine(QLineF(posX4 + myPlotWidth/2., posY, posX6 + myPlotWidth/2., posY));
         }
      }
   }

// major & minor x-axis ticks

   double posY1 = cornerY;
   double posY2 = cornerY + iPlotTickMajor;
   double posY3 = cornerY + iPlotTickMinor;
   double posY4 = cornerY + myPlotHeight/2.;
   double posY5 = cornerY + myPlotHeight/2. -  iPlotTickMajor;
   double posY6 = cornerY + myPlotHeight/2. -  iPlotTickMinor;

   iNumberWritten = 0;
   for (int i=-20; i<20; i++)
   {  
      double xValue = (double)i;
      double posX;
         
      if (xValue > SafeLog(yAxisMin) && xValue < SafeLog(yAxisMax))
      {  
         posX = cornerX + myPlotWidth/2. + scaleFactorC*(xValue - ccAxisMin);
         paint.drawLine(QLineF(posX, posY1, posX, posY2));
         paint.drawLine(QLineF(posX, posY4, posX, posY5));
         paint.drawLine(QLineF(posX, posY1 + myPlotHeight/2., posX, posY2 + myPlotHeight/2.));
         paint.drawLine(QLineF(posX, posY4 + myPlotHeight/2., posX, posY5 + myPlotHeight/2.));

         posX = cornerX + myPlotWidth/2. - scaleFactorC*(xValue - ccAxisMin);
         paint.drawLine(QLineF(posX, posY1, posX, posY2));
         paint.drawLine(QLineF(posX, posY4, posX, posY5));
         paint.drawLine(QLineF(posX, posY1 + myPlotHeight/2., posX, posY2 + myPlotHeight/2.));
         paint.drawLine(QLineF(posX, posY4 + myPlotHeight/2., posX, posY5 + myPlotHeight/2.));

	 posX = cornerX + myPlotWidth/2. + scaleFactorC*(xValue - ccAxisMin);
	 double realValue = pow(10.0, (double)i);
	 QString text = QString("%1").arg(realValue, 0, 'e', 0);
         paint.translate(posX, cornerY + myPlotHeight + gapBetweenAxisAndTick + tickSize.height()/2.);
         drawRichText(text, fontTicks, ALIGN_CENTRE, &paint, 1);
         paint.resetMatrix();

	 posX = cornerX + myPlotWidth/2. - scaleFactorC*(xValue - ccAxisMin);
         paint.translate(posX, cornerY + myPlotHeight + gapBetweenAxisAndTick + tickSize.height()/2.);
         drawRichText(text, fontTicks, ALIGN_CENTRE, &paint, 1);
         paint.resetMatrix();
      }

      for (int j=2; j<10; j++)
      {
         xValue = SafeLog((double)j*pow(10.0, (double)i));
         if (xValue > SafeLog(yAxisMin) && xValue < SafeLog(yAxisMax))
         {
            posX = cornerX + myPlotWidth/2. + scaleFactorC*(xValue - ccAxisMin);
            paint.drawLine(QLineF(posX, posY1, posX, posY3));
            paint.drawLine(QLineF(posX, posY4, posX, posY6));
            paint.drawLine(QLineF(posX, posY1 + myPlotHeight/2., posX, posY3 + myPlotHeight/2.));
            paint.drawLine(QLineF(posX, posY4 + myPlotHeight/2., posX, posY6 + myPlotHeight/2.));
            posX = cornerX + myPlotWidth/2. - scaleFactorC*(xValue - ccAxisMin);
            paint.drawLine(QLineF(posX, posY1, posX, posY3));
            paint.drawLine(QLineF(posX, posY4, posX, posY6));
            paint.drawLine(QLineF(posX, posY1 + myPlotHeight/2., posX, posY3 + myPlotHeight/2.));
            paint.drawLine(QLineF(posX, posY4 + myPlotHeight/2., posX, posY6 + myPlotHeight/2.));
         }
      }
   }

// labels indicating 0 and 180

   QString label1 = QString("0");
   QString label2 = QString("180");
   paint.translate(cornerX + myPlotWidth/2., cornerY + myPlotHeight/17.);
   drawRichText(label1, fontMarkers, ALIGN_CENTRE, &paint);
   paint.resetMatrix();
   paint.translate(cornerX + myPlotWidth/2., cornerY + myPlotHeight - myPlotHeight/17.);
   drawRichText(label2, fontMarkers, ALIGN_CENTRE, &paint);
   paint.resetMatrix();
}

//-------------------------------------------------------------------------------------------------
//
// draw the wheel plot data
//
//-------------------------------------------------------------------------------------------------

void QWheelPlot::drawPlot()
{
   int k;
   float zval;

   for (int i=0; i<numR; i++)
   {

      float rrlo, rrhi;
      if (rlo[i] < yAxisMin) rrlo = yAxisMin;
      else rrlo = rlo[i];
      if (rhi[i] < yAxisMin) rrhi = yAxisMin;
      else rrhi = rhi[i];

      rrhi = SafeLog(rrhi);
      rrlo = SafeLog(rrlo);

      rrlo = (rrlo - SafeLog(yAxisMin))/(SafeLog(yAxisMax) - SafeLog(yAxisMin))*myPlotHeight/2.;
      rrhi = (rrhi - SafeLog(yAxisMin))/(SafeLog(yAxisMax) - SafeLog(yAxisMin))*myPlotHeight/2.;

      for (int j=0; j<numT; j++)
      {
         bool dontPlot = false;
         k = numR*j+i;
         zval = SafeLog(z[k]);
         zval = (zval < SafeLog(cAxisMin)) ? SafeLog(cAxisMin) : zval;
         zval = (zval > SafeLog(cAxisMax)) ? SafeLog(cAxisMax) : zval;
         double col = (zval - SafeLog(cAxisMin))/(SafeLog(cAxisMax) - SafeLog(cAxisMin));

         if (SafeLog(z[k]) < SafeLog(cAxisMin)) dontPlot = true;
         if (z[k] < 0)  dontPlot = true;
	 if (col > 1.0) dontPlot = true;
	 if (col < 0.0) dontPlot = true;

         if (rrhi > 0. && !dontPlot)
         {
            if (rrlo < 0.0) rrlo = 0.0; 
            if (rrlo != rrhi) drawArc(rrlo, rrhi, thlo[j], thhi[j], col);
         }
      }
   }
}

//-------------------------------------------------------------------------------------------------
//
// draw a bin
// - the definition of the polygon used here was taken from PLOTAZ, part of QPEACE,
//   written by Steve Schwartz
//
//-------------------------------------------------------------------------------------------------

void QWheelPlot::drawArc(float rlo, float rhi, float thlo, float thhi, double col)
{
   QPointF points[6] = 
   {
      QPointF(rlo*cos(QPGDEG2RAD*thlo), rlo*sin(QPGDEG2RAD*thlo)),
      QPointF(rhi*cos(QPGDEG2RAD*thlo), rhi*sin(QPGDEG2RAD*thlo)),
      QPointF(rhi*cos(QPGDEG2RAD*0.5*(thlo+thhi)), rhi*sin(QPGDEG2RAD*0.5*(thlo+thhi))),
      QPointF(rhi*cos(QPGDEG2RAD*thhi), rhi*sin(QPGDEG2RAD*thhi)),
      QPointF(rlo*cos(QPGDEG2RAD*thhi), rlo*sin(QPGDEG2RAD*thhi)),
      QPointF(rlo*cos(QPGDEG2RAD*0.5*(thlo+thhi)), rlo*sin(QPGDEG2RAD*0.5*(thlo+thhi)))
   };
   
   QPainter paint(plot);
   paint.setClipRect(QRectF(cornerX, cornerY, myPlotWidth, myPlotHeight));
   col = 1.0 - col;
   if (col != 1.0)
   {
      QColor myColor = colourBarDefinition(col);
      paint.setPen(myColor);
      paint.translate(cornerX + myPlotWidth/2., cornerY + myPlotHeight/2.);
      QBrush myBrush(myColor);
      paint.setBrush(myBrush);
      paint.drawPolygon(points, 6, Qt::OddEvenFill);
   }
}

//-------------------------------------------------------------------------------------------------
//
// save the current plot to a file
//
//-------------------------------------------------------------------------------------------------

void QWheelPlot::saveToFile(QString fileName, QString text)
{
   if (text.size() > 0) plot->setText("Description", text);
   plot->save(fileName, 0, 10);
}

//----------------------------------------------------------------------------------------
//    
// safe log 10
// 
//----------------------------------------------------------------------------------------
      
double QWheelPlot::SafeLog(double x)
{  
   return (x < 1.0e-33) ? -33.0 : log10(x);
}    

//----------------------------------------------------------------------------------------
//
// set value not to plot
// 
//----------------------------------------------------------------------------------------
      
void QWheelPlot::setInvisibleValue(double x)
{
   invisibleValue = x;
}

//-------------------------------------------------------------------------------------------------
//
// draw rich text
//
//-------------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------------
// 
// draw rich text
//    
//-------------------------------------------------------------------------------------------------
      
QSizeF QWheelPlot::drawRichText(QString textIn, QFont font, int alignment, QPainter *painter, int process)
{
   QSizeF returnF;

   if (useGUI)
   {

   QString text = textIn;
   if (process == 1) 
   {
      QRegExp expNorm1("^1e\\+0\\d$");
      QRegExp expNorm2("^1e\\-0\\d$");
      QRegExp expNorm1a("^1e\\+\\d\\d$");
      QRegExp expNorm2a("^1e\\-\\d\\d$");

      if (expNorm1.indexIn(textIn) == 0)
      {
         QString exponent = textIn.mid(4, 1);
         text = QString("10<sup>%1</sup>").arg(exponent);
      }  
      else if (expNorm2.indexIn(textIn) == 0)
      {
         QString exponent = textIn.mid(4, 1);
         text = QString("10<sup>-%1</sup>").arg(exponent);
      }     
      else if (expNorm1a.indexIn(textIn) == 0)
      {
         QString exponent = textIn.mid(3, 2);
         text = QString("10<sup>%1</sup>").arg(exponent);
      }  
      else if (expNorm2a.indexIn(textIn) == 0)
      {  
         QString exponent = textIn.mid(3, 2);
         text = QString("10<sup>-%1</sup>").arg(exponent);
      }

   }
   else if (process == 2)
   {
      if (textIn.contains("^"))
      {
         QString textTmp = textIn;
         bool done = false;
         while (!done)
         {
            if (!textTmp.contains("^")) done = true;
            if (!done)
            {
               int index = textTmp.indexOf("^", 0);
               QString bit = textTmp.mid(index+1, 1);
               if (bit == "{")
               {
                  int beginBr = textTmp.indexOf("{", index);
                  int endBr = textTmp.indexOf("}", beginBr);
                  if (beginBr > -1 && endBr > -1)
                  {
                     bit = textTmp.mid(beginBr+1, (endBr-beginBr));
                     text = textTmp.replace(index, (endBr-beginBr+2), "<sup>"+bit+"</sup>");
	          }
               }
               else if (bit == "-")
               {
                  bit = textTmp.mid(index+1, 2);
                  text = textTmp.replace(index, 3, "<sup>"+bit+"</sup>");
               }
               else
               text = textTmp.replace(index, 2, "<sup>"+bit+"</sup>");
            }
         }
         textTmp = text;
      }  
   }

   QTextDocument doc;
   doc.setHtml(text);
   doc.setDefaultFont(font);

   double offsetV = doc.size().height()*0.5;
   double offsetH = doc.size().width();
   if (alignment == ALIGN_LEFT) offsetH = 0.0;
   else if (alignment == ALIGN_RIGHT) offsetH = -offsetH;
   else if (alignment == ALIGN_CENTRE) offsetH = -offsetH/2.;

   painter->translate(QPointF(offsetH, -offsetV));
   doc.drawContents(painter);
   painter->translate(QPointF(-offsetH, offsetV));

   returnF = QSizeF(doc.size().width(), doc.size().height());
   }
   else
   {
      QFontMetrics fm(font);
      double textHeight = fm.boundingRect(textIn).height();
      double textWidth = fm.boundingRect(textIn).width();

      double offsetV = textHeight*0.5;
      double offsetH = textWidth;

      if (alignment == ALIGN_LEFT) offsetH = 0.0;
      else if (alignment == ALIGN_RIGHT) offsetH = -offsetH;
      else if (alignment == ALIGN_CENTRE) offsetH = -offsetH/2.;

      painter->drawText(QPointF(offsetH, offsetV), textIn);

      returnF = QSizeF(textWidth*1.1, textHeight*1.1);

   }

   return(returnF);
}

//-------------------------------------------------------------------------------------------------
//
// set colour table
//
//-------------------------------------------------------------------------------------------------

void QWheelPlot::setColourBarDefinition(int in)
{
   colourBarDef = in;
}

//-------------------------------------------------------------------------------------------------
//    
// convert value between 0 and 1 to a colour for spectrograms
//
//-------------------------------------------------------------------------------------------------

QColor QWheelPlot::colourBarDefinition(double in)
{
   QColor myColor;

   switch(colourBarDef)
   {
      case 1: // CAA
      double ca[10], cb[10], cc[10];
      ca[0] = 0.0; // red
      ca[1] = 0.18; // orange
      ca[2] = 0.24; // yellow
      ca[3] = 0.43; // green
      ca[4] = 0.55; // cyan
      ca[5] = 0.75; // blue
      ca[6] = 0.85; // indigo
      ca[7] = 0.95; // violet
      ca[8] = 0.98; // violet (black)
      ca[9] = 1.00; // black
      for (int i=0; i<10; i++) cb[i] = 1.0;
      cb[7] = 0.3; cb[8] = 0.1; cb[9] = 0.0;

      cc[0] = 0.0; // red
      cc[1] = 0.1; // orange
      cc[2] = 0.18; // yellow
      cc[3] = 0.26; // green
      cc[4] = 0.48; // cyan
      cc[5] = 0.64; // blue
      cc[6] = 0.72; // indigio
      cc[7] = 0.78; // violet
      cc[8] = 0.85; // violet (black)
      cc[9] = 0.85; // black

      for (int i=0; i<9; i++)
      {
         if (in >= ca[i] && in < ca[i+1])
         {
            double u = cc[i] + (cc[i+1]-cc[i])*(in-ca[i])/(ca[i+1]-ca[i]);
            double uu = cb[i] + (cb[i+1]-cb[i])*(in-ca[i])/(ca[i+1]-ca[i]);
            myColor.setHsvF(u, 1.0, uu);
         }
      }
      if (in == 1.0) myColor.setHsvF(0.0, 0.0, 1.0);
      break;

      case 2: // IDL red-blue
      {
          double red[5][2], green[6][2], blue[5][2];
          
          red[0][0] = 0.0;  red[0][1] = 0.0;
          red[1][0] = 0.35; red[1][1] = 0.0;
          red[2][0] = 0.66; red[2][1] = 1.0;
          red[3][0] = 0.89; red[3][1] = 1.0;
          red[4][0] = 1.0;  red[4][1] = 0.5;
          
          green[0][0] = 0.0;   green[0][1] = 0.0;
          green[1][0] = 0.125; green[1][1] = 0.0;
          green[2][0] = 0.375; green[2][1] = 1.0;
          green[3][0] = 0.64;  green[3][1] = 1.0;
          green[4][0] = 0.91;  green[4][1] = 0.0;
          green[5][0] = 1.0;   green[5][1] = 0.0;

          blue[0][0] = 0.0;  blue[0][1] = 0.5;
          blue[1][0] = 0.11; blue[1][1] = 1.0;
          blue[2][0] = 0.34; blue[2][1] = 1.0;
          blue[3][0] = 0.65; blue[3][1] = 0.0;
          blue[4][0] = 1.0;  blue[4][1] = 0.0;

          double uR = 0., uG = 0., uB = 0.;

          in = 1.0 - in;

          for (int i=0; i<4; ++i)
          {
             if (in >= red[i][0] && in < red[i+1][0])
             {
                uR = red[i][1] + (red[i+1][1]-red[i][1])*(in-red[i][0])/(red[i+1][0]-red[i][0]);
             }
          }
          if (in <= red[0][0]) uR = red[0][1];
          if (in >= red[4][0]) uR = red[4][1];

          for (int i=0; i<5; ++i)
          {  
             if (in >= green[i][0] && in < green[i+1][0])
             {  
                uG = green[i][1] + (green[i+1][1]-green[i][1])*(in-green[i][0])/(green[i+1][0]-green[i][0]);
             }
          }
          if (in <= green[0][0]) uG = green[0][1];
          if (in >= green[5][0]) uG = green[5][1];
          
          for (int i=0; i<4; ++i)
          {  
             if (in >= blue[i][0] && in < blue[i+1][0])
             {  
                uB = blue[i][1] + (blue[i+1][1]-blue[i][1])*(in-blue[i][0])/(blue[i+1][0]-blue[i][0]);
             }
          }
          if (in <= blue[0][0]) uB = blue[0][1];
          if (in >= blue[4][0]) uB = blue[4][1];

          in = 1.0 - in;
          
          myColor.setRgbF(uR, uG, uB);
      
      }
      break;


      default:
      if (in == 1.0) myColor.setHsvF(0.0, 0.0, 1.0);
      else myColor.setHsvF(in*0.85, 1.0, 1.0);
      break;
   }
   return(myColor);
}

//-------------------------------------------------------------------------------------------------
//
// draw label on plot
//
//-------------------------------------------------------------------------------------------------

void QWheelPlot::drawLabel(QString text)
{
   if (!noShowLabel)
   {
   QPainter painter(plot);
   painter.translate(QPointF(imageWidthPlot*0.99, imageHeightPlot*0.98));
   painter.setPen(QColor("gainsboro"));
   drawRichText(text, fontPL, ALIGN_RIGHT, &painter);
   painter.resetMatrix();
   }
}

