/*****************************************************************************
 **
 ** plotcf - command line tool for quickly & easily plotting CAA CDF and CEF
 **          files.
 **
 **          Plot types include time-series, spectrograms, Sauvaud style
 **          plots (similar to QJAS) and wheel plots (similar to PLOTAZ).
 **
 **          Uses Trolltech Qt 4.3 or above and NASA's CDF library. Wheel
 **          plots are made using polygons as developed by Steve Schwartz
 **          in QPEACE/PLOTAZ.
 **
 ******************************************************************************
 **
 ** $Id: main.C 184 2009-11-01 07:46:35Z AndrewLahiff $
 **
 ******************************************************************************
 ** 
 ** Copyright (C) 2007-2010 Andrew Lahiff.
 ** 
 ** This file is part of plotcf.
 **
 ** plotcf is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 ** 
 ** plotcf is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 ** 
 ** You should have received a copy of the GNU General Public License
 ** along with plotcf.  If not, see <http://www.gnu.org/licenses/>.
 ** 
 *******************************************************************************/

#include <QStringList>

#include "plotcf.h"
#include "wheel.h"
#include "misc.h"
#include "QsiPlot.h"
#include "readfile.h"
#include "QWheelPlot.h"
#include "plotwindow.h"

using namespace std;

/*******************************************************************************
 **
 ** main program
 **
 *******************************************************************************/

int main(int argc, char *argv[])
{

//--------------------------------------------------------
// check for no arguments specified; print usage then quit
//--------------------------------------------------------

   if (argc == 1)
   {
      cout << "Usage: plotcf [mode] [options] [files]" << endl;
      cout << "For help, type: plotcf -help" << endl << endl;
      exit(0);
   }
   else
   {
      cout << "[plotcf] This is plotcf " << VERSION << " " << "[" << PLOTCF_DATE << "]" << std::endl;
   }
   
//-----------------------
// storage of plot panels
//-----------------------

   struct panelPlotStruct plotPanels[MAX_SIZE];   
   int numPlotPanels = 0; // number of plot panels on the page

//---------------------------------------
// declarations required for user options
//---------------------------------------

   double userDefinedStartTime = -1, userDefinedStopTime = -1;
   double plotWidth = 700, plotHeight = 700;
   int userDefinedMaxVars = -1;
   QString userPlotTitle, userOnlyList;
   int userAnodeChoice = -1;
   int userMode = 0;
   int listAll = 0;
   
   QString userCminList, userCmaxList, userYminList, userYmaxList;
   QString userYlabelsList, userClabelsList;
   QString outputFileName, userPolarList, userScaleList, userColourList, userSymbolList,
           userLegendList, userPlotTitles, userSpectroTypes;
   
   bool noDisplay = false;
   bool userModifiedPlotHeight = false;
   
   int pListMultiPlots[MAX_SIZE][MAX_PLOTS_PER_PANEL] = {0};
   int pListMultiPlotsComp[MAX_SIZE][MAX_PLOTS_PER_PANEL] = {0};
   int pListMultiNumber[MAX_SIZE] = {0};
   
   int pListSymbolPlots[MAX_SIZE][MAX_PLOTS_PER_PANEL] = {0};
   
   QString pListLegends[MAX_SIZE][MAX_PLOTS_PER_PANEL] = {0};

   bool autoPlotN = false, autoPlotV = false, autoPlotNV = false, autoPlotNVT = false,
        autoPlotT = false, autoPlotNT = false;
   QString autoPlotVarg, autoPlotTarg, autoPlotVTarg;

   bool autoPlotPPA = false, autoPlotAPP = false;

   bool multipleSpacecraftPlot = false;
   bool makeWheel = false, makeSauvaud = false, makeSlice = false;

   QString userWheelType, userWheelTime, outputFileType, userMultFactors;
   QString userSliceType, userSliceWhich;

   bool scpShow = false, scpCorr = false;
   double scpOffset = 0.0;

   bool smoothTimeSeriesData = false;
   double smoothTimeSeriesSecs = 16.0;

   bool gridOnOff = false;

   QString userLineY;

   int colourBarType = 0;

   bool autoPlotPRS = false;
   
   bool rtControl = false;
   
   int showingAllPolarBins = 0;

   QString userEnergyRegion;

   bool guiOnOff = true;
   
   bool showStatus = false;

   bool inclSupportData = false;

   bool userHeights = false;
   QString userHeightsStr;

//--------------------------------
// generate text for plot metadata
//--------------------------------

   QString userName;
   char *envtmp;
   if ((envtmp = getenv("USER")) != NULL)
   {
      userName = QString(envtmp);
   }

  time_t rawtime;
  struct tm * timeinfo;

  time ( &rawtime );
  timeinfo = localtime ( &rawtime );

   QString plotMetaData;
   plotMetaData = "Generated by plotcf " + QString(VERSION) + " by " + userName
                  + " on " + QString(asctime (timeinfo));

   QString plotLabel = "plotcf " + QString(VERSION);

   int wheelUseHalfSpin = -1;

//-----------------
// search arguments
//-----------------

   int numFiles = 0;
   int fileLocArg[MAX_FILES] = {0};
   for (int i = 1; i < argc; i++)
   {
      if (strcmp(argv[i], "-help") == 0 || strcmp(argv[i], "--help") == 0) // look for help
      {
         printHelp();
         exit(0);
      }
      if (strcmp(argv[i], "-list") == 0) // look for listing
      {  
         userMode = 1;
         if ((i+1) < argc) 
         {
            if (strcmp(argv[i+1], "all") == 0) listAll = 1;
         }
      }
      if (strcmp(argv[i], "-ctable") == 0) // look for colour table type
      {
         if ((i+1) < argc) colourBarType = atoi(argv[i+1]);
      }
      if (strcmp(argv[i], "-grid") == 0) // look for grid
      {
         gridOnOff = true;
      }
      if (strcmp(argv[i], "-rt") == 0) // look for real-time control
      {
         rtControl = true;
      }      
      if (strcmp(argv[i], "-height") == 0) // look for height
      {
         userHeights = true;
         if ((i+1) < argc) userHeightsStr = QString(argv[i+1]);
      }
      if (strcmp(argv[i], "-sensor") == 0) // look for sensors to be used (previously energy region)
      {
	if ((i+1) < argc) userEnergyRegion = argv[i+1];
      }
      if (strcmp(argv[i], "-prs") == 0) // look for PEACE-RAPID spectrogram
      {
         autoPlotPRS = true;
      }
      if (strcmp(argv[i], "-status") == 0) // look for status
      {
         showStatus = true;
      }      
      if (strcmp(argv[i], "-smooth") == 0) // look for smoothing
      {
         smoothTimeSeriesData = true;
         if ((i+1) < argc) smoothTimeSeriesSecs = atof(argv[i+1]);
      }
      if (strcmp(argv[i], "-ppa") == 0) // look for para/perp/anti spectrograms
      {
         autoPlotPPA = true;
         scpShow = true;
      }
      if (strcmp(argv[i], "-app") == 0) // look for anti/perp/para spectrograms
      {
         autoPlotAPP = true;
         scpShow = true;
      }
      if (strcmp(argv[i], "-plotlabel") == 0)
      {
         if ((i+1) < argc)
         {
             plotLabel = QString(argv[i+1]);
         }
      }
      if (strcmp(argv[i], "-noplotlabel") == 0)
      {
         plotLabel = "";
      }
      if (strcmp(argv[i], "-nodisplay") == 0) // don't display plot on screen
      {
         noDisplay = true;
         guiOnOff = false;
      }
      if (strcmp(argv[i], "-n") == 0) // plot all PROPERY = Number_Density variables
      {
         autoPlotN = true;
      }
      if (strcmp(argv[i], "-t") == 0) // plot all PROPERY = Temperature variables
      {
         autoPlotT = true;
         if ((i+1) < argc) autoPlotVarg = QString(argv[i+1]);
      }
      if (strcmp(argv[i], "-v") == 0) // plot all PROPERY = Velocity variables
      {
         autoPlotV = true;
         if ((i+1) < argc) autoPlotVarg = QString(argv[i+1]);
      }
      if (strcmp(argv[i], "-halfspin") == 0) // plot this half spin oly
      {
         if ((i+1) < argc) wheelUseHalfSpin = atoi(argv[i+1]);
      }
      if (strcmp(argv[i], "-nt") == 0) // plot all Number_Density & Temperature variables
      {  
         autoPlotNT = true;
         if ((i+1) < argc) autoPlotVarg = QString(argv[i+1]);
      }
      if (strcmp(argv[i], "-nv") == 0) // plot all Number_Density & Velocity variables
      {
         autoPlotNV = true;
         if ((i+1) < argc) autoPlotVarg = QString(argv[i+1]);
      }
      if (strcmp(argv[i], "-nvt") == 0) // plot all Number_Density & Velocity & Temperature variables
      {
         autoPlotNVT = true;
         if ((i+1) < argc) autoPlotVarg = QString(argv[i+1]);
      }
      if (strcmp(argv[i], "-o") == 0) // output filename
      {
         if ((i+1) < argc) outputFileName = QString(argv[i+1]);
      }  
      if (strcmp(argv[i], "-wheel") == 0) // wheel plot
      {
         if ((i+1) < argc) userWheelType = QString(argv[i+1]);
         makeWheel = true;
      }
      if (strcmp(argv[i], "-sauvaud") == 0) // Sauvaud style plot
      {
         makeSauvaud = true;
      }
      if (strcmp(argv[i], "-slice") == 0) // make slice
      {
         if ((i+1) < argc) userSliceType = QString(argv[i+1]);
	 if ((i+2) < argc) userSliceWhich = QString(argv[i+2]);
         makeSlice = true;
      }      
      if (strcmp(argv[i], "-time") == 0) // wheel plot timestamp requested
      {  
         if ((i+1) < argc) userWheelTime = QString(argv[i+1]);
      }
      if (strcmp(argv[i], "-plot") == 0) // plot only specific things
      {
         if ((i+1) < argc) userOnlyList = QString(argv[i+1]);
      }
      if (strcmp(argv[i], "-colour") == 0) // specify plot colours
      {
         if ((i+1) < argc) userColourList = QString(argv[i+1]);
      }	  
      if (strcmp(argv[i], "-style") == 0) // specify plot symbols
      {
         if ((i+1) < argc) userSymbolList = QString(argv[i+1]);
      }	  
      if (strcmp(argv[i], "-scp-show") == 0) // show S/C potential
      {
         scpShow = true;
      }
      if (strcmp(argv[i], "-scp-corr") == 0) // S/C potential correction
      {
         if ((i+1) < argc) scpOffset = atof(argv[i+1]);
         scpCorr = true;
      }
      if (strcmp(argv[i], "-polar") == 0) // plot specific polar bins
      {
         if ((i+1) < argc) userPolarList = QString(argv[i+1]);
      }
      if (strcmp(argv[i], "-incl-support") == 0) // include support data
      {
         inclSupportData = true;
      }
      if (strcmp(argv[i], "-st") == 0) // look for start time
      {
         if ((i+1) < argc) userDefinedStartTime = time_to_tis(argv[i+1]);
      }
      if (strcmp(argv[i], "-sp") == 0) // look for stop time
      {  
         if ((i+1) < argc) userDefinedStopTime = time_to_tis(argv[i+1]);
      }
      if (strcmp(argv[i], "-yline") == 0) // look for yline
      {
         if ((i+1) < argc) userLineY = argv[i+1];
      }
      if (strcmp(argv[i], "-w") == 0) // look for plot width
      {  
         if ((i+1) < argc) plotWidth = atof(argv[i+1]);
      }
      if (strcmp(argv[i], "-h") == 0) // look for plot height
      {
         if ((i+1) < argc)
         {
            plotHeight = atof(argv[i+1]);
            userModifiedPlotHeight = true;
         }
      }
      if (strcmp(argv[i], "-cmin") == 0) // look for colour box min
      {  
         if ((i+1) < argc) userCminList = QString(argv[i+1]);
      }      
      if (strcmp(argv[i], "-cmax") == 0) // look for colour box min
      {  
         if ((i+1) < argc) userCmaxList = QString(argv[i+1]);
      }      
      if (strcmp(argv[i], "-ymin") == 0) // look for y-axis min
      {  
         if ((i+1) < argc) userYminList = QString(argv[i+1]);
      }      
      if (strcmp(argv[i], "-ymax") == 0) // look for y-axis min
      {  
         if ((i+1) < argc) userYmaxList = QString(argv[i+1]);
      }            
      if (strcmp(argv[i], "-n") == 0) // number of variables to plot
      {  
         if ((i+1) < argc) userDefinedMaxVars = atoi(argv[i+1]);
      }
      if (strcmp(argv[i], "-ylabel") == 0) // user defined y-axis labels
      {  
         if ((i+1) < argc) userYlabelsList = QString(argv[i+1]);
      }
      if (strcmp(argv[i], "-clabel") == 0) // user defined colour bar labels
      {  
         if ((i+1) < argc) userClabelsList = QString(argv[i+1]);
      }
      if (strcmp(argv[i], "-yscale") == 0) // user defined scale type
      {
         if ((i+1) < argc) userScaleList = QString(argv[i+1]);
      }  
      if (strcmp(argv[i], "-legend") == 0) // user defined legends
      {
         if ((i+1) < argc) userLegendList = QString(argv[i+1]);
      }  
      if (strcmp(argv[i], "-plabel") == 0) // user defined plot titles
      {
         if ((i+1) < argc) userPlotTitles = QString(argv[i+1]);
      }  	  
      if (strcmp(argv[i], "-sptype") == 0) // spectrogram types
      {
         if ((i+1) < argc) userSpectroTypes = QString(argv[i+1]);
      }
      if (strcmp(argv[i], "-mult") == 0) // apply multiplication factors
      {
         if ((i+1) < argc) userMultFactors = QString(argv[i+1]);
      }
      if (QString(argv[i]).endsWith(".cef", Qt::CaseInsensitive) ||
          QString(argv[i]).endsWith(".cdf", Qt::CaseInsensitive))
      {
      	 if (numFiles < MAX_FILES)
      	 {
      	    fileLocArg[numFiles] = i;
      	    numFiles++;
      	 }
      }
      if (strcmp(argv[i], "-title") == 0) // look for user plot title
      {  
         if ((i+1) < argc) userPlotTitle = QString(argv[i+1]);
      }      
      if (strcmp(argv[i], "-a") == 0) // look for anode choice
      {
         if ((i+1) < argc)
         {
            if (strcmp(argv[i+1], "all") == 0) userAnodeChoice = 0;
            else userAnodeChoice = -1;
         }
      }
   }

//----------------------------------------------
// check if the user specified any CDF/CEF files
//----------------------------------------------

   if (numFiles == 0)
   {
      cerr << "[plotcf] ERROR: at least one valid CDF or CEF file must be specified." << endl;
      exit(1);
   }

//-----------------------------------------------
// initializations for y-axis & colour bar limits
//-----------------------------------------------

   double userYminA[MAX_SIZE], userYmaxA[MAX_SIZE];
   double userCminA[MAX_SIZE], userCmaxA[MAX_SIZE];
   
   for (int i=0; i<MAX_SIZE; i++)
   {
      userYminA[i] = MY_FILL;
      userYmaxA[i] = MY_FILL;
      userCminA[i] = MY_FILL;
      userCmaxA[i] = MY_FILL;
   }

//--------------------------------
// user-defined plot panel heights
//--------------------------------

   std::vector<PlotHeightStruct> userHeightsList;
   if (userHeights)
   {
      if (userHeightsStr.contains(","))
      {
         QStringList a1 = userHeightsStr.split(",");
         for (int i=0; i<a1.size(); ++i)
         {
            QStringList a2 = a1.at(i).split(":");
            PlotHeightStruct v1;
            v1.plot = a2.at(0).toInt();
            v1.height = a2.at(1).toDouble();
            userHeightsList.push_back(v1);
         }
      }
      else
      {
         QStringList a1 = userHeightsStr.split(":");
         PlotHeightStruct v1;
         v1.plot = a1.at(0).toInt();
         v1.height = a1.at(1).toDouble();
         userHeightsList.push_back(v1);
      }
   }

//---------------------------
// user-defined y-axis limits
//---------------------------

   double yMinAll = MY_FILL, yMaxAll = MY_FILL;

   if (!userYminList.isEmpty())
   {
     if (userYminList.contains(":") || userYminList.contains(","))
     {
         QStringList a1 = userYminList.split(",");
         for (int i=0; i<a1.size(); i++)
         {
            QStringList a2 = a1.at(i).split(":");
            userYminA[a2.at(0).toInt()] = a2.at(1).toFloat();
         }
      }
      else
      {
  	 yMinAll = userYminList.toDouble();
      }
   }

   if (!userYmaxList.isEmpty())
   {
      if (userYmaxList.contains(":") || userYmaxList.contains(","))
      {
         QStringList a1 = userYmaxList.split(",");
         for (int i=0; i<a1.size(); i++)
         {
            QStringList a2 = a1.at(i).split(":");
            userYmaxA[a2.at(0).toInt()] = a2.at(1).toFloat();
         }
      }
      else
      {
 	 yMaxAll = userYmaxList.toDouble();
      }
   }

//-------------------------------
// user-defined colour bar limits
//-------------------------------

   double cMinAll = MY_FILL, cMaxAll = MY_FILL;
   
   if (!userCminList.isEmpty())
   {
      if (userCminList.contains(":") || userCminList.contains(","))
      {
         QStringList a1 = userCminList.split(",");
         for (int i=0; i<a1.size(); i++)
         {
            QStringList a2 = a1.at(i).split(":");
            userCminA[a2.at(0).toInt()] = a2.at(1).toFloat();
         }
      }
      else
      {
  	 cMinAll = userCminList.toDouble();
      }
   }

   if (!userCmaxList.isEmpty())
   {
      if (userCmaxList.contains(":") || userCmaxList.contains(","))
      {
         QStringList a1 = userCmaxList.split(",");
         for (int i=0; i<a1.size(); i++)
         {
            QStringList a2 = a1.at(i).split(":");
            userCmaxA[a2.at(0).toInt()] = a2.at(1).toFloat();
         } 
      }
      else
      {
 	 cMaxAll = userCmaxList.toDouble();
      }
   }

//---------------------------
// user-defined y-axis labels
//---------------------------

   QString userListLabelsY[MAX_SIZE];
   if (!userYlabelsList.isEmpty())
   {
      QStringList a1 = userYlabelsList.split(",");
      for (int i=0; i<a1.size(); i++)
      {
         QStringList a2 = a1.at(i).split(":");
         userListLabelsY[a2.at(0).toInt()] = a2.at(1); 	
      }
   }

//--------------------------
// user-defined y-axis lines
//--------------------------
         
   double allLinesY = MY_FILL;
   QString userLinesY[MAX_SIZE];
   if (!userLineY.isEmpty())
   {     
      if (userLineY.contains(":"))
      {
         if (userLineY.contains(","))
         {
            QStringList a1 = userLineY.split(",");
            for (int i=0; i<a1.size(); i++)
            {
               QStringList a2 = a1.at(i).split(":");
               userLinesY[a2.at(0).toInt()] = a2.at(1);
            }
         }
         else
         {
            QStringList a2 = userLineY.split(":");
            userLinesY[a2.at(0).toInt()] = a2.at(1);
         }
      }
      else
      {
         allLinesY = userLineY.toFloat();
      }
   }

//-------------------------
// user-defined plot titles
//-------------------------

   QString userListPlotTitles[MAX_SIZE];
   if (!userPlotTitles.isEmpty())
   {
      QStringList a1 = userPlotTitles.split(",");
      for (int i=0; i<a1.size(); i++)
      {
         QStringList a2 = a1.at(i).split(":");
         userListPlotTitles[a2.at(0).toInt()] = a2.at(1); 	
      }
   }

   QString userListPlotTitlesRHS[MAX_SIZE];

//-------------------------------
// user-defined colour bar labels
//-------------------------------

   QString userListLabelsC[MAX_SIZE];
   if (!userClabelsList.isEmpty())
   {
      QStringList a1 = userClabelsList.split(",");
      for (int i=0; i<a1.size(); i++)
      {
         QStringList a2 = a1.at(i).split(":");
         userListLabelsC[a2.at(0).toInt()] = a2.at(1);
      }  
   }  

//-----------
// scale type
//-----------

   QString userScaleTypeList[MAX_SIZE];
   if (!userScaleList.isEmpty())
   {
      QStringList a1 = userScaleList.split(",");
      for (int i=0; i<a1.size(); i++)
      {
         QStringList a2 = a1.at(i).split(":");
         if (a2.at(1) == "log") userScaleTypeList[a2.at(0).toInt()] = "1";
         else if (a2.at(1) == "linear") userScaleTypeList[a2.at(0).toInt()] = "0";
	 else userScaleTypeList[a2.at(0).toInt()] = "-1";
      }
   }

//----------------
// start/end times
//----------------

   if (userDefinedStartTime > -1 && userDefinedStopTime > -1)
   {
      if (userDefinedStopTime < userDefinedStartTime) userDefinedStopTime = userDefinedStopTime + 86400.0;
   }

//-----------------------------------------------------------
// check if user has specified exactly what they want plotted
//-----------------------------------------------------------

   int usersVarListSize = 0;
   if (!userOnlyList.isEmpty())
   {
      QStringList li = userOnlyList.split(",");
      for (int i=0; i<li.size(); i++)
      {
	 if (li.at(i).contains(":"))
	 {
	    QStringList liu = li.at(i).split(":");
            QStringList lli = liu.at(1).split("/");
            for (int ii=0; ii<lli.size(); ii++)
            {
	       if (lli.at(ii).contains("x")) pListMultiPlotsComp[liu.at(0).toInt()][ii] = 0;
	       else if (lli.at(ii).contains("y")) pListMultiPlotsComp[liu.at(0).toInt()][ii] = 1;
	       else if (lli.at(ii).contains("z")) pListMultiPlotsComp[liu.at(0).toInt()][ii] = 2;
	       else pListMultiPlotsComp[liu.at(0).toInt()][ii] = -1;
	       QString tt = lli.at(ii);
	       tt.remove(QRegExp("[xyz]"));
               pListMultiPlots[liu.at(0).toInt()][ii] = tt.toInt();	
            }
            pListMultiNumber[liu.at(0).toInt()] = lli.size();
         }
	 else
	 {
	    QString tt = li.at(i);
	    tt.remove(QRegExp("[xyz]"));		 
	    pListMultiPlots[i][0] = tt.toInt();
	    if (li.at(i).contains("x")) pListMultiPlotsComp[i][0] = 0;
	    else if (li.at(i).contains("y")) pListMultiPlotsComp[i][0] = 1;
	    else if (li.at(i).contains("z")) pListMultiPlotsComp[i][0] = 2;
	    else pListMultiPlotsComp[i][0] = -1;
	    pListMultiNumber[i] = 1;
         }
      }
      usersVarListSize = li.size();
   }
   else usersVarListSize = 0;

//-----------------------------------------------------------
// check if user has specified exactly what colours they want
//-----------------------------------------------------------

   QString usersColList[MAX_SIZE][MAX_PLOTS_PER_PANEL];
   int usersColListSize = 0;
   int iyc=0;
   if (!userColourList.isEmpty())
   {
      QStringList li = userColourList.split(",");
      for (int i=0; i<li.size(); i++)
      {
	 QStringList liu = li.at(i).split(":");	  
         QStringList lli = liu.at(1).split("/");
         for (int ii=0; ii<lli.size(); ii++)
         {
            usersColList[liu.at(0).toInt()][ii] = lli.at(ii);
         }
         iyc++;
      }
      usersColListSize = li.size();
   }
   else usersColListSize = 0;
   
//-----------------------------------------------------------------
// check if user has specified exactly how they want things plotted
//-----------------------------------------------------------------

   int usersSymbolListSize = 0;

   if (!userSymbolList.isEmpty())
   {
      QStringList li = userSymbolList.split(",");
      for (int i=0; i<li.size(); i++)
      {
         QStringList liu = li.at(i).split(":");
         QStringList lli = liu.at(1).split("/");
         for (int ii=0; ii<lli.size(); ii++)
         {
	    if (lli.at(ii) == "lineSolid")
	    {
               pListSymbolPlots[liu.at(0).toInt()][ii] = PLOT_LINE;
            }
            else if (lli.at(ii) == "lineDash")
            {
               pListSymbolPlots[liu.at(0).toInt()][ii] = PLOT_LINE_DASHED;
            }
            else if (lli.at(ii) == "lineDot")
            {
               pListSymbolPlots[liu.at(0).toInt()][ii] = PLOT_LINE_DOTTED;
            }
            else if (lli.at(ii) == "lineDashDot")
            {
               pListSymbolPlots[liu.at(0).toInt()][ii] = PLOT_LINE_DASH_DOT;
            }
            else if (lli.at(ii) == "lineDashDotDot")
            {
               pListSymbolPlots[liu.at(0).toInt()][ii] = PLOT_LINE_DASH_DOT_DOT;
            }
	    else if (lli.at(ii) == "circleFilled")
	    {
	       pListSymbolPlots[liu.at(0).toInt()][ii] = PLOT_CIRCLE_FILLED;
            }
            else if (lli.at(ii) == "circle")
            {
               pListSymbolPlots[liu.at(0).toInt()][ii] = PLOT_CIRCLE;
            }
	    else if (lli.at(ii) == "cross")
	    {
	       pListSymbolPlots[liu.at(0).toInt()][ii] = PLOT_CROSS;
            }
            else if (lli.at(ii) == "plus")
            {
               pListSymbolPlots[liu.at(0).toInt()][ii] = PLOT_PLUS;
            }
            else if (lli.at(ii) == "square")
            {
               pListSymbolPlots[liu.at(0).toInt()][ii] = PLOT_SQUARE;
            }
            else if (lli.at(ii) == "squareFilled")
            {
               pListSymbolPlots[liu.at(0).toInt()][ii] = PLOT_SQUARE_FILLED;
            }
            else
            {
               cerr << "ERROR: Invalid line/symbol type: " << lli.at(ii).toStdString() << endl;
               exit(1);
            }
         }
      }
      usersSymbolListSize = li.size();
   }
   else usersSymbolListSize = 0;

//-----------------------------------------------------------------
// check if user has specified legends
//-----------------------------------------------------------------

   int usersLegendListSize = 0;

   if (!userLegendList.isEmpty())
   {
      QStringList li = userLegendList.split(",");
      for (int i=0; i<li.size(); i++)
      {
	 QStringList liu = li.at(i).split(":");
         QStringList lli = liu.at(1).split("/");
         for (int ii=0; ii<lli.size(); ii++)
         {
            pListLegends[liu.at(0).toInt()][ii] = lli.at(ii);
         }
      }
      usersLegendListSize = li.size();
   }
   else usersLegendListSize = 0;

//-------------------------------
// user-defined spectrogram types
//-------------------------------
   
   int userListSpectroType[MAX_SIZE] = {0};
   if (!userSpectroTypes.isEmpty())
   {     
      QStringList a1 = userSpectroTypes.split(",");
      for (int i=0; i<a1.size(); i++)
      {     
         QStringList a2 = a1.at(i).split(":");
         if (a2.at(1) == "et") userListSpectroType[a2.at(0).toInt()] = SPECTRO_ENERGY_TIME;
         else if (a2.at(1) == "pt") userListSpectroType[a2.at(0).toInt()] = SPECTRO_PA_TIME;
         else if (a2.at(1) == "ept") userListSpectroType[a2.at(0).toInt()] = SPECTRO_ENERGY_PA_TIME;
         else
         {
            cerr << "ERROR: Invalid spectrogram type: " << a2.at(1).toStdString() << endl;
            exit(1);
         }
      }
   }  

//------------------------------------------
// handle spectrogram pitch-angle short cuts
//------------------------------------------
      
   if (autoPlotPPA)
   {  
      userPolarList = "0:0/5/11";
      if (userEnergyRegion.isEmpty() && QString(argv[fileLocArg[0]]).contains("PITCH_FULL"))
      {
         userListPlotTitles[0] = "HEEA";
         userListPlotTitles[3] = "LEEA";
      }
      else
      {
         userListPlotTitles[0] = "Parallel";
         userListPlotTitles[1] = "Perpendicular";
         userListPlotTitles[2] = "Anti-parallel";
         //if (showStatus) userListPlotTitles[3] = "Pitch Angle Selection Quality";
      }
   }     
   else if (autoPlotAPP)
   {     
      userPolarList = "0:11/5/0";
      if (userEnergyRegion.isEmpty() && QString(argv[fileLocArg[0]]).contains("PITCH_FULL"))
      {
         userListPlotTitles[0] = "HEEA";
         userListPlotTitles[3] = "LEEA";
      }
      else
      {
         userListPlotTitles[0] = "Anti-parallel";
         userListPlotTitles[1] = "Perpendicular";
         userListPlotTitles[2] = "Parallel";
         //if (showStatus) userListPlotTitles[3] = "Pitch Angle Selection Quality";
      }
   }    

//------------------------------------------------------------
// handle energy region selections for PITCH_FULL*
//------------------------------------------------------------

   if (!userEnergyRegion.isEmpty())
   {
      if (userEnergyRegion == "ol")
      {
         if (!autoPlotPPA && !autoPlotAPP)
         {
            pListMultiPlots[0][0] = 2;
            pListMultiNumber[0] = 1;
            pListMultiPlotsComp[0][0] = -1;
            usersVarListSize = 1;
         }
         else
         {
            pListMultiPlots[0][0] = 2;
            pListMultiNumber[0] = 1;
            pListMultiPlotsComp[0][0] = -1;

            pListMultiPlots[1][0] = 2;
            pListMultiNumber[1] = 1;
            pListMultiPlotsComp[1][0] = -1;

            pListMultiPlots[2][0] = 2;
            pListMultiNumber[2] = 1;
            pListMultiPlotsComp[2][0] = -1;

            usersVarListSize = 3;
         }
      }
      else if (userEnergyRegion == "oh")
      {
         if (!autoPlotPPA && !autoPlotAPP)
         {
            pListMultiPlots[0][0] = 1;
            pListMultiNumber[0] = 1;
            pListMultiPlotsComp[0][0] = -1;
            usersVarListSize = 1;
         }
         else
         {
            pListMultiPlots[0][0] = 1;
            pListMultiNumber[0] = 1;
            pListMultiPlotsComp[0][0] = -1;

            pListMultiPlots[1][0] = 1;
            pListMultiNumber[1] = 1;
            pListMultiPlotsComp[1][0] = -1;

            pListMultiPlots[2][0] = 1;
            pListMultiNumber[2] = 1;
            pListMultiPlotsComp[2][0] = -1;

            usersVarListSize = 3;
         }
      }
      else if (userEnergyRegion == "oloh")
      {
         if (!autoPlotPPA && !autoPlotAPP)
         {
            pListMultiPlots[0][0] = 1;
            pListMultiPlots[0][1] = 2;
            pListMultiNumber[0] = 2;
            pListMultiPlotsComp[0][0] = -1;
            pListMultiPlotsComp[0][1] = -1;

            usersVarListSize = 1;
         }  
         else
         {
            pListMultiPlots[0][0] = 1;
            pListMultiPlots[0][1] = 2;
            pListMultiNumber[0] = 2;
            pListMultiPlotsComp[0][0] = -1;
            pListMultiPlotsComp[0][1] = -1;
            
            pListMultiPlots[1][0] = 1;
            pListMultiPlots[1][1] = 2;
            pListMultiNumber[1] = 2;
            pListMultiPlotsComp[1][0] = -1;
            pListMultiPlotsComp[1][1] = -1;
            
            pListMultiPlots[2][0] = 1;
            pListMultiPlots[2][1] = 2;
            pListMultiNumber[2] = 2;
            pListMultiPlotsComp[2][0] = -1;
            pListMultiPlotsComp[2][1] = -1;
            
            usersVarListSize = 3;
         }  
      }  
      else if (userEnergyRegion == "b")
      {
         if (!autoPlotPPA && !autoPlotAPP)
         {
            pListMultiPlots[0][0] = 3;
            pListMultiNumber[0] = 1;
            pListMultiPlotsComp[0][0] = -1;
            usersVarListSize = 1;
         }
         else
         {
            pListMultiPlots[0][0] = 3;
            pListMultiNumber[0] = 1;
            pListMultiPlotsComp[0][0] = -1;

            pListMultiPlots[1][0] = 3;
            pListMultiNumber[1] = 1;
            pListMultiPlotsComp[1][0] = -1;

            pListMultiPlots[2][0] = 3;
            pListMultiNumber[2] = 1;
            pListMultiPlotsComp[2][0] = -1;

            usersVarListSize = 3; 
         }
      }
      else if (userEnergyRegion == "t")
      {
         if (!autoPlotPPA && !autoPlotAPP)
         {
            pListMultiPlots[0][0] = 0;
            pListMultiNumber[0] = 1;
            pListMultiPlotsComp[0][0] = -1;
            usersVarListSize = 1;
         }
         else
         {
            pListMultiPlots[0][0] = 0;
            pListMultiNumber[0] = 1;
            pListMultiPlotsComp[0][0] = -1;

            pListMultiPlots[1][0] = 0;
            pListMultiNumber[1] = 1;
            pListMultiPlotsComp[1][0] = -1;

            pListMultiPlots[2][0] = 0;
            pListMultiNumber[2] = 1;
            pListMultiPlotsComp[2][0] = -1;

            usersVarListSize = 3;
         }
      }
      else if (userEnergyRegion == "tlb")
      {
         if (!autoPlotPPA && !autoPlotAPP)
         {
            pListMultiPlots[0][0] = 0;
            pListMultiPlots[0][1] = 2;
            pListMultiPlots[0][2] = 3;
            pListMultiNumber[0] = 3;
            pListMultiPlotsComp[0][0] = -1;
            pListMultiPlotsComp[0][1] = -1;
            pListMultiPlotsComp[0][2] = -1;
            usersVarListSize = 1;
         }
         else
         {
            pListMultiPlots[0][0] = 0;
            pListMultiPlots[0][1] = 2;
            pListMultiPlots[0][2] = 3;
            pListMultiNumber[0] = 3;
            pListMultiPlotsComp[0][0] = -1;
            pListMultiPlotsComp[0][1] = -1;
            pListMultiPlotsComp[0][2] = -1;

            pListMultiPlots[1][0] = 0;
            pListMultiPlots[1][1] = 2;
            pListMultiPlots[1][2] = 3;
            pListMultiNumber[1] = 3;
            pListMultiPlotsComp[1][0] = -1;
            pListMultiPlotsComp[1][1] = -1;
            pListMultiPlotsComp[1][2] = -1;

            pListMultiPlots[2][0] = 0;
            pListMultiPlots[2][1] = 2;
            pListMultiPlots[2][2] = 3;
            pListMultiNumber[2] = 3;
            pListMultiPlotsComp[2][0] = -1;
            pListMultiPlotsComp[2][1] = -1;
            pListMultiPlotsComp[2][2] = -1;
            usersVarListSize = 3;
         }
      }
      else if (userEnergyRegion == "thb")
      {
         if (!autoPlotPPA && !autoPlotAPP)
         {
            pListMultiPlots[0][0] = 0;
            pListMultiPlots[0][1] = 1;
            pListMultiPlots[0][2] = 3;
            pListMultiNumber[0] = 3;
            pListMultiPlotsComp[0][0] = -1;
            pListMultiPlotsComp[0][1] = -1;
            pListMultiPlotsComp[0][2] = -1;
            usersVarListSize = 1;
         }
         else
         {
            pListMultiPlots[0][0] = 0;
            pListMultiPlots[0][1] = 1;
            pListMultiPlots[0][2] = 3;
            pListMultiNumber[0] = 3;
            pListMultiPlotsComp[0][0] = -1;
            pListMultiPlotsComp[0][1] = -1;
            pListMultiPlotsComp[0][2] = -1;

            pListMultiPlots[1][0] = 0;
            pListMultiPlots[1][1] = 1;
            pListMultiPlots[1][2] = 3;
            pListMultiNumber[1] = 3;
            pListMultiPlotsComp[1][0] = -1;
            pListMultiPlotsComp[1][1] = -1;
            pListMultiPlotsComp[1][2] = -1;

            pListMultiPlots[2][0] = 0;
            pListMultiPlots[2][1] = 1;
            pListMultiPlots[2][2] = 3;
            pListMultiNumber[2] = 3;
            pListMultiPlotsComp[2][0] = -1;
            pListMultiPlotsComp[2][1] = -1;
            pListMultiPlotsComp[2][2] = -1;

            usersVarListSize = 3;
         }
      }
      else if (userEnergyRegion == "tlhb")
      {     
         if (!autoPlotPPA && !autoPlotAPP)
         {  
            pListMultiPlots[0][0] = 0;
            pListMultiPlots[0][1] = 1;
            pListMultiPlots[0][2] = 2;
            pListMultiPlots[0][3] = 3;
            pListMultiNumber[0] = 4;
            pListMultiPlotsComp[0][0] = -1;
            pListMultiPlotsComp[0][1] = -1;
            pListMultiPlotsComp[0][2] = -1;
            pListMultiPlotsComp[0][3] = -1;
            usersVarListSize = 1;
         }
         else
         {
            pListMultiPlots[0][0] = 0;
            pListMultiPlots[0][1] = 1;
            pListMultiPlots[0][2] = 2;
            pListMultiPlots[0][3] = 3;
            pListMultiNumber[0] = 4;
            pListMultiPlotsComp[0][0] = -1;
            pListMultiPlotsComp[0][1] = -1;
            pListMultiPlotsComp[0][2] = -1;
            pListMultiPlotsComp[0][3] = -1;
            
            pListMultiPlots[1][0] = 0;
            pListMultiPlots[1][1] = 1;
            pListMultiPlots[1][2] = 2;
            pListMultiPlots[1][3] = 3;
            pListMultiNumber[1] = 4;
            pListMultiPlotsComp[1][0] = -1;
            pListMultiPlotsComp[1][1] = -1;
            pListMultiPlotsComp[1][2] = -1;
            pListMultiPlotsComp[1][3] = -1;
      
            pListMultiPlots[2][0] = 0;
            pListMultiPlots[2][1] = 1;
            pListMultiPlots[2][2] = 2;
            pListMultiPlots[2][3] = 3;
            pListMultiNumber[2] = 4;
            pListMultiPlotsComp[2][0] = -1;
            pListMultiPlotsComp[2][1] = -1;
            pListMultiPlotsComp[2][2] = -1;
            pListMultiPlotsComp[2][3] = -1;

            usersVarListSize = 3; 
         }
      }
      else if (userEnergyRegion == "tl")
      {
         if (!autoPlotPPA && !autoPlotAPP)
         {
            pListMultiPlots[0][0] = 0;
            pListMultiPlots[0][1] = 2;
            pListMultiNumber[0] = 2;
            pListMultiPlotsComp[0][0] = -1;
            pListMultiPlotsComp[0][1] = -1;
            usersVarListSize = 1;
         }
         else
         {
            pListMultiPlots[0][0] = 0;
            pListMultiPlots[0][1] = 2;
            pListMultiNumber[0] = 2;
            pListMultiPlotsComp[0][0] = -1;
            pListMultiPlotsComp[0][1] = -1;

            pListMultiPlots[1][0] = 0;
            pListMultiPlots[1][1] = 2;
            pListMultiNumber[1] = 2;
            pListMultiPlotsComp[1][0] = -1;
            pListMultiPlotsComp[1][1] = -1;

            pListMultiPlots[2][0] = 0;
            pListMultiPlots[2][1] = 2;
            pListMultiNumber[2] = 2;
            pListMultiPlotsComp[2][0] = -1;
            pListMultiPlotsComp[2][1] = -1;

            usersVarListSize = 3;
         }
      }
      else if (userEnergyRegion == "th")
      {
         if (!autoPlotPPA && !autoPlotAPP)
         {
            pListMultiPlots[0][0] = 0;
            pListMultiPlots[0][1] = 1;
            pListMultiNumber[0] = 2;
            pListMultiPlotsComp[0][0] = -1;
            pListMultiPlotsComp[0][1] = -1;
            usersVarListSize = 1;
         }
         else
         {
            pListMultiPlots[0][0] = 0;
            pListMultiPlots[0][1] = 1;
            pListMultiNumber[0] = 2;
            pListMultiPlotsComp[0][0] = -1;
            pListMultiPlotsComp[0][1] = -1;

            pListMultiPlots[1][0] = 0;
            pListMultiPlots[1][1] = 1;
            pListMultiNumber[1] = 2;
            pListMultiPlotsComp[1][0] = -1;
            pListMultiPlotsComp[1][1] = -1;

            pListMultiPlots[2][0] = 0;
            pListMultiPlots[2][1] = 1;
            pListMultiNumber[2] = 2;
            pListMultiPlotsComp[2][0] = -1;
            pListMultiPlotsComp[2][1] = -1;

            usersVarListSize = 3;
         }
      }
   }
   
//----------------------------------------------------------
// create the required ReadFile's and open the CDF/CEF files
//----------------------------------------------------------
   
   ReadFile caaRead[MAX_FILES];

// open file(s)

   for (int i=0; i<numFiles; i++)
   {
      if (caaRead[i].open(argv[fileLocArg[i]]) == _READ_FAILED)
      {
         cerr << "ERROR opening file: " << argv[fileLocArg[i]] << std::endl;
         exit(1);
      }   
   }

//------------------------------------------------
// read the header(s) & extract useful information
//------------------------------------------------

   QString beginDateTime[MAX_FILES], endDateTime[MAX_FILES];
   double beginDateTimeDec[MAX_FILES], endDateTimeDec[MAX_FILES];
   double beginDateTimeUsed = -1, endDateTimeUsed = -1;
   QString beginDateUsed;

   QString dataSetTitle, gbObservatory, gbExperiment, dataSetIdArr[MAX_FILES];

   cout << "[plotcf] reading header(s)..." << endl;

   for (int i=0; i<numFiles; i++)
   {
 
 // read in the header
 
      caaRead[i].readHeader();
      
// get the DATASET_TITLE, OBSERVATORY, EXPERIMENT from the first file      

      dataSetIdArr[i] = caaRead[i].getGlobalAttr("DATASET_ID");
      if (i==0)
      {
 	 dataSetTitle = caaRead[i].getGlobalAttr("DATASET_TITLE");
         gbObservatory = caaRead[i].getGlobalAttr("OBSERVATORY");
         gbExperiment = caaRead[i].getGlobalAttr("EXPERIMENT");
      }
      else
      {
	 if (gbObservatory != caaRead[i].getGlobalAttr("OBSERVATORY"))
	 {
	    multipleSpacecraftPlot = true;
	 }
      }
 
 // read in the begin & end times
      
      beginDateTime[i] = caaRead[i].getFileBeginDateTime();
      endDateTime[i] = caaRead[i].getFileEndDateTime();
      
      beginDateTimeDec[i] = time_to_tis((char *)(beginDateTime[i].mid(11, 12).toStdString().c_str()));
      endDateTimeDec[i] = time_to_tis((char *)(endDateTime[i].mid(11, 12).toStdString().c_str()));

 // determine the begin & end times to use in the plot
      
      if (i==0)
      {
         beginDateTimeUsed = beginDateTimeDec[i];
         beginDateUsed = beginDateTime[i];
         endDateTimeUsed = endDateTimeDec[i];	
      }
      else
      {
         if (beginDateTimeDec[i] < beginDateTimeUsed)
         {
            beginDateTimeUsed = beginDateTimeDec[i];
            beginDateUsed = beginDateTime[i];
         }
         if (endDateTimeDec[i] > endDateTimeUsed)
         {
            endDateTimeUsed = endDateTimeDec[i];	
         }
      }     
   }

   cout << "[plotcf] determining what to plot..." << endl;
   
   if (endDateTimeUsed < beginDateTimeUsed) endDateTimeUsed = endDateTimeUsed + 86400.0;
  
   int numData = 0;
   
   int pListPanels[MAX_SIZE] = {0};
   
   int fileHasSpectro[MAX_FILES] = {0};
   
   int pListPolarOnly[MAX_SIZE][20] = {0}; // asumming no more than 20 polar bins possible
   
// initialization

   for (int i=0; i<MAX_SIZE; i++)
   {
      plotPanels[i].numTimeSeries = 0;
      plotPanels[i].numSpectrograms = 0;
      plotPanels[i].yAxisMin = MY_FILL;
      plotPanels[i].yAxisMax = MY_FILL;
   }   

//-----------------------
// multiplication factors
//-----------------------

   double userMultFactorsListSp[MAX_SIZE][MAX_PLOTS_PER_PANEL] = {0};
   double userMultFactorsListTs[MAX_SIZE][MAX_PLOTS_PER_PANEL] = {0};

   double userMultFactorsListSpY[MAX_SIZE][MAX_PLOTS_PER_PANEL] = {0};

//--------------------------------------------
// special treatment for PITCH_FULL & PITCH_3D
//--------------------------------------------

   int specialModeFlag = -1;

   if ((numFiles == 1 && !autoPlotPPA && !autoPlotAPP && 
       userEnergyRegion.isEmpty() && userPolarList.isEmpty()) &&
       (dataSetIdArr[0].contains("PITCH_FULL")))
   {
      specialModeFlag = 0;

      // we want a single colour box covering all 4 plot panels, and labels indicating
      // the different energy regions

      userListPlotTitles[0] = "HEEA";
      userListPlotTitles[1] = "LEEA";
   }
   else if (numFiles == 1 && (autoPlotPPA || autoPlotAPP) &&
       userEnergyRegion.isEmpty() && 
       (dataSetIdArr[0].contains("PITCH_FULL")))
   {
      userListLabelsY[0] = "E (eV)";
      userListLabelsY[3] = "E (eV)";

      if (autoPlotPPA)
      {
         userListPlotTitlesRHS[0] = "0";
         userListPlotTitlesRHS[1] = "90";
         userListPlotTitlesRHS[2] = "180";
         userListPlotTitlesRHS[3] = "0";
         userListPlotTitlesRHS[4] = "90";
         userListPlotTitlesRHS[5] = "180";
      }
      else if (autoPlotAPP)
      {
         userListPlotTitlesRHS[0] = "180";
         userListPlotTitlesRHS[1] = "90";
         userListPlotTitlesRHS[2] = "0";
         userListPlotTitlesRHS[3] = "180";
         userListPlotTitlesRHS[4] = "90";
         userListPlotTitlesRHS[5] = "0";
      }
   }

//--------------------------------------------------
// handle combined PEACE-RAPID spectrogram short-cut
//--------------------------------------------------
   
   int iPrsPeace = -1, iPrsRapid = -1;
   if (autoPlotPRS)
   {
   
// find the PEACE file(s) & RAPID files(s)

      for (int i=0; i<numFiles; i++)
      {
         if (dataSetIdArr[i].contains("PITCH_SPIN"))
         {
            iPrsPeace = i;
         }
         else if (dataSetIdArr[i].contains("ESPCT"))
         {
            iPrsRapid = i;
         }
      }

      pListMultiPlots[0][0] = iPrsPeace;
      pListMultiPlots[0][1] = iPrsRapid;
      pListMultiNumber[0] = 2;
      usersVarListSize = 1;

// multiplication factors to ensure units are the same for both
// PEACE and RAPID

      userMultFactorsListSp[0][0] = 6.24150974E+15;
      userMultFactorsListSp[0][1] = 6.242E+19;

      userMultFactorsListSpY[0][1] = 1000.0;

// y-axis & c-axis labels

      userListLabelsY[0] = "energy (eV)";
      userListLabelsC[0] = "Diff. energy flux ()";
 
   }

//-------------------------------------------------------------------
// for auto moments plots, we need to determine what variables to use
//-------------------------------------------------------------------

   int nLoc = 0, vLoc = 0, tLoc = 0, sLoc = 0, bLoc = 0;
   if (autoPlotNV)
   {
      vLoc = 1;
      bLoc = 4;
   }
   if (autoPlotNVT)
   {
      vLoc = 1;
      tLoc = 4;
      bLoc = 5;
   }
   if (autoPlotNT)
   {
      tLoc = 1;
      bLoc = 2;
   }
   if (autoPlotN)
   {
      nLoc = 0;
      bLoc = 1;
   }
   if (autoPlotT)
   {
      tLoc = 0;
      bLoc = 1;
   }
   if (autoPlotV)
   {
      vLoc = 0;
      bLoc = 3;
   }

   int iPreCount = 0, iPreCountN = 0, iPreCountV = 0, iPreCountT;
   int iFoundN = 0, iFoundV = 0, iFoundT = 0, iFoundS = 0, iFoundB = 0;

   // store file, variable number, data type, fill value of pitch angle selection quality flag
   int PitchAngleQuality[3] = {-1, -1, -1};
   double PitchAngleQualityFill = 0;

   QString suggestedLegendsN[MAX_PLOTS_PER_PANEL];
   QString suggestedLegendsV[MAX_PLOTS_PER_PANEL];
   QString suggestedLegendsT[MAX_PLOTS_PER_PANEL];
   QString arrObservatory[MAX_FILES], arrExperiment[MAX_FILES];

   bool autoPlotOnlyOne = true;
   
   if (!makeWheel && !makeSauvaud && !makeSlice) // the checks below are not required for wheel plots or Sauvaud plots
   {
      for (int j = 0; j < numFiles; j++) // <-- loop over all CDF/CEF files specified by user
      {      
         QString DATASET_ID = caaRead[j].getGlobalAttr("DATASET_ID");
         QString OBSERVATORY = caaRead[j].getGlobalAttr("OBSERVATORY");
         arrObservatory[j] = OBSERVATORY;
         QString EXPERIMENT = caaRead[j].getGlobalAttr("EXPERIMENT");
         arrExperiment[j] = EXPERIMENT;
         for (int i=0; i<caaRead[j].getNumVariables(); i++) // <-- loop over variables in each CDF/CEF file
         {     	    	
 	    QString PROPERTY = caaRead[j].getVariableAttr(i, "PROPERTY");
            QString PARAMETER_TYPE = caaRead[j].getVariableAttr(i, "PARAMETER_TYPE");
            QString FIELDNAM = caaRead[j].getVariableAttr(i, "FIELDNAM");	 
            QString NAME = caaRead[j].getVariableAttr(i, "NAME");
            QString DISPLAYTYPE = caaRead[j].getVariableAttr(i, "DISPLAYTYPE");
            QString ENTITY = caaRead[j].getVariableAttr(i, "ENTITY");           

            if (PARAMETER_TYPE.startsWith("DATA", Qt::CaseInsensitive) ||
                (PARAMETER_TYPE.startsWith("SUPPORT_DATA", Qt::CaseInsensitive) && !NAME.contains("LABEL") &&
                 !NAME.contains("REPRESENTATION") && NAME.startsWith("Status_")))
	    {
               if ((autoPlotN || autoPlotNV || autoPlotV || autoPlotNVT) &&
                   DISPLAYTYPE.startsWith("Spectrogram", Qt::CaseInsensitive))
               {
                  iFoundS++;
                  pListMultiPlots[sLoc][0] = iPreCount;
                  pListMultiNumber[sLoc] = 1;
                  nLoc++; vLoc++, tLoc++; bLoc++;
                  scpShow = true; // ensure S/C potential is overlaid on spectrogram if a file is specified
               }
	       if ((autoPlotN || autoPlotNV || autoPlotNVT || autoPlotNT) &&
                   PROPERTY.contains("Density", Qt::CaseInsensitive))
	       {
	          iFoundN++;
	          pListMultiPlots[nLoc][iPreCountN] = iPreCount;
                  suggestedLegendsN[iPreCountN] = OBSERVATORY + " > " + EXPERIMENT + " > " + FIELDNAM;     
                  iPreCountN++;
                  pListMultiNumber[nLoc] = iPreCountN;
	       }
	       if ((autoPlotV || autoPlotNV || autoPlotNVT) &&
                   PROPERTY.startsWith("Velocity", Qt::CaseInsensitive) &&
                   !NAME.contains("Velocity_Magnitude"))
	       {
	          iFoundV++;
                  if ((FIELDNAM.contains("Para", Qt::CaseInsensitive) && autoPlotVarg == "para") ||
                      (FIELDNAM.contains("Perp", Qt::CaseInsensitive) && autoPlotVarg == "perp") || 
                      (!FIELDNAM.contains("Para", Qt::CaseInsensitive) &&
                       !FIELDNAM.contains("Perp", Qt::CaseInsensitive) && 
		       autoPlotVarg != "para" && autoPlotVarg != "perp" && autoPlotVarg != "paraperp") ||
                      (autoPlotVarg == "paraperp" && (FIELDNAM.contains("Para", Qt::CaseInsensitive) ||
                                                      FIELDNAM.contains("Perp", Qt::CaseInsensitive))))
	          {
                     pListMultiPlots[vLoc][iPreCountV] = iPreCount;
                     pListMultiPlots[vLoc+1][iPreCountV] = iPreCount;
                     pListMultiPlots[vLoc+2][iPreCountV] = iPreCount;
                     pListMultiPlotsComp[vLoc][iPreCountV] = 0;
                     pListMultiPlotsComp[vLoc+1][iPreCountV] = 1;
                     pListMultiPlotsComp[vLoc+2][iPreCountV] = 2;
                     suggestedLegendsV[iPreCountV] = OBSERVATORY + " > " + EXPERIMENT + " > " + FIELDNAM;
                     iPreCountV++;
                     pListMultiNumber[vLoc] = iPreCountV;
                     pListMultiNumber[vLoc+1] = iPreCountV;
                     pListMultiNumber[vLoc+2] = iPreCountV;
                  }	      
	       }
               if ((autoPlotT || autoPlotNVT || autoPlotNT) &&
                   PROPERTY.startsWith("Temperature", Qt::CaseInsensitive))
               {
                  iFoundT++;
                  if ((FIELDNAM.contains("Para", Qt::CaseInsensitive) && autoPlotVarg == "para") ||
                      (FIELDNAM.contains("Perp, Qt::CaseInsensitive") && autoPlotVarg == "perp") ||
		      (!FIELDNAM.contains("Para", Qt::CaseInsensitive) &&
                       !FIELDNAM.contains("Perp", Qt::CaseInsensitive) &&
		       autoPlotVarg != "para" && autoPlotVarg != "perp" & autoPlotVarg != "paraperp") ||
                      (autoPlotVarg == "paraperp" && (FIELDNAM.contains("Para", Qt::CaseInsensitive) ||
                                                      FIELDNAM.contains("Perp", Qt::CaseInsensitive))))
		  {
                     pListMultiPlots[tLoc][iPreCountT] = iPreCount;
                     suggestedLegendsT[iPreCountT] = OBSERVATORY + " > " + EXPERIMENT + " > " + FIELDNAM;   
 
                     iPreCountT++;
                     pListMultiNumber[tLoc] = iPreCountT;
                  }

               }
               if ((autoPlotN || autoPlotV || autoPlotT || autoPlotNT || autoPlotNV || autoPlotNV) &&
                   DATASET_ID.contains("FGM_SPIN", Qt::CaseInsensitive) && PROPERTY == "Vector" &&
                   ENTITY == "Magnetic_Field")
               {
                  pListMultiPlots[bLoc][0] = iPreCount;
                  pListMultiPlots[bLoc][1] = iPreCount;
                  pListMultiPlots[bLoc][2] = iPreCount;
                  pListMultiPlotsComp[bLoc][0] = 0;
                  pListMultiPlotsComp[bLoc][1] = 1;
                  pListMultiPlotsComp[bLoc][2] = 2;
                  pListMultiNumber[bLoc] = 3;
                  iFoundB++;
               }
	       iPreCount++;
	    }
            if (NAME.startsWith("Status_P") && NAME.contains("SelectionQuality__") &&
                     !NAME.contains("LABEL"))
            {
               PitchAngleQuality[0] = j; // file
               PitchAngleQuality[1] = i; // variable
            }
         }
      }

      if (autoPlotN || autoPlotV || autoPlotNV || autoPlotNVT || autoPlotNT) usersVarListSize = 0;
      if (iFoundN > 0 && (autoPlotN || autoPlotNV || autoPlotNVT || autoPlotNT)) usersVarListSize++;
      if (iFoundV > 0 && (autoPlotV || autoPlotNV || autoPlotNVT)) usersVarListSize += 3;
      if (iFoundT > 0 && (autoPlotV || autoPlotT || autoPlotNVT || autoPlotNT)) usersVarListSize++;
      if (iFoundS > 0) usersVarListSize++;
      if (iFoundB > 0) usersVarListSize++;

      // now set nice y-axis labels

      if (userYlabelsList.isEmpty())
      {
         if (autoPlotN || autoPlotNV || autoPlotNVT || autoPlotNT) userListLabelsY[nLoc] = "N (1/cm^3)";
         if (autoPlotV || autoPlotNV || autoPlotNVT) userListLabelsY[vLoc] = "Vx (km/s)";
         if (autoPlotV || autoPlotNV || autoPlotNVT) userListLabelsY[vLoc+1] = "Vy (km/s)";
         if (autoPlotV || autoPlotNV || autoPlotNVT) userListLabelsY[vLoc+2] = "Vz (km/s)";
         if (autoPlotT || autoPlotNVT || autoPlotNT) userListLabelsY[tLoc] = "T (MK)";
         if (iFoundS > 0) userListLabelsY[sLoc] = "energy (eV)";
         if (iFoundB > 0) userListLabelsY[bLoc] = "B (nT)";
      }

      // adjust colour-axis label if necessary

      if (userClabelsList.isEmpty())
      {
         if (iFoundS > 0 && iFoundV > 0) userListLabelsC[sLoc] = "Diff. Energy Flux";
      }

      // now set y-axis scale types (all to linear)

      if (userScaleList.isEmpty())
      {
         if (autoPlotN || autoPlotNV || autoPlotNVT || autoPlotNT) userScaleTypeList[nLoc] = "0";
         if (autoPlotV || autoPlotNV || autoPlotNVT) userScaleTypeList[vLoc] = "0";
         if (autoPlotV || autoPlotNV || autoPlotNVT) userScaleTypeList[vLoc+1] = "0";
         if (autoPlotV || autoPlotNV || autoPlotNVT) userScaleTypeList[vLoc+2] = "0";
         if (autoPlotT || autoPlotNVT || autoPlotNT) userScaleTypeList[tLoc] = "0";
         if (iFoundS > 0) userScaleTypeList[sLoc] = "1";
         if (iFoundB > 0) userScaleTypeList[bLoc] = "0";
      }

      // now set nice legends if the user hasn't specified any

      if (userLegendList.isEmpty())
      {
         usersLegendListSize = 0;
         if (iFoundN > 0 && (autoPlotN || autoPlotNV || autoPlotNVT || autoPlotNT)) usersLegendListSize++;
         if (iFoundV > 0 && (autoPlotV || autoPlotNVT)) usersLegendListSize += 3;
         if (iFoundT > 0 && (autoPlotV || autoPlotT || autoPlotNVT || autoPlotNT)) usersLegendListSize++; 
         if (iFoundB > 0) usersLegendListSize++;

         if (autoPlotN || autoPlotNV || autoPlotNVT || autoPlotNT)
         {
            for (int k=0; k<iPreCountN; k++)
	    {
	       pListLegends[nLoc][k] = suggestedLegendsN[k];
	    }
         }
         if (autoPlotV || autoPlotNV || autoPlotNVT)
         {
            for (int k=0; k<iPreCountV; k++)
	    {
	       pListLegends[vLoc][k] = suggestedLegendsV[k];
	       pListLegends[vLoc+1][k] = suggestedLegendsV[k];
	       pListLegends[vLoc+2][k] = suggestedLegendsV[k];
	    }
         }
         if (autoPlotT || autoPlotNVT || autoPlotNT)
         {
            for (int k=0; k<iPreCountT; k++)
            {
               pListLegends[tLoc][k] = suggestedLegendsT[k];
            }
         }
         if ((autoPlotN || autoPlotV || autoPlotT || autoPlotNV || autoPlotNT || autoPlotNVT) &&
             iFoundB > 0)
         {
            pListLegends[bLoc][0] = "Bx";
            pListLegends[bLoc][1] = "By";
            pListLegends[bLoc][2] = "Bz";
         }
      }

// check if all plots are from the same spacecraft
// - we don't want to specify a plot title containing a S/C number or experiment
//   if different S/C or different experiments are specified

      if (numFiles > 1)
      {
         for (int j=1; j<numFiles; j++)
         {
            if (arrObservatory[j] != arrObservatory[j-1]) autoPlotOnlyOne = false;
            if (arrExperiment[j] != arrExperiment[j-1]) autoPlotOnlyOne = false;
         }
      }

// add pitch angle selection quality

   if (autoPlotPPA || autoPlotAPP)
   {

   }

   } // <-- end if not a wheel plot or Sauvaud plot

//-------------------------------------------------------------------   
// read useful information from the header and determine what to plot
//-------------------------------------------------------------------   

   int iCount = 0; // counter over all DATA
   int numVariables[MAX_FILES] = {0}; // number of variables in each file

   bool ignoreThisVar = false; // for S/C potential modes, we might want to ignore the potential & use it as we require later

// required additional data for -scp-show option

   int scpFile[MAX_DISTINCT_SPACECRAFT] = {0}, scpVar[MAX_DISTINCT_SPACECRAFT] = {0}; // location of S/C potential data for each S/C
   int scpType[MAX_DISTINCT_SPACECRAFT] = {0};    // S/C potential data type
   double scpFill[MAX_DISTINCT_SPACECRAFT] = {0}; // S/C potential fill value
   int scpAssign[MAX_DISTINCT_SPACECRAFT] = {0};  // S/C potential assigned

// required additional data for -scp-corr option

   int scpFileCorr[MAX_DISTINCT_SPACECRAFT] = {0}, scpVarCorr[MAX_DISTINCT_SPACECRAFT] = {0};
   int scpTypeCorr[MAX_DISTINCT_SPACECRAFT] = {0};    // S/C potential data type
   double scpFillCorr[MAX_DISTINCT_SPACECRAFT] = {0}; // S/C potential fill value
   int scpAssignCorr[MAX_DISTINCT_SPACECRAFT] = {0};  // S/C potential assigned or not
   int scpDataRefCorr[MAX_DISTINCT_SPACECRAFT] = {0}; // reference to location of data
   
   for (int j = 0; j < numFiles; j++) // <-- loop over all CDF/CEF files specified by user
   {
      QString dataSetID = caaRead[j].getGlobalAttr("DATASET_ID");

      QString observatory = caaRead[j].getGlobalAttr("OBSERVATORY");
      int clScN = -1;
      if (!observatory.isEmpty()) clScN = observatory.mid(observatory.size()-1,1).toInt();
      numVariables[j] = caaRead[j].getNumVariables();
      
      for (int i=0; i<numVariables[j]; i++) // <-- loop over variables in each CDF/CEF file
      {     	    	
         QString name = caaRead[j].getVariableAttr(i, "NAME");
         QString scaleType = caaRead[j].getVariableAttr(i, "SCALETYP");
         DataType valueType = caaRead[j].getVariableDataType(i);
         QString displayType = caaRead[j].getVariableAttr(i, "DISPLAYTYPE");
         QString parameterType = caaRead[j].getVariableAttr(i, "PARAMETER_TYPE");
         QString varType = caaRead[j].getVariableAttr(i, "VAR_TYPE");
         QString labelAxis = caaRead[j].getVariableAttr(i, "LABLAXIS");
       	 QString property = caaRead[j].getVariableAttr(i, "PROPERTY");
         QString units = caaRead[j].getVariableAttr(i, "UNITS");
	 QString field = caaRead[j].getVariableAttr(i, "FIELDNAM");
         QString dataset_id = dataSetID;
                   
         float scaleMin = caaRead[j].getVariableAttr(i, "SCALEMIN").toFloat();
         float scaleMax = caaRead[j].getVariableAttr(i, "SCALEMAX").toFloat();
         float fillValue = caaRead[j].getVariableAttr(i, "FILLVAL").toFloat();

         QString yLabelDefault, cLabelDefault;

// current data type

         int currentDataType = valueType;
 
 // determine y-axis scale type, based on info in meta-data as well as user request
         
         int yAxisScaleType = 0;
         if (scaleType.startsWith("LOG", Qt::CaseInsensitive)) yAxisScaleType = 1;   

 // pitch angle status quality - we already know the file number & variable number,
 // we just need to get the data type & fill value

         if (j == PitchAngleQuality[0] && i == PitchAngleQuality[1])
         {
            PitchAngleQuality[3] = currentDataType;
            PitchAngleQualityFill = fillValue;
         }
         
 // determine if spectrogram or time-series
  
         int currentPlotType = 0, numPolarsInFile = -1;
         if (displayType.startsWith("Spectrogram", Qt::CaseInsensitive) || isSpectro(dataSetID))
         {
            currentPlotType = SPECTROGRAM_PLOT;
	    fileHasSpectro[j] = 1;

// determine how many polar bins there are

            if (!userPolarList.isEmpty())
            {
	       QString tDEPEND[4];
               tDEPEND[0] = caaRead[j].getVariableAttr(i, "DEPEND_0");
               tDEPEND[1] = caaRead[j].getVariableAttr(i, "DEPEND_1");
               tDEPEND[2] = caaRead[j].getVariableAttr(i, "DEPEND_2");
               tDEPEND[3] = caaRead[j].getVariableAttr(i, "DEPEND_3");       
         
               int izPolar = -1;
               for (int ib=0; ib<4; ib++)
               {
                  if (tDEPEND[ib].contains("Polar", Qt::CaseInsensitive) ||
                      tDEPEND[ib].contains("Theta", Qt::CaseInsensitive) ||
                      tDEPEND[ib].contains("Sweep_Pitch", Qt::CaseInsensitive))
                  {
                     izPolar = ib;	
                  }	
               }    
	       numPolarsInFile = caaRead[j].getVariableSize(i, izPolar-1);  	
			   	
               QStringList a1 = userPolarList.split(",");
               for (int ix=0; ix<a1.size(); ix++)
               {
                  QStringList a2 = a1.at(ix).split(":");
                  if (a2.at(0).toInt() == numData)
                  {
                     if (a2.at(1) == "all")
                     {
                        pListPanels[numPlotPanels] = numPolarsInFile;	
                        for (int iy=0;iy<numPolarsInFile; iy++)
                        {
                           pListPolarOnly[numPlotPanels][iy] = iy;	
                        }               
			showingAllPolarBins = numPolarsInFile;  	     
                     }
                     else
		     {		        
		        QStringList a3 = a2.at(1).split("/");
                        pListPanels[numPlotPanels] = a3.size();
                        for (int iz=0; iz < a3.size(); iz++)
			{
			   pListPolarOnly[numPlotPanels][iz] = a3.at(iz).toInt();
		        }
		     }
		  }
	       }					
	    }
					
// determine default c-axis label

            if (labelAxis.isEmpty() || labelAxis == "NULL")
            {
               cLabelDefault = property + " (" + units + ")";	
            }
            else
            {
               cLabelDefault = labelAxis + " (" + units + ")";	
            }                                      
         }
         else
         {
             currentPlotType = TIMESERIES_PLOT;
         	
// determine default y-axis label

             if (!labelAxis.isEmpty() && labelAxis != "NULL") yLabelDefault = labelAxis + " (" + units + ")";
             else if (!field.isEmpty() && field != "NULL") yLabelDefault = field + " (" + units + ")";  
             else yLabelDefault = "(" + units + ")";
         }
 
// add (if necessary) to plot panel list
         
	 if ((((parameterType.startsWith("DATA", Qt::CaseInsensitive) && !name.contains("_SD_")) ||
             varType.startsWith("DATA", Qt::CaseInsensitive)) && !name.startsWith("BackgroundLevel") ||
             (parameterType.startsWith("SUPPORT_DATA", Qt::CaseInsensitive) && 
              !name.contains("LABEL", Qt::CaseInsensitive) && name.startsWith("Status_"))) &&
              !name.contains("LABEL", Qt::CaseInsensitive) &&
              !name.contains("REPRESENTATION", Qt::CaseInsensitive))
	 {			           				  

            if (userMode == 1) // print variables names (and file names) if requested by user
            {
               if (!parameterType.startsWith("SUPPORT_DATA", Qt::CaseInsensitive) ||
                   (parameterType.startsWith("SUPPORT_DATA", Qt::CaseInsensitive) && (listAll == 1 || inclSupportData)))
               {
                  cout << "   [" << iCount << "] " << name.toStdString() << endl;
                  cout << "       from file: " << argv[fileLocArg[j]] << endl;
               }
            }  

	    // is this an EFW spacecraft potential?

            if ((name.contains("CP_EFW_L3_P") || name.contains("CP_EFW_L2_P")) && numFiles > 1)
	    {
	       ignoreThisVar = true;
               if (scpShow)
               {
                  scpFile[clScN-1] = j;
                  scpVar[clScN-1] = i;
                  scpType[clScN-1] = currentDataType;
                  scpFill[clScN-1] = fillValue;
                  scpAssign[clScN-1] = 1;
               }
               else if (scpCorr)
               {
                  scpFileCorr[clScN-1] = j;
                  scpVarCorr[clScN-1] = i;
                  scpTypeCorr[clScN-1] = currentDataType;
                  scpFillCorr[clScN-1] = fillValue;
                  scpAssignCorr[clScN-1] = 1;
               }
	    }
	    else ignoreThisVar = false; // we need to ignore the EFW potential now as we don't want to
                                        // plot it in it's own panel

            // determine the total size of vectors/tensors

            int iVarTotalSize = 0;
	    int iVariableDims = caaRead[j].getVariableDimension(i);
            for (int ja=0; ja<iVariableDims; ja++)
	    {
               int thisDimSize = caaRead[j].getVariableSize(i, ja);
	       for (int jb=0; jb<thisDimSize; jb++)
               {
	          iVarTotalSize++;
	       }
	    }

	    if (usersVarListSize == 0) // plot all variables since user hasn't specified any
            {
               plotPanels[numPlotPanels].yAxisScaleType = 
	          (userScaleTypeList[numPlotPanels].isEmpty()) ? yAxisScaleType : userScaleTypeList[numPlotPanels].toInt();                        				
               if (currentPlotType == TIMESERIES_PLOT && !ignoreThisVar &&
                   (!parameterType.startsWith("SUPPORT_DATA", Qt::CaseInsensitive) ||
                   inclSupportData && parameterType.startsWith("SUPPORT_DATA", Qt::CaseInsensitive)))
               {						   
                  for (int ja=0; ja<iVarTotalSize; ja++)
		  {
                     plotPanels[numPlotPanels].tsFile[ja] = j;
                     plotPanels[numPlotPanels].tsVar[ja] = i;	
		     plotPanels[numPlotPanels].tsType[ja] = currentDataType;
		     plotPanels[numPlotPanels].tsFillValue[ja] = fillValue;  		
		     plotPanels[numPlotPanels].tsVarComp[ja] = ja;	   
                     plotPanels[numPlotPanels].multFactor[ja] = 1.0;  
                     plotPanels[numPlotPanels].tsDATASET_ID[ja] = dataset_id;
		  }
			   
                  plotPanels[numPlotPanels].numTimeSeries = iVarTotalSize;
                  plotPanels[numPlotPanels].numSpectrograms = 0;
                  plotPanels[numPlotPanels].yAxisMin = (userYminA[numPlotPanels] == MY_FILL) ? scaleMin : userYminA[numPlotPanels];
                  plotPanels[numPlotPanels].yAxisMax = (userYmaxA[numPlotPanels] == MY_FILL) ? scaleMax : userYmaxA[numPlotPanels];
                  plotPanels[numPlotPanels].yLabel = 
                     (userListLabelsY[numPlotPanels].isEmpty()) ? yLabelDefault : userListLabelsY[numPlotPanels];
				  numPlotPanels++;
               }
               else if (currentPlotType == SPECTROGRAM_PLOT)
               {
	          int iPmax = (pListPanels[numPlotPanels] > 0) ? pListPanels[numPlotPanels] : 1;
		  int iPmaster = (pListPanels[numPlotPanels] > 0) ? numPlotPanels : -1;
				  
		  for (int ja=0; ja < iPmax; ja++)
		  {
                     plotPanels[numPlotPanels].numTimeSeries = 0;
                     plotPanels[numPlotPanels].numSpectrograms = 1; 
                     plotPanels[numPlotPanels].spFile[0] = j;
                     plotPanels[numPlotPanels].spVar[0] = i;     
                     plotPanels[numPlotPanels].multFactorSp[0] = 1.0;
                     plotPanels[numPlotPanels].yAxisMin = (userYminA[numPlotPanels] == MY_FILL) ? 1 : userYminA[numPlotPanels];
                     plotPanels[numPlotPanels].yAxisMax = (userYmaxA[numPlotPanels] == MY_FILL) ? 800 : userYmaxA[numPlotPanels];	   
                     plotPanels[numPlotPanels].cBarMin = (userCminA[numPlotPanels] == MY_FILL) ? scaleMin : userCminA[numPlotPanels];
                     plotPanels[numPlotPanels].cBarMax = (userCmaxA[numPlotPanels] == MY_FILL) ? scaleMax : userCmaxA[numPlotPanels];    
		     plotPanels[numPlotPanels].tsType[0] = currentDataType;
                     plotPanels[numPlotPanels].yLabel = 
                        (userListLabelsY[numPlotPanels].isEmpty()) ? yLabelDefault : userListLabelsY[numPlotPanels];
                     plotPanels[numPlotPanels].cLabel =
                        (userListLabelsC[numPlotPanels].isEmpty()) ? cLabelDefault : userListLabelsC[numPlotPanels];  
		     plotPanels[numPlotPanels].spFillValue[0] = fillValue;
                     plotPanels[numPlotPanels].spDATASET_ID[0] = dataset_id;
                     plotPanels[numPlotPanels].scN[0] = clScN;

		     // handle multiple polar plots

		     plotPanels[numPlotPanels].spPolarComp = pListPolarOnly[iPmaster][ja];
		     if (ja == 0 && iPmaster != -1)
		     {
		        plotPanels[numPlotPanels].spMaster = iPmax;
			plotPanels[numPlotPanels].spSlave = 0;
		     }
		     else if (iPmaster == -1)
		     {
		        plotPanels[numPlotPanels].spMaster = -2;
			plotPanels[numPlotPanels].spSlave = 0;					 	
		     }
		     else
		     {
		        plotPanels[numPlotPanels].spMaster = -1;
			plotPanels[numPlotPanels].spSlave = 1;
		     }
		     numPlotPanels++;
		  }
               }         	
            }
            else // plot only user's choices of variables
            {				
			   
	       for (int iu=0; iu < usersVarListSize; iu++) // list of user's panels
	       {
	          for (int ie=0; ie < pListMultiNumber[iu]; ie++) // loop over user's plots per panel
		  {
		     if (iCount == pListMultiPlots[iu][ie])
		     {			 
                        plotPanels[numPlotPanels].yAxisScaleType = 
		           (userScaleTypeList[numPlotPanels].isEmpty()) ? yAxisScaleType : userScaleTypeList[numPlotPanels].toInt();							   				      					 
                        if (currentPlotType == TIMESERIES_PLOT && !ignoreThisVar)
                        {						
                           int iCompMax = iVarTotalSize;
		           if (pListMultiPlotsComp[iu][plotPanels[iu].numTimeSeries] != -1) iCompMax = 1;
						
			   for (int io=0; io<iCompMax; io++)
			   {
                              plotPanels[iu].tsVarComp[plotPanels[iu].numTimeSeries] =
							     (iCompMax == 1 && pListMultiPlotsComp[iu][plotPanels[iu].numTimeSeries] != -1) ? pListMultiPlotsComp[iu][plotPanels[iu].numTimeSeries] : io;				
                              plotPanels[iu].tsFile[plotPanels[iu].numTimeSeries] = j;
                              plotPanels[iu].tsVar[plotPanels[iu].numTimeSeries] = i;
                              plotPanels[iu].yAxisMin = (userYminA[iu] == MY_FILL) ? scaleMin : userYminA[iu];
                              plotPanels[iu].yAxisMax = (userYmaxA[iu] == MY_FILL) ? scaleMax : userYmaxA[iu];
                              plotPanels[iu].yLabel = 
                                 (userListLabelsY[iu].isEmpty()) ? yLabelDefault : userListLabelsY[iu];
			      plotPanels[iu].tsType[plotPanels[iu].numTimeSeries] = currentDataType;
		              plotPanels[iu].tsFillValue[plotPanels[iu].numTimeSeries] = fillValue;
			      plotPanels[iu].legLabel[plotPanels[iu].numTimeSeries] = field;
                              plotPanels[iu].multFactor[plotPanels[iu].numTimeSeries] = 1.0;
                              plotPanels[iu].tsDATASET_ID[plotPanels[iu].numTimeSeries] = dataset_id;
			      plotPanels[iu].numTimeSeries++;                             
			   }
                        }
                        else if (currentPlotType == SPECTROGRAM_PLOT)
                        {		
/*
                           plotPanels[iu].spFile[plotPanels[iu].numSpectrograms] = j;
                           plotPanels[iu].spVar[plotPanels[iu].numSpectrograms] = i;     
                           plotPanels[numPlotPanels].multFactorSp[plotPanels[iu].numSpectrograms] = 1.0;
                           plotPanels[iu].yAxisMin = (userYminA[numPlotPanels] == MY_FILL) ? 1 : userYminA[numPlotPanels];
                           plotPanels[iu].yAxisMax = (userYmaxA[numPlotPanels] == MY_FILL) ? 800 : userYmaxA[numPlotPanels];			   
                           plotPanels[iu].cBarMin = (userCminA[numPlotPanels] == MY_FILL) ? scaleMin : userCminA[numPlotPanels];
                           plotPanels[iu].cBarMax = (userCmaxA[numPlotPanels] == MY_FILL) ? scaleMax : userCmaxA[numPlotPanels];    
			   plotPanels[iu].yLabel = 
                              (userListLabelsY[numPlotPanels].isEmpty()) ? yLabelDefault : userListLabelsY[numPlotPanels];
                           plotPanels[iu].cLabel =
                              (userListLabelsC[numPlotPanels].isEmpty()) ? cLabelDefault : userListLabelsC[numPlotPanels];  
		           plotPanels[iu].tsType[plotPanels[iu].numTimeSeries] = currentDataType;
		           plotPanels[iu].spFillValue[plotPanels[iu].numSpectrograms] = fillValue;
                           plotPanels[numPlotPanels].spMaster = -2;
                           plotPanels[iu].spDATASET_ID[plotPanels[iu].numSpectrograms] = dataset_id;	
                           plotPanels[iu].scN[plotPanels[iu].numSpectrograms] = clScN;
                           plotPanels[iu].numSpectrograms++;
*/

                  int iPmax = (pListPanels[numPlotPanels] > 0) ? pListPanels[numPlotPanels] : 1;
                  int iPmaster = (pListPanels[numPlotPanels] > 0) ? numPlotPanels : -1;

                  for (int ja=0; ja < iPmax; ja++)
                  {
                     plotPanels[iu].numTimeSeries = 0;
                     plotPanels[iu].spFile[plotPanels[iu].numSpectrograms] = j;
                     plotPanels[iu].spVar[plotPanels[iu].numSpectrograms] = i;
                     plotPanels[iu].multFactorSp[plotPanels[iu].numSpectrograms] = 1.0;
                     plotPanels[iu].yAxisMin = (userYminA[numPlotPanels] == MY_FILL) ? 1 : userYminA[numPlotPanels];
                     plotPanels[iu].yAxisMax = (userYmaxA[numPlotPanels] == MY_FILL) ? 800 : userYmaxA[numPlotPanels];
                     plotPanels[iu].cBarMin = (userCminA[numPlotPanels] == MY_FILL) ? scaleMin : userCminA[numPlotPanels];
                     plotPanels[iu].cBarMax = (userCmaxA[numPlotPanels] == MY_FILL) ? scaleMax : userCmaxA[numPlotPanels];
                     plotPanels[iu].tsType[0] = currentDataType;
                     plotPanels[iu].yLabel =
                        (userListLabelsY[numPlotPanels].isEmpty()) ? yLabelDefault : userListLabelsY[numPlotPanels];
                     plotPanels[numPlotPanels].cLabel =
                        (userListLabelsC[numPlotPanels].isEmpty()) ? cLabelDefault : userListLabelsC[numPlotPanels];
                     plotPanels[iu].spFillValue[plotPanels[iu].numSpectrograms] = fillValue;
                     plotPanels[iu].spDATASET_ID[plotPanels[iu].numSpectrograms] = dataset_id;
                     plotPanels[iu].scN[plotPanels[iu].numSpectrograms] = clScN;

                     // handle multiple polar plots

                     plotPanels[numPlotPanels].spPolarComp = pListPolarOnly[iPmaster][ja];
                     if (ja == 0 && iPmaster != -1)
                     {
                        plotPanels[numPlotPanels].spMaster = iPmax;
                        plotPanels[numPlotPanels].spSlave = 0;
                     }
                     else if (iPmaster == -1)
                     {
                        plotPanels[numPlotPanels].spMaster = -2;
                        plotPanels[numPlotPanels].spSlave = 0;
                     }
                     else
                     {
                        plotPanels[numPlotPanels].spMaster = -1;
                        plotPanels[numPlotPanels].spSlave = 1;
                     }
                     numPlotPanels++;
                     } // <-- loop for polars

                     plotPanels[iu].numSpectrograms++;

    	           } 

	             }
	          }
	       }
            }
	    iCount++;
         } // <-- end if PARAMETER_TYPE is DATA or SUPPORT_DATA (with some conditions)                
      } // <-- end loop over variables
   } // <-- end loop over files

   if (userMode == 1) exit(0);
   
   if (usersVarListSize > 0) numPlotPanels = usersVarListSize;

//----------------------------------------------------------------------------------
// set various things for instruments which don't have much useful variable metadata
//----------------------------------------------------------------------------------

// this enables sensible plots to be made without users having to specify things
// like axis limits, y-axis scale types, etc

   for (int i=0; i<numPlotPanels; i++)
   {
      for (int j=0; j<plotPanels[i].numSpectrograms; j++)
      {
         if (plotPanels[i].spDATASET_ID[j].contains("CP_CIS-HIA_HS_MAG_IONS_CS") ||
             plotPanels[i].spDATASET_ID[j].contains("CP_CIS-HIA_HS_SW_IONS_CS"))
         {
            plotPanels[i].yAxisScaleType = 1;
            if (plotPanels[i].cBarMin == 0 && plotPanels[i].cBarMax == 0)
            {
               plotPanels[i].cBarMin = 100;
               plotPanels[i].cBarMax = 100000;
            }
         }
         else if (plotPanels[i].spDATASET_ID[j].contains("CP_CIS-HIA_HS_MAG_IONS_PF"))
         {
            plotPanels[i].yAxisScaleType = 1;
            if (plotPanels[i].cBarMin == 0 && plotPanels[i].cBarMax == 0)
            {
               plotPanels[i].cBarMin = 1.0e+02;
               plotPanels[i].cBarMax = 1.0e+08;
            }
         }
         else if (plotPanels[i].spDATASET_ID[j].contains("CP_CIS-HIA_HS_MAG_IONS_PEF") ||
                  plotPanels[i].spDATASET_ID[j].contains("CP_CIS-CODIF_HS_O1_PEF") ||
                  plotPanels[i].spDATASET_ID[j].contains("CP_CIS-CODIF_HS_H1_PEF") ||
                  plotPanels[i].spDATASET_ID[j].contains("CP_CIS-CODIF_HS_He2_PEF"))
         {
            plotPanels[i].yAxisScaleType = 1;
            if (plotPanels[i].cBarMin == 0 && plotPanels[i].cBarMax == 0)
            {  
               plotPanels[i].cBarMin = 1.0e+05;
               plotPanels[i].cBarMax = 1.0e+07;
            }
         }
         else if (plotPanels[i].spDATASET_ID[j].contains("CP_WHI_NATURAL") ||
                  plotPanels[i].spDATASET_ID[j].contains("CP_WHI_ACTIVE"))
         {
            plotPanels[i].yAxisScaleType = 0;
            if (plotPanels[i].cBarMin == 0 && plotPanels[i].cBarMax == 0)
            {
               plotPanels[i].cBarMin = 1.0e-17;
               plotPanels[i].cBarMax = 1.0e-12;
            }
         }
      }
   }

//--------------------------------------------
// set up multiplication factors if they exist
//--------------------------------------------

   for (int i=0; i<numPlotPanels; i++)
   {
      for (int j=0; j<plotPanels[i].numSpectrograms; j++)
      {
         if (userMultFactorsListSp[i][j] != 0.0)
            plotPanels[i].multFactor[j] = userMultFactorsListTs[i][j];
      }
      for (int j=0; j<plotPanels[i].numSpectrograms; j++)
      {
         if (userMultFactorsListSp[i][j] != 0.0)
            plotPanels[i].multFactorSp[j] = userMultFactorsListSp[i][j];
      }
   }

//--------------------------------------------------
// do we have PEACE & EFW together in a single panel
//-------------------------------------------------- 

// we need to change the sign of EFW potential if plotting it with a PEACE spectrogram

   for (int i=0; i<numPlotPanels; i++)
   {
      int haveEFW = -1, havePEA = -1;
      for (int j=0; j<plotPanels[i].numTimeSeries; j++)
      {
 	 if (plotPanels[i].tsDATASET_ID[j].contains("CP_EFW_L3_P")) haveEFW = j;
      }
      for (int j=0; j<plotPanels[i].numSpectrograms; j++)
      {
 	 if (plotPanels[i].spDATASET_ID[j].contains("PEA")) havePEA = j;
      }
      if (haveEFW > -1 && havePEA > -1) plotPanels[i].multFactor[haveEFW] = -1.;
   }

//---------------------
// spacecraft potential
//---------------------

   if (scpShow)
   {

// we need to show the spacecraft potential on any PEACE spectrograms; note that the
// EFW potential with the appropriate spacecraft number will be used

      for (int i=0; i<numPlotPanels; i++)
      {
         //for (int j=0; j<plotPanels[i].numSpectrograms; j++)
         if (plotPanels[i].numSpectrograms > 0)
         {
            int j =0;
 	    if (plotPanels[i].spDATASET_ID[j].contains("PEA") && scpAssign[plotPanels[i].scN[j]-1] == 1)
	    {
	       plotPanels[i].tsFile[plotPanels[i].numTimeSeries] = scpFile[plotPanels[i].scN[j]-1];
               plotPanels[i].tsVar[plotPanels[i].numTimeSeries] = scpVar[plotPanels[i].scN[j]-1];
               plotPanels[i].tsType[plotPanels[i].numTimeSeries] = scpType[plotPanels[i].scN[j]-1];
               plotPanels[i].tsFillValue[plotPanels[i].numTimeSeries] = scpFill[plotPanels[i].scN[j]-1];
               plotPanels[i].multFactor[plotPanels[i].numTimeSeries] = -1.;
               plotPanels[i].numTimeSeries++;
	    }
         }      
      }
   }
   else if (scpCorr)
   {

// we need to spacecraft potential correct any PEACE spectrograms; note that the
// EFW potential with the appropriate spacecraft number will be used

      for (int i=0; i<numPlotPanels; i++)
      {
         for (int j=0; j<plotPanels[i].numSpectrograms; j++)
         {
            if (plotPanels[i].spDATASET_ID[j].contains("PEA") && scpAssignCorr[plotPanels[i].scN[j]-1] == 1)
            {
               plotPanels[i].spPotentialCorrect[j] = 1;
            }
            else plotPanels[i].spPotentialCorrect[j] = 0;
         }       
      }
   }

//--------------------------
// show PAD status variables
//--------------------------

   if (showStatus && PitchAngleQuality[0] != -1)
   {
      plotPanels[numPlotPanels].tsFile[0] = PitchAngleQuality[0];
      plotPanels[numPlotPanels].tsVar[0] = PitchAngleQuality[1];
      plotPanels[numPlotPanels].yLabel = "PAD Status Quality";
      plotPanels[numPlotPanels].tsType[0] = PitchAngleQuality[2];
      plotPanels[numPlotPanels].numTimeSeries = 1;
      plotPanels[numPlotPanels].numSpectrograms = 0;
      plotPanels[numPlotPanels].tsFillValue[0] = PitchAngleQualityFill;
      plotPanels[numPlotPanels].multFactor[0] = 1.0;
      plotPanels[numPlotPanels].tsVarComp[0] = 0;
      userYminA[numPlotPanels] = -1;
      userYmaxA[numPlotPanels] = 4;
      pListSymbolPlots[numPlotPanels][0] = PLOT_PLUS;
      usersColList[numPlotPanels][0] = "red";
      plotPanels[numPlotPanels].yAxisScaleType = 0;
      numPlotPanels++;
      usersColListSize = 1;
   }

//----------------------
// begin setting up plot 
//---------------------- 

// plot height
  
    if (numPlotPanels == 1 && !userModifiedPlotHeight)
    {
       plotHeight = 400; // for one panel, the plot page doesn't need to be as tall
    }  

    if (makeWheel)
    {
       plotWidth = 800;
       plotHeight = 700;
    }
    else if (makeSauvaud)
    {
       plotWidth = 800;
       plotHeight = 700;
    }
    else if (makeSlice)
    {
       plotWidth = 800;
       plotHeight = 700;
    }

// plot title

    QString myTitle, iniBit;
    
    if (autoPlotOnlyOne)
    {
       myTitle = gbObservatory + " > ";
       if (!gbExperiment.isEmpty()) myTitle = myTitle + gbExperiment + " > ";
    }
    if (userPlotTitle.isEmpty() && !autoPlotN & !autoPlotV && !autoPlotNV && !autoPlotT &&
        !autoPlotNVT && !autoPlotNT && !autoPlotPRS)
    {
       if (!gbObservatory.isEmpty()) myTitle = gbObservatory + " > ";
       if (!gbExperiment.isEmpty()) myTitle = myTitle + gbExperiment + " > ";
       myTitle = myTitle + dataSetTitle;
    }
    else if (!userPlotTitle.isEmpty()) myTitle = userPlotTitle;
    else if (autoPlotPRS)
    {
       myTitle = myTitle + "PEACE & RAPID omnidirectional flux";
    }
    else if (autoPlotN)
    {      
       myTitle = myTitle + "Density";
    }
    else if (autoPlotV && autoPlotVarg != "para" && autoPlotVarg != "perp")
    {
       myTitle = myTitle + "Velocity";
    }
    else if (autoPlotV && autoPlotVarg == "para")
    {
       myTitle = myTitle + "Velocity (para)";
    }
    else if (autoPlotV && autoPlotVarg == "perp")
    {
       myTitle = myTitle + "Velocity (perp)";
    }
    else if (autoPlotNV && autoPlotVarg != "para" && autoPlotVarg != "perp")
    {
       myTitle = myTitle + "Density & Velocity";
    }
    else if (autoPlotNV && autoPlotVarg == "para")
    {
       myTitle = myTitle + "Density & Velocity (para)";
    }
    else if (autoPlotNV && autoPlotVarg == "perp")
    {
       myTitle = myTitle + "Density & Velocity (perp)";
    }
    else if (autoPlotNVT && autoPlotVarg != "para" && autoPlotVarg != "perp")
    {
       myTitle = myTitle + "Density, Velocity & Temperature";
    }
    else if (autoPlotNVT && autoPlotVarg == "para")
    {
       myTitle = myTitle + "Density, Velocity & Temperature (para)";
    }
    else if (autoPlotNVT && autoPlotVarg == "perp")
    {
       myTitle = myTitle + "Density, Velocity & Temperature (perp)";
    }
    else if (autoPlotT && autoPlotVarg != "para" && autoPlotVarg != "perp")
    {
       myTitle = myTitle + "Temperature";
    }
    else if (autoPlotT && autoPlotVarg == "para")
    {
       myTitle = myTitle + "Temperature (para)";
    }
    else if (autoPlotT && autoPlotVarg == "perp")
    {
       myTitle = myTitle + "Temperature (perp)";
    }
    else if (autoPlotNT && autoPlotVarg == "para")
    {
       myTitle = myTitle + "Density & Temperature (para)";
    }
    else if (autoPlotNT && autoPlotVarg == "perp")
    {
       myTitle = myTitle + "Density & Temperature (perp)";
    }
    else if (autoPlotNT && autoPlotVarg != "para" && autoPlotVarg != "perp")
    {
       myTitle = myTitle + "Density & Temperature";
    }

#ifdef Q_WS_X11

#else
    guiOnOff = true; // for Mac & Windows, always enable GUI
#endif

// setup the QApplication

    QApplication app(argc, argv, guiOnOff);
    QFont defaultFont = QApplication::font();
    int defaultFontSize = defaultFont.pointSize();
    if (noDisplay) defaultFontSize = 12;

    int numColsIndex = 0;

// setup the QsiPlot

    QsiPlot myPlot(plotWidth, plotHeight, myTitle, defaultFontSize, guiOnOff);
    myPlot.setPlotLineWidth(1);
    myPlot.setCurveLineWidth(1); 
    myPlot.setBackgroundColor(Qt::white);
    myPlot.setGrid(gridOnOff, gridOnOff);
    myPlot.setColourBarDefinition(colourBarType);

// setup the QWheelPlot
// - we'll set the number of bins, y-axis & c-axis limits later

    QWheelPlot myWheelPlot(plotWidth, plotHeight, -1, -1, -1, -1, -1, -1, defaultFontSize, guiOnOff);
    myWheelPlot.setColourBarDefinition(colourBarType);

    if (makeWheel)
    {

    /*******************************************************************************
    **
    ** make wheel plots _ADL_WHEEL_
    **
    *******************************************************************************/

// get requested time from user

    int userTimeDefined = -1;
    QDateTime dateTimeUser;
    QTime timeOnlyUser;
    if (!userWheelTime.isEmpty())
    {
       QRegExp DateAndTime("^\\d\\d\\d\\d-\\d\\d-\\d\\dT\\d\\d:\\d\\d:\\d\\d.\\d\\d\\dZ$");
       QRegExp TimeOnlyMilli("^\\d\\d:\\d\\d:\\d\\d.\\d\\d\\d$");
       QRegExp TimeOnly("^\\d\\d:\\d\\d:\\d\\d$");

       if (DateAndTime.indexIn(userWheelTime) > -1)
       {
          dateTimeUser = QDateTime::fromString(userWheelTime, "yyyy-MM-ddTHH:mm:ss.zzzZ");
          dateTimeUser = dateTimeUser.toUTC();
          userTimeDefined = 0;
       }
       else if (TimeOnlyMilli.indexIn(userWheelTime) > -1)
       {
          timeOnlyUser = QTime::fromString(userWheelTime, "HH:mm:ss.zzz");
          userTimeDefined = 1;
       }
       else if (TimeOnly.indexIn(userWheelTime) > -1)
       {
          timeOnlyUser = QTime::fromString(userWheelTime, "HH:mm:ss");
          userTimeDefined = 1;
       }
    }

// get requested type

   bool inclBackLevel = false, inclReflection = false;
   if (!userWheelType.isEmpty())
   {
      if (userWheelType == "bl") inclBackLevel = true;
      if (userWheelType == "ga") inclReflection = true;
   }

// declarations

    std::vector<WheelData> wheelData;
    QString wheelLabel;
   
// decide what to plot

   for (int j = 0; j < numFiles; j++) // <-- loop over all CDF/CEF files specified by user
   {      
      QString OBSERVATORY = caaRead[j].getGlobalAttr("OBSERVATORY");
      QString EXPERIMENT = caaRead[j].getGlobalAttr("EXPERIMENT");
      for (int i=0; i<caaRead[j].getNumVariables(); i++) // <-- loop over variables in each CDF/CEF file
      {                 
         QString NAME = caaRead[j].getVariableAttr(i, "NAME");
         QString PROPERTY = caaRead[j].getVariableAttr(i, "PROPERTY");
         QString PARAMETER_TYPE = caaRead[j].getVariableAttr(i, "PARAMETER_TYPE");
         QString FIELDNAM = caaRead[j].getVariableAttr(i, "FIELDNAM");

         if (PARAMETER_TYPE.startsWith("DATA", Qt::CaseInsensitive) &&
             (NAME.contains("PITCH_FULL") ))
	 {
	    if (NAME.startsWith("Data_HEEA"))
	    {
                WheelData wdata;
                wdata.file = j;
                wdata.var = i;
                wdata.az_req = true;
                wdata.name = NAME;
                wdata.bl = false;
                wheelData.push_back(wdata);
            }

	    if (NAME.startsWith("Data_LEEA"))
	    {
                WheelData wdata;
                wdata.file = j;
                wdata.var = i;
                wdata.az_req = true;
                wdata.name = NAME;
                wdata.bl = false;
                wheelData.push_back(wdata);
            }
	 }
         else if (PARAMETER_TYPE.startsWith("DATA", Qt::CaseInsensitive) && !NAME.startsWith("BackgroundLevel__") &&
		  (NAME.contains("PITCH_SPIN") || NAME.contains("PITCH_3D")))
	 {
                WheelData wdata;
                wdata.file = j;
                wdata.var = i;
                wdata.name = NAME;
                wdata.bl = false;
                if (NAME.contains("PITCH_SPIN")) wdata.az_req = false;
                else if (NAME.contains("PITCH_3D")) wdata.az_req = true;
                wheelData.push_back(wdata);
	 }
         else if (NAME.startsWith("BackgroundLevel__") && inclBackLevel)
         {
             WheelData wdata;
             wdata.file = j;
             wdata.var = i;
             wdata.name = NAME;
             wdata.bl = true;
             if (NAME.contains("PITCH_SPIN")) wdata.az_req = false;
             else if (NAME.contains("PITCH_3D")) wdata.az_req = true;
             wheelData.push_back(wdata);
         }
         
         if (PARAMETER_TYPE.startsWith("DATA", Qt::CaseInsensitive) &&
             (NAME.startsWith("Data__") && NAME.contains("PITCH_SPIN") ||
	      (NAME.startsWith("Data_HEEA") && (NAME.contains("PITCH_FULL") ))))
         {
            // colour bar label

            QString units = caaRead[j].getVariableAttr(i, "UNITS");   
            QString lablaxis = caaRead[j].getVariableAttr(i, "LABLAXIS");
            wheelLabel = lablaxis + " (" + units + ")";
         }
      }
    }

// switch on/off

   for (int i=0; i<wheelData.size(); ++i)
   {
      wheelData.at(i).enabled = true;

      // include reflection

      if (inclReflection) wheelData.at(i).reflection = true;
      else wheelData.at(i).reflection = false;
   }

   QString subtitle;

// array to store what files we've read

   std::vector<int> filesRead;

// read in data; make plot when we get to the right record

   int icc = 0;
   for (int i=0; i<wheelData.size(); ++i)
   {

      filesRead.push_back(wheelData.at(i).file); // record file read

      if (i > 0)
      {
         caaRead[wheelData.at(i).file].rewindData();
      }

      if (!wheelData.at(i).enabled) continue;

      if (wheelData.at(i).name.contains("LEEA"))
      {
          if (userEnergyRegion.size() == 0 ||
              userEnergyRegion == "l" || userEnergyRegion == "lh")
          {}
          else continue;
      }

      if (wheelData.at(i).name.contains("HEEA"))
      {
          if (userEnergyRegion.size() == 0 ||
              userEnergyRegion == "h" || userEnergyRegion == "lh")
          {}
          else continue;
      }

      bool endNow = false;
      int iC = 0;
      while (caaRead[wheelData.at(i).file].readRecord() != _READ_FAILED && !endNow)
      {

  	 // get timing info

 	 QString time_stamp, time_stampBegin, time_stampEnd,
	         time_stampBegin2, time_stampEnd2;
         caaRead[wheelData.at(i).file].getDataEpoch(0, &time_stamp);
         float timeDeltaLo, timeDeltaHi;

         caaRead[wheelData.at(i).file].getDataValueDelta(0, _MINUS, &timeDeltaLo);
         caaRead[wheelData.at(i).file].getDataValueDelta(0, _PLUS, &timeDeltaHi);
         time_stampBegin = timeBounds(time_stamp, -timeDeltaLo);
         time_stampEnd = timeBounds(time_stamp, timeDeltaHi);
         QDateTime dateTimeIn = QDateTime::fromString(time_stamp, "yyyy-MM-ddTHH:mm:ss.zzzZ");
         QTime timeOnlyIn = QTime::fromString(time_stamp.mid(11, 12), "HH:mm:ss.zzz");
         dateTimeIn = dateTimeIn.toUTC();
       
         if (wheelData.at(i).az_req)  // determine the begin & end times of the two snapshots (PITCH_FULL & PITCH_3D only)
	 {
            int SPOS;
            caaRead[wheelData.at(i).file].getDataValue(4, &SPOS);
            double SUNAZ = 26.2 - ((double)SPOS)*360.0/(16.0*1024.0);

            float dataAZ[2] = {0}, dataAZp[2] = {0}, dataAZm[2] = {0};
            int iVarAzi = getVarAzi(caaRead[wheelData.at(i).file], wheelData.at(i).var);
            caaRead[wheelData.at(i).file].getDataValue(iVarAzi, &dataAZ[0]);
            caaRead[wheelData.at(i).file].getDataValueDelta(iVarAzi, _PLUS, &dataAZp[0]);
	    caaRead[wheelData.at(i).file].getDataValueDelta(iVarAzi, _MINUS, &dataAZm[0]);      

            double angleStart[2] = {0}, angleEnd[2] = {0};
            for (int iy=0; iy<2; iy++)
	    {
	       angleStart[iy] = dataAZ[iy] - dataAZm[iy] + SUNAZ;
	       angleEnd[iy]   = dataAZ[iy] + dataAZp[iy] + SUNAZ;

               if (angleStart[iy] < 0.) angleStart[iy] += 360.;
               if (angleEnd[iy] < 0.) angleEnd[iy] += 360.;
               if (angleStart[iy] > 360.) angleStart[iy] -= 360.;
               if (angleEnd[iy] > 360.) angleEnd[iy] -= 360.;

	       angleStart[iy] = angleStart[iy]*(timeDeltaLo + timeDeltaHi)/360.0 - timeDeltaLo;
	       angleEnd[iy]   = angleEnd[iy]*(timeDeltaLo + timeDeltaHi)/360.0 - timeDeltaLo;
	    }   

            time_stampBegin = timeBounds(time_stamp, angleStart[0]);
            time_stampEnd   = timeBounds(time_stamp, angleEnd[0]);
            time_stampBegin2= timeBounds(time_stamp, angleStart[1]);
            time_stampEnd2  = timeBounds(time_stamp, angleEnd[1]);
         }

         if ((dateTimeIn >= dateTimeUser && userTimeDefined == 0) || userTimeDefined == -1 ||
             (timeOnlyIn >= timeOnlyUser && userTimeDefined == 1))
	 {

         if ((icc == 0 && wheelUseHalfSpin == -1) || (wheelUseHalfSpin > -1 && icc == wheelUseHalfSpin - 1))
         {
             double wCmin, wCmax;
             wCmin = caaRead[wheelData.at(i).file].getVariableAttr(wheelData.at(i).var, "SCALEMIN").toFloat();
             wCmax = caaRead[wheelData.at(i).file].getVariableAttr(wheelData.at(i).var, "SCALEMAX").toFloat();

             myWheelPlot.yAxisMin = yMinAll;
             myWheelPlot.yAxisMax = yMaxAll;
             myWheelPlot.cAxisMin = (cMinAll != MY_FILL) ? cMinAll : wCmin;
             myWheelPlot.cAxisMax = (cMaxAll != MY_FILL) ? cMaxAll : wCmax;
         }

         Wheel wheel;
         wheel.getData(wheelData.at(i).var, wheelData.at(i).az_req, caaRead[wheelData.at(i).file]);
         wheel.includeReflection(wheelData.at(i).reflection);

         bool autoSetEn = true;
         if (icc > 0) autoSetEn = false;

         if ((wheelUseHalfSpin > -1 && icc == wheelUseHalfSpin - 1) || wheelUseHalfSpin == -1)
         {
            if (wheelData.at(i).bl)
            {
                if (inclBackLevel) wheel.plot(myWheelPlot, true, yMinAll, yMaxAll, autoSetEn);
            }
            else wheel.plot(myWheelPlot, false, yMinAll, yMaxAll, autoSetEn);
         }

         if (!wheelData.at(i).az_req) subtitle = time_stampBegin + " - " + time_stampEnd;
         else
         {
            if (wheelUseHalfSpin == -1)
            {
               subtitle = time_stampBegin + " - " + time_stampEnd + " | " + time_stampBegin2 + " - " + time_stampEnd2;
            }
            else
            {
               if (wheelUseHalfSpin == 1)
               {
                  subtitle = time_stampBegin + " - " + time_stampEnd;
               }
               else if (wheelUseHalfSpin == 2)
               {
                  subtitle = time_stampBegin2 + " - " + time_stampEnd2;
               }
            }
         }

         endNow = true;

         ++icc;

         }
         iC++;

      }
   }

         myWheelPlot.drawFrame("Energy (eV)", "Energy (eV)", wheelLabel, myTitle, subtitle);
               
         if (inclBackLevel) 
         {     
            myWheelPlot.drawSideTitle(RHS, "Background level");
         }


   } // <-- end make wheel plot
   else if (makeSauvaud)
   {

     cout << "[plotcf] making Sauvaud plot..." << endl;

    /*******************************************************************************
    **
    ** make Sauvaud plot
    **
    *******************************************************************************/

// declarations

    int numSauvaud = 0;
    int wFile[MAX_SIZE] = {0};      // which file contains the data
    int wVarNum[MAX_SIZE][4] = {0};    // variable containing data
    int wVarPolar[MAX_SIZE][4] = {0};  // number of polar bins
    int wVarEnergy[MAX_SIZE][4] = {0}; // number of energy bins
    int wVarAzimuth[MAX_SIZE][4] = {0}; // number of azimuth bins

    double sScaleMin[MAX_SIZE] = {0};
    double sScaleMax[MAX_SIZE] = {0};    
   
    int sizeE[4] = {0}, sizeP[4] = {0}; // size of energy & polar arrays
    int sizeA[4] = {0}; // size of azimuth array
    QString DEPEND_P[MAX_SIZE][4];     // name of variable containing polar angles
    QString DEPEND_E[MAX_SIZE][4];     // name of variable containing energy values
    QString DEPEND_A[MAX_SIZE][4];     // name of variable containing azimuth values
    QString sauvClabels[MAX_SIZE];  // units of each plot

    QString sauvBeginDateTime, sauvUsingBeginDateTime;
    QString sauvEndDateTime, sauvUsingEndDateTime;
    
    bool fourRegions = false;
    int modeRealSize[4][3] = {0};
    
    int vnPAD = -1;

    int ieu = 0, iieeuu = 0;
    if (userEnergyRegion == "l") ieu = 0;
    else if (userEnergyRegion == "h") ieu = 1;
    else if (userEnergyRegion == "lh") ieu = 2;

    iieeuu = ieu;
    if (ieu == 4) iieeuu = 1;

   for (int j = 0; j < numFiles; j++) // <-- loop over all CDF/CEF files specified by user
   {     
   	
       QString DSID = caaRead[j].getGlobalAttr("DATASET_ID");
       if (DSID.contains("PITCH_FULL")) fourRegions = true;   	
   	 
      QString OBSERVATORY = caaRead[j].getGlobalAttr("OBSERVATORY");
      QString EXPERIMENT = caaRead[j].getGlobalAttr("EXPERIMENT");
      for (int i=0; i<caaRead[j].getNumVariables(); i++) // <-- loop over variables in each CDF/CEF file
      {                 
         QString NAME = caaRead[j].getVariableAttr(i, "NAME");
         QString PROPERTY = caaRead[j].getVariableAttr(i, "PROPERTY");
         QString PARAMETER_TYPE = caaRead[j].getVariableAttr(i, "PARAMETER_TYPE");
         QString FIELDNAM = caaRead[j].getVariableAttr(i, "FIELDNAM");
         QString LABLAXIS = caaRead[j].getVariableAttr(i, "LABLAXIS");
         QString UNITS = caaRead[j].getVariableAttr(i, "UNITS");        
         
              if ((NAME.contains("Mode_RealSize_HEEA") || NAME.contains("Mode_RealSize__"))  && !NAME.contains("LABEL"))
             {
                int izx[3];
                caaRead[j].getDataValue(i, izx);
             	modeRealSize[0][0] = izx[0];
             	modeRealSize[0][1] = izx[1];
             	modeRealSize[0][2] = izx[2];
             }
             else if (NAME.contains("Mode_RealSize_LEEA") && !NAME.contains("LABEL"))
             {
                int izx[3];
                caaRead[j].getDataValue(i, izx);
             	modeRealSize[1][0] = izx[0];
             	modeRealSize[1][1] = izx[1];
             	modeRealSize[1][2] = izx[2];          	
             }        
         
         if (NAME.startsWith("Status_PAD") && !NAME.contains("LABEL") && vnPAD == -1) vnPAD = i;
         
         if (PARAMETER_TYPE.startsWith("DATA", Qt::CaseInsensitive) &&
             NAME.startsWith("Data_") && (DSID.contains("PITCH_SPIN") || DSID.contains("PITCH_3D")) ||
             DSID.contains("PITCH_FULL"))
         {
            wFile[numSauvaud] = j;                   

            if (NAME.startsWith("Data_HEEA"))
            {
               wVarNum[numSauvaud][0] = i;
               if (caaRead[j].getVariableAttr(i, "DEPEND_1").contains("Energy"))
                  DEPEND_E[numSauvaud][0] = caaRead[j].getVariableAttr(i, "DEPEND_1");
               if (caaRead[j].getVariableAttr(i, "DEPEND_2").contains("Energy"))
                  DEPEND_E[numSauvaud][0] = caaRead[j].getVariableAttr(i, "DEPEND_2");
               if (caaRead[j].getVariableAttr(i, "DEPEND_3").contains("Energy"))
                  DEPEND_E[numSauvaud][0] = caaRead[j].getVariableAttr(i, "DEPEND_3");

               if (caaRead[j].getVariableAttr(i, "DEPEND_1").contains("PitchAngle"))
                  DEPEND_P[numSauvaud][0] = caaRead[j].getVariableAttr(i, "DEPEND_1");
               if (caaRead[j].getVariableAttr(i, "DEPEND_2").contains("PitchAngle"))
                  DEPEND_P[numSauvaud][0] = caaRead[j].getVariableAttr(i, "DEPEND_2");
               if (caaRead[j].getVariableAttr(i, "DEPEND_3").contains("PitchAngle"))
                  DEPEND_P[Sauvaud][0] = caaRead[j].getVariableAttr(i, "DEPEND_3");

               if (caaRead[j].getVariableAttr(i, "DEPEND_1").contains("Azimuth"))
                  DEPEND_A[numSauvaud][0] = caaRead[j].getVariableAttr(i, "DEPEND_1");
               if (caaRead[j].getVariableAttr(i, "DEPEND_2").contains("Azimuth"))
                  DEPEND_A[numSauvaud][0] = caaRead[j].getVariableAttr(i, "DEPEND_2");
               if (caaRead[j].getVariableAttr(i, "DEPEND_3").contains("Azimuth"))
                  DEPEND_A[numSauvaud][0] = caaRead[j].getVariableAttr(i, "DEPEND_3");
             }
            if (NAME.startsWith("Data_HEEA"))
            {
            	wVarNum[numSauvaud][1] = i;
               if (caaRead[j].getVariableAttr(i, "DEPEND_1").contains("Energy"))
                  DEPEND_E[numSauvaud][1] = caaRead[j].getVariableAttr(i, "DEPEND_1");
               if (caaRead[j].getVariableAttr(i, "DEPEND_2").contains("Energy"))
                  DEPEND_E[numSauvaud][1] = caaRead[j].getVariableAttr(i, "DEPEND_2");
               if (caaRead[j].getVariableAttr(i, "DEPEND_3").contains("Energy"))
                  DEPEND_E[numSauvaud][1] = caaRead[j].getVariableAttr(i, "DEPEND_3");
               
               if (caaRead[j].getVariableAttr(i, "DEPEND_1").contains("PitchAngle"))
                  DEPEND_P[numSauvaud][1] = caaRead[j].getVariableAttr(i, "DEPEND_1");
               if (caaRead[j].getVariableAttr(i, "DEPEND_2").contains("PitchAngle"))
                  DEPEND_P[numSauvaud][1] = caaRead[j].getVariableAttr(i, "DEPEND_2");
               if (caaRead[j].getVariableAttr(i, "DEPEND_3").contains("PitchAngle"))
                  DEPEND_P[numSauvaud][1] = caaRead[j].getVariableAttr(i, "DEPEND_3");
               
               if (caaRead[j].getVariableAttr(i, "DEPEND_1").contains("Azimuth"))
                  DEPEND_A[numSauvaud][1] = caaRead[j].getVariableAttr(i, "DEPEND_1");
               if (caaRead[j].getVariableAttr(i, "DEPEND_2").contains("Azimuth"))
                  DEPEND_A[numSauvaud][1] = caaRead[j].getVariableAttr(i, "DEPEND_2");
               if (caaRead[j].getVariableAttr(i, "DEPEND_3").contains("Azimuth"))
                  DEPEND_A[numSauvaud][1] = caaRead[j].getVariableAttr(i, "DEPEND_3");    
             } 
             if (NAME.contains("PITCH_SPIN") || NAME.contains("PITCH_3D"))
             {
             	 wVarNum[numSauvaud][0] = i;
               if (caaRead[j].getVariableAttr(i, "DEPEND_1").contains("Energy"))
                  DEPEND_E[numSauvaud][0] = caaRead[j].getVariableAttr(i, "DEPEND_1");
               if (caaRead[j].getVariableAttr(i, "DEPEND_2").contains("Energy"))
                  DEPEND_E[numSauvaud][0] = caaRead[j].getVariableAttr(i, "DEPEND_2");
            
               if (caaRead[j].getVariableAttr(i, "DEPEND_1").contains("PitchAngle"))
                  DEPEND_P[numSauvaud][0] = caaRead[j].getVariableAttr(i, "DEPEND_1");
               if (caaRead[j].getVariableAttr(i, "DEPEND_2").contains("PitchAngle"))
                  DEPEND_P[numSauvaud][0] = caaRead[j].getVariableAttr(i, "DEPEND_2");
             }            
             
             
             if ((ieu == 0 && (NAME.contains("PITCH_SPIN") || NAME.contains("PITCH_3D") || NAME.contains("Data_Top"))) ||
                 ((ieu == 1 || ieu == 4) && NAME.contains("Data_Overlap_HEEA")) ||
                 (ieu == 2 && NAME.contains("Data_Overlap_LEEA")) ||
                 (ieu == 3 && NAME.contains("Data_Bottom")))
                 {
               
             
                       // colour bar label

            sauvClabels[numSauvaud] = LABLAXIS + " (" + UNITS + ")";

	    // get SCALEMIN & SCALEMAX

            sScaleMin[numSauvaud] = caaRead[j].getVariableAttr(i, "SCALEMIN").toFloat();
            sScaleMax[numSauvaud] = caaRead[j].getVariableAttr(i, "SCALEMAX").toFloat();           
                 }
             
         }
         
      }
     

	    // begin & end date/time

            sauvBeginDateTime = caaRead[j].getFileBeginDateTime();
            sauvEndDateTime = caaRead[j].getFileEndDateTime();
            numSauvaud++;      
      
    }
  
   if (numSauvaud > 1) numSauvaud = 1;

// set begin/end date & time

   if (userDefinedStartTime == -1) userDefinedStartTime = time_to_tis((char *)(sauvBeginDateTime.mid(11, 12).toStdString().c_str()));
   if (userDefinedStopTime == -1) userDefinedStopTime = time_to_tis((char *)(sauvEndDateTime.mid(11, 12).toStdString().c_str())); 

// use HH:MM only for time durations > 60 minutes

   double totalDuration = userDefinedStopTime - userDefinedStartTime;
   if (totalDuration > 3600.0)
   {
      myPlot.setTimeStringFormat("hh:mm");
   }

// find variables corresponding to polar, energy & azimuthal info

   for (int i=0; i < numSauvaud; i++)
   {
      for (int j=0; j < caaRead[wFile[i]].getNumVariables(); j++)
      {
         QString NAME = caaRead[wFile[i]].getVariableAttr(j, "NAME");
         for (int vi = 0; vi < (fourRegions ? 4 : 1); vi++)
         {
            if (NAME == DEPEND_E[i][vi])
            {
               wVarEnergy[i][vi] = j;
            }
            if (NAME == DEPEND_P[i][vi])
            {
               wVarPolar[i][vi] = j;
            }
            if (NAME == DEPEND_A[i][vi])
            {
               wVarAzimuth[i][vi] = j;
            }
         }  
      }     
   }

// plot setup

   int inum_panels = 1;
   if (showStatus) inum_panels++;

   myPlot.setLayout(inum_panels, 1, myTitle);
   if (showStatus)
   {
      myPlot.setSize(0, 0, 0.90);
      myPlot.setSize(1, 0, 0.10);
      myPlot.showValuesX(0, 0, pFalse);
      myPlot.showValuesX(1, 0, pTrue);  
   }
     
   myPlot.drawFrame(0, 0, true);
   myPlot.setScaleTypeY(0, 0, Log);
   
   if (showStatus)
   { 
      myPlot.drawFrame(1, 0, true);
      myPlot.setScaleTypeY(1, 0, Linear);      
   }

   double tMin, tMax, durMajX;
   int nMinX;
   myPlot.getAutoScaleX(0, -1, userDefinedStartTime, userDefinedStopTime, &durMajX, &nMinX, &tMin, &tMax); 

// set the initial date

    int beginYr = sauvBeginDateTime.mid(0, 4).toInt();
    int beginMn = sauvBeginDateTime.mid(5, 2).toInt();
    int beginDy = sauvBeginDateTime.mid(8, 2).toInt();

    QDate fileInitialDate(beginYr, beginMn, beginDy);
    myPlot.initialDate = fileInitialDate;
    
    double *statusVarStoreX, *statusVarStoreY;
    
    if (showStatus)
    {
       statusVarStoreX = new double[86400];
       statusVarStoreY = new double[86400];	
    }

// read in the data

   bool doOnlyOnce = false;
   int vnPADi = 0;
   for (int i=0; i< numSauvaud; i++)
   {

      if (cMinAll != MY_FILL) sScaleMin[i] = cMinAll;
      if (cMaxAll != MY_FILL) sScaleMax[i] = cMaxAll;

      while (caaRead[wFile[i]].readRecord() != _READ_FAILED)
      {
  	 // get timing info

	 QString time_stamp, time_stampBegin, time_stampEnd;
         caaRead[wFile[i]].getDataEpoch(0, &time_stamp);
         float timeDeltaLo, timeDeltaHi;
         caaRead[wFile[i]].getDataValueDelta(0, _MINUS, &timeDeltaLo);
         caaRead[wFile[i]].getDataValueDelta(0, _PLUS, &timeDeltaHi);
         float tisSpinCentre = time_to_tis((char *)(time_stamp.mid(11,12).toStdString()).c_str());
         double xBegin = tisSpinCentre - timeDeltaLo;
         double xEnd = tisSpinCentre + timeDeltaHi;
         
         float statusVar[3] = {0};
         if (showStatus)
         {
            caaRead[wFile[i]].getDataValue(vnPAD, statusVar);
            statusVarStoreY[vnPADi] = statusVar[0];
            statusVarStoreX[vnPADi] = tisSpinCentre;
            vnPADi++;
         }

         // allocate memory

         float binsElo[4][60], binsEhi[4][60], binsEcen[4][60], binsPlo[4][12], binsPhi[4][12];
         int eref[4][60] = {0};
         float binsAlo[4][12], binsAhi[4][12];

         for (int vi=0; vi < (fourRegions ? 4 : 1); vi++)
         {
            sizeE[vi] = caaRead[wFile[i]].getVariableSize(wVarEnergy[i][vi], 0);
            sizeP[vi] = caaRead[wFile[i]].getVariableSize(wVarPolar[i][vi], 0);
            if (fourRegions) sizeA[vi] = caaRead[wFile[i]].getVariableSize(wVarAzimuth[i][vi], 0);
         }	      

         float *data;
         float *dataBL = new float[sizeP[0]*sizeE[0]];

         float *dataP = new float[sizeP[0]];
         float *dataPp = new float[sizeP[0]];
         float *dataPm = new float[sizeP[0]];

         float *dataE = new float[sizeE[0]];
         float *dataEp = new float[sizeE[0]];
         float *dataEm = new float[sizeE[0]];
         
         float *dataA = new float[sizeA[0]];
         float *dataAp = new float[sizeA[0]];
         float *dataAm = new float[sizeA[0]];
         
         float *data_OL, *data_OH, *data_B;
         
         float *dataP_OL, *dataPp_OL, *dataPm_OL;
         float *dataP_OH, *dataPp_OH, *dataPm_OH;
         float *dataP_B, *dataPp_B, *dataPm_B;
         
         float *dataE_OL, *dataEp_OL, *dataEm_OL;
         float *dataE_OH, *dataEp_OH, *dataEm_OH;
         float *dataE_B, *dataEp_B, *dataEm_B;
         
         float *dataA_OL, *dataAp_OL, *dataAm_OL;
         float *dataA_OH, *dataAp_OH, *dataAm_OH;
         float *dataA_B, *dataAp_B, *dataAm_B;
         
         if (fourRegions)
         {
            dataE_OL = new float[sizeE[2]];
            dataEp_OL = new float[sizeE[2]];
            dataEm_OL = new float[sizeE[2]];
            
            dataE_OH = new float[sizeE[1]];
            dataEp_OH = new float[sizeE[1]];
            dataEm_OH = new float[sizeE[1]];
            
            dataE_B = new float[sizeE[3]];
            dataEp_B = new float[sizeE[3]];
            dataEm_B = new float[sizeE[3]];
            
            dataP_OL = new float[sizeP[2]];
            dataPp_OL = new float[sizeP[2]];
            dataPm_OL = new float[sizeP[2]];
            
            dataP_OH = new float[sizeP[1]];
            dataPp_OH = new float[sizeP[1]];
            dataPm_OH = new float[sizeP[1]];

            dataP_B = new float[sizeP[3]];
            dataPp_B = new float[sizeP[3]];
            dataPm_B = new float[sizeP[3]];          
            
            dataA_OL = new float[sizeA[2]];
            dataAp_OL = new float[sizeA[2]];
            dataAm_OL = new float[sizeA[2]];
            
            dataA_OH = new float[sizeA[1]];
            dataAp_OH = new float[sizeA[1]];
            dataAm_OH = new float[sizeA[1]];

            dataA_B = new float[sizeA[3]];
            dataAp_B = new float[sizeA[3]];
            dataAm_B = new float[sizeA[3]];          
            
            data = new float[sizeP[0]*sizeE[0]*sizeA[0]];
            data_OH = new float[sizeP[1]*sizeE[1]*sizeA[1]];
            data_OL = new float[sizeP[2]*sizeE[2]*sizeA[2]];
            data_B = new float[sizeP[3]*sizeE[3]*sizeA[3]];   
             	
         }
         else
         {
            data = new float[sizeP[0]*sizeE[0]];	
         }

         // get the pitch angle bin data
        
         caaRead[wFile[i]].getDataValue(wVarPolar[i][0], &dataP[0]);
         caaRead[wFile[i]].getDataValueDelta(wVarPolar[i][0], _PLUS, &dataPp[0]);
	     caaRead[wFile[i]].getDataValueDelta(wVarPolar[i][0], _MINUS, &dataPm[0]);
	 
	     if (fourRegions)
	     {
            caaRead[wFile[i]].getDataValue(wVarPolar[i][2], &dataP_OL[0]);
            caaRead[wFile[i]].getDataValueDelta(wVarPolar[i][2], _PLUS, &dataPp_OL[0]);
	        caaRead[wFile[i]].getDataValueDelta(wVarPolar[i][2], _MINUS, &dataPm_OL[0]);
	     
            caaRead[wFile[i]].getDataValue(wVarPolar[i][1], &dataP_OH[0]);
            caaRead[wFile[i]].getDataValueDelta(wVarPolar[i][1], _PLUS, &dataPp_OH[0]);
	        caaRead[wFile[i]].getDataValueDelta(wVarPolar[i][1], _MINUS, &dataPm_OH[0]);
	     
            caaRead[wFile[i]].getDataValue(wVarPolar[i][3], &dataP_B[0]);
            caaRead[wFile[i]].getDataValueDelta(wVarPolar[i][3], _PLUS, &dataPp_B[0]);
	        caaRead[wFile[i]].getDataValueDelta(wVarPolar[i][3], _MINUS, &dataPm_B[0]);	     	     	     	
	     }

         for (int k=0; k<sizeP[0]; k++)
	     {
	        binsPlo[0][k] = dataP[k] - dataPm[k];
	        binsPhi[0][k] = dataP[k] + dataPp[k];
	     }
	 
	     if (fourRegions)
	     {
            for (int k=0; k<sizeP[1]; k++)
	        {
	           binsPlo[1][k] = dataP_OH[k] - dataPm_OH[k];
	           binsPhi[1][k] = dataP_OH[k] + dataPp_OH[k];
	        }	  
            for (int k=0; k<sizeP[2]; k++)
	        {
	           binsPlo[2][k] = dataP_OL[k] - dataPm_OL[k];
	           binsPhi[2][k] = dataP_OL[k] + dataPp_OL[k];
	        }	 
            for (int k=0; k<sizeP[3]; k++)
	        {
	           binsPlo[3][k] = dataP_B[k] - dataPm_B[k];
	           binsPhi[3][k] = dataP_B[k] + dataPp_B[k];
	        }	 	        	           	
	     }

         // get the azimuth bin data if necessary
         
	     if (fourRegions)
	     {        
            caaRead[wFile[i]].getDataValue(wVarAzimuth[i][0], &dataA[0]);
            caaRead[wFile[i]].getDataValueDelta(wVarAzimuth[i][0], _PLUS, &dataAp[0]);
	        caaRead[wFile[i]].getDataValueDelta(wVarAzimuth[i][0], _MINUS, &dataAm[0]);
	
            caaRead[wFile[i]].getDataValue(wVarAzimuth[i][2], &dataA_OL[0]);
            caaRead[wFile[i]].getDataValueDelta(wVarAzimuth[i][2], _PLUS, &dataAp_OL[0]);
	        caaRead[wFile[i]].getDataValueDelta(wVarAzimuth[i][2], _MINUS, &dataAm_OL[0]);
	     
            caaRead[wFile[i]].getDataValue(wVarAzimuth[i][1], &dataA_OH[0]);
            caaRead[wFile[i]].getDataValueDelta(wVarAzimuth[i][1], _PLUS, &dataAp_OH[0]);
	        caaRead[wFile[i]].getDataValueDelta(wVarAzimuth[i][1], _MINUS, &dataAm_OH[0]);
	     
            caaRead[wFile[i]].getDataValue(wVarAzimuth[i][3], &dataA_B[0]);
            caaRead[wFile[i]].getDataValueDelta(wVarAzimuth[i][3], _PLUS, &dataAp_B[0]);
	        caaRead[wFile[i]].getDataValueDelta(wVarAzimuth[i][3], _MINUS, &dataAm_B[0]);	     	     	     	
	        
            for (int k=0; k<sizeA[0]; k++)
	        {
	           binsAlo[0][k] = dataA[k] - dataAm[k];
	           binsAhi[0][k] = dataA[k] + dataAp[k];
	        }
	 
            for (int k=0; k<sizeA[1]; k++)
	        {
	           binsAlo[1][k] = dataA_OH[k] - dataAm_OH[k];
	           binsAhi[1][k] = dataA_OH[k] + dataAp_OH[k];
	        }	  
            for (int k=0; k<sizeA[2]; k++)
	        {
	           binsAlo[2][k] = dataA_OL[k] - dataAm_OL[k];
	           binsAhi[2][k] = dataA_OL[k] + dataAp_OL[k];
	        }	 
            for (int k=0; k<sizeA[3]; k++)
	        {
	           binsAlo[3][k] = dataA_B[k] - dataAm_B[k];
	           binsAhi[3][k] = dataA_B[k] + dataAp_B[k];
	        }	 	        	           	
	     }


         // get the energy bin data & determine lower & upper limits

         caaRead[wFile[i]].getDataValue(wVarEnergy[i][0], &dataE[0]);
         caaRead[wFile[i]].getDataValueDelta(wVarEnergy[i][0], _PLUS, &dataEp[0]);
	     caaRead[wFile[i]].getDataValueDelta(wVarEnergy[i][0], _MINUS, &dataEm[0]);
       
         int sizeEactual[4] = {0};
         for (int k=0; k<sizeE[0]; k++)
	     { 
	        if (dataE[k] > 0)
	        {
	           binsElo[0][sizeEactual[0]] = dataE[k] - dataEm[k];
	           binsEhi[0][sizeEactual[0]] = dataE[k] + dataEp[k];
               binsEcen[0][sizeEactual[0]] = dataE[k];
               eref[0][sizeEactual[0]] = k;
               sizeEactual[0]++;
	        }
	     }
	     
	     if (fourRegions)
	     {
            caaRead[wFile[i]].getDataValue(wVarEnergy[i][1], &dataE_OH[0]);
            caaRead[wFile[i]].getDataValueDelta(wVarEnergy[i][1], _PLUS, &dataEp_OH[0]);
	        caaRead[wFile[i]].getDataValueDelta(wVarEnergy[i][1], _MINUS, &dataEm_OH[0]);	 

            caaRead[wFile[i]].getDataValue(wVarEnergy[i][2], &dataE_OL[0]);
            caaRead[wFile[i]].getDataValueDelta(wVarEnergy[i][2], _PLUS, &dataEp_OL[0]);
	        caaRead[wFile[i]].getDataValueDelta(wVarEnergy[i][2], _MINUS, &dataEm_OL[0]);	
	        
            caaRead[wFile[i]].getDataValue(wVarEnergy[i][3], &dataE_B[0]);
            caaRead[wFile[i]].getDataValueDelta(wVarEnergy[i][3], _PLUS, &dataEp_B[0]);
	        caaRead[wFile[i]].getDataValueDelta(wVarEnergy[i][3], _MINUS, &dataEm_B[0]);
	        
            for (int k=0; k<sizeE[1]; k++)
	        { 
	           if (dataE_OH[k] > 0)
	           {
	              binsElo[1][sizeEactual[1]] = dataE_OH[k] - dataEm_OH[k];
	              binsEhi[1][sizeEactual[1]] = dataE_OH[k] + dataEp_OH[k];
                  binsEcen[1][sizeEactual[1]] = dataE_OH[k];
                  eref[1][sizeEactual[1]] = k;
                  sizeEactual[1]++;
	           }
	        }	    
	        
            for (int k=0; k<sizeE[2]; k++)
	        { 
	           if (dataE_OL[k] > 0)
	           {
	              binsElo[2][sizeEactual[2]] = dataE_OL[k] - dataEm_OL[k];
	              binsEhi[2][sizeEactual[2]] = dataE_OL[k] + dataEp_OL[k];
                  binsEcen[2][sizeEactual[2]] = dataE_OL[k];
                  eref[2][sizeEactual[2]] = k;
                  sizeEactual[2]++;
	           }
	        }		        

            for (int k=0; k<sizeE[3]; k++)
	        { 
	           if (dataE_B[k] > 0)
	           {
	              binsElo[3][sizeEactual[3]] = dataE_B[k] - dataEm_B[k];
	              binsEhi[3][sizeEactual[3]] = dataE_B[k] + dataEp_B[k];
                  binsEcen[3][sizeEactual[3]] = dataE_B[k];
                  eref[3][sizeEactual[3]] = k;
                  sizeEactual[3]++;
	           }
	        }		        
	           
	     } // <-- end if (fourRegions)
	     

         if (!doOnlyOnce)
	      {
            double realMin = 0, realMax = 0;
            realMin = binsElo[iieeuu][sizeEactual[iieeuu]-1];
            realMax = binsEhi[iieeuu][0];
            if (yMinAll > MY_FILL) realMin = yMinAll;
            if (yMaxAll > MY_FILL) realMax = yMaxAll;
            myPlot.setPlotAxisLimits(0, 0, tMin, tMax, realMin, realMax);
            if (showStatus) myPlot.setPlotAxisLimits(1, 0, tMin, tMax, -1, 4);
	        myPlot.setSauvaudEnergyTable(0, 0, sizeEactual[iieeuu], &binsElo[iieeuu][0], &binsEhi[iieeuu][0], &binsEcen[iieeuu][0]);
	       doOnlyOnce = true;
	 }

         // get the data

         if (ieu == 0) caaRead[wFile[i]].getDataValue(wVarNum[i][0], &data[0]);
         else if (ieu == 1) caaRead[wFile[i]].getDataValue(wVarNum[i][1], &data_OH[0]);
         else if (ieu == 2) caaRead[wFile[i]].getDataValue(wVarNum[i][2], &data_OL[0]);
         else if (ieu == 3) caaRead[wFile[i]].getDataValue(wVarNum[i][3], &data_B[0]);
         else if (ieu == 4)
         {
            caaRead[wFile[i]].getDataValue(wVarNum[i][1], &data_OH[0]);
            caaRead[wFile[i]].getDataValue(wVarNum[i][2], &data_OL[0]);
         }


         // draw the data
         
         for (int ia=0; ia < (fourRegions ? sizeA[iieeuu] : 1); ia++)
         {
            double xBeginThis, xEndThis;
         	if (fourRegions)
         	{
         	   xBeginThis = xBegin + (double)ia*(xEnd - xBegin)/((double)modeRealSize[iieeuu][0]);
         	   xEndThis = xBegin + (double)(ia+1)*(xEnd - xBegin)/((double)modeRealSize[iieeuu][0]);
         	}
         	else
         	{
         	   xBeginThis = xBegin;
         	   xEndThis = xEnd;	
         	}
            for (int ie=0; ie<sizeEactual[iieeuu]; ie++)
	        {
	           for (int ip=0; ip<sizeP[iieeuu]; ip++)
	           {
                          double val = 0;
	           	  if (ieu == 0) val = data[eref[iieeuu][ie] + ip*sizeE[iieeuu] + sizeP[iieeuu]*sizeE[iieeuu]*ia];
	              else if (ieu == 1 || ieu == 4) val = data_OH[eref[iieeuu][ie] + ip*sizeE[iieeuu] + sizeP[iieeuu]*sizeE[iieeuu]*ia];
	              else if (ieu == 2) val = data_OL[eref[iieeuu][ie] + ip*sizeE[iieeuu] + sizeP[iieeuu]*sizeE[iieeuu]*ia];
	              else if (ieu == 3) val = data_B[eref[iieeuu][ie] + ip*sizeE[iieeuu] + sizeP[iieeuu]*sizeE[iieeuu]*ia];
//cout << "draw [1] = " << val << " :: "<<  ia << " : " << ie << " : " << ip << " x b/e = " << xBegin << " : " << xEnd << endl;
	              myPlot.drawSauvaudSlice(0, 0, val, xBeginThis, xEndThis, binsPlo[iieeuu][ip], binsPhi[iieeuu][ip],
	                                      binsEcen[iieeuu][ie], sScaleMin[i], sScaleMax[i]);
	           }
	        }
         }

         if (ieu == 4) // _ADL_MARKER_
         {
         ieu = 2; iieeuu = 2;
         for (int ia=0; ia < (fourRegions ? sizeA[iieeuu] : 1); ia++)
         {   
            double xBeginThis, xEndThis;
                if (fourRegions)
                {
                   xBeginThis = xBegin + (double)ia*(xEnd - xBegin)/((double)modeRealSize[iieeuu][0]);
                   xEndThis = xBegin + (double)(ia+1)*(xEnd - xBegin)/((double)modeRealSize[iieeuu][0]);
                }
                else
                {
                   xBeginThis = xBegin;
                   xEndThis = xEnd;     
                }
            for (int ie=0; ie<sizeEactual[iieeuu]; ie++)
                {
                   for (int ip=0; ip<sizeP[iieeuu]; ip++)
                   {
                          double val;
                          if (ieu == 0) val = data[eref[iieeuu][ie] + ip*sizeE[iieeuu] + sizeP[iieeuu]*sizeE[iieeuu]*ia];
                      else if (ieu == 1 || ieu == 4) val = data_OH[eref[iieeuu][ie] + ip*sizeE[iieeuu] + sizeP[iieeuu]*sizeE[iieeuu]*ia];
                      else if (ieu == 2) val = data_OL[eref[iieeuu][ie] + ip*sizeE[iieeuu] + sizeP[iieeuu]*sizeE[iieeuu]*ia];
                      else if (ieu == 3) val = data_B[eref[iieeuu][ie] + ip*sizeE[iieeuu] + sizeP[iieeuu]*sizeE[iieeuu]*ia];
//cout << "draw [2] = " << val << " :: "<<  ia << " : " << ie << " : " << ip << " x b/e = " << xBegin << " : " << xEnd << endl;
                      myPlot.drawSauvaudSlice(0, 0, val, xBeginThis, xEndThis, binsPlo[iieeuu][ip], binsPhi[iieeuu][ip],
                                              binsEcen[iieeuu][ie], sScaleMin[i], sScaleMax[i]);
                   }
                }
         }  
         ieu = 4; iieeuu = 1;
         }

         // de-allocate memory

         delete [] data;
         delete [] dataBL;
         delete [] dataP;
         delete [] dataPp;
         delete [] dataPm;
         delete [] dataE;
         delete [] dataEp;
         delete [] dataEm;
         
         delete [] data_OL;
         delete [] data_OH;
         delete [] data_B;
         
         delete [] dataP_OL;
         delete [] dataPp_OL;
         delete [] dataPm_OL;
         
         delete [] dataP_OH;
         delete [] dataPp_OH;
         delete [] dataPm_OH;
         
         delete [] dataP_B;
         delete [] dataPp_B;
         delete [] dataPm_B;
         
         delete [] dataE_OL;
         delete [] dataEp_OL;
         delete [] dataEm_OL;
         
         delete [] dataE_OH;
         delete [] dataEp_OH;
         delete [] dataEm_OH;
         
         delete [] dataE_B;
         delete [] dataEp_B;
         delete [] dataEm_B;
 
         delete [] dataA;
         delete [] dataAp;
         delete [] dataAm;
         
         delete [] dataA_OL;
         delete [] dataAp_OL;
         delete [] dataAm_OL;
         
         delete [] dataA_OH;
         delete [] dataAp_OH;
         delete [] dataAm_OH;
         
         delete [] dataA_B;
         delete [] dataAp_B;
         delete [] dataAm_B;         

      }
   }

   // colour box, energy lines, 

   myPlot.drawColourBox(0, 0, sScaleMin[0], sScaleMax[0], sauvClabels[0], 0);
   myPlot.drawSauvaudEnergyLines(0, 0);
   myPlot.drawFrame(0, 0);
   myPlot.setPlotType(0, 0, TimeSeries);
   myPlot.setNumTicksX(0, 0, durMajX, nMinX);
   myPlot.setLabelY(0, 0, "Energy (eV)");
   
   if (showStatus)
   {
      myPlot.drawFrame(1, 0);
      myPlot.setNumTicksY(1, 0, 5, 0);
      myPlot.setPlotType(1, 0, TimeSeries);
      myPlot.setNumTicksX(1, 0, durMajX, nMinX);
      myPlot.setLabelY(1, 0, "PAD Quality");      
      myPlot.drawCurve(1, 0, "black", vnPADi, statusVarStoreX, statusVarStoreY, "", PLOT_PLUS);
   }

   } // <-- end make Sauvaud plot
   else if (makeSlice)
   {

     cout << "[plotcf] making slice plot..." << endl;

    /*******************************************************************************
    **
    ** make slice
    **
    *******************************************************************************/

// get requested time from user

    int userTimeDefined = -1;
    QDateTime dateTimeUser;
    QTime timeOnlyUser;
    if (!userWheelTime.isEmpty())
    {
       QRegExp DateAndTime("^\\d\\d\\d\\d-\\d\\d-\\d\\dT\\d\\d:\\d\\d:\\d\\d.\\d\\d\\dZ$");
       QRegExp TimeOnlyMilli("^\\d\\d:\\d\\d:\\d\\d.\\d\\d\\d$");
       QRegExp TimeOnly("^\\d\\d:\\d\\d:\\d\\d$");

       if (DateAndTime.indexIn(userWheelTime) > -1)
       {
          dateTimeUser = QDateTime::fromString(userWheelTime, "yyyy-MM-ddTHH:mm:ss.zzzZ");
          dateTimeUser = dateTimeUser.toUTC();
          userTimeDefined = 0;
       }
       else if (TimeOnlyMilli.indexIn(userWheelTime) > -1)
       {
          timeOnlyUser = QTime::fromString(userWheelTime, "HH:mm:ss.sss");
          userTimeDefined = 1;
       }
       else if (TimeOnly.indexIn(userWheelTime) > -1)
       {
          timeOnlyUser = QTime::fromString(userWheelTime, "HH:mm:ss");
          userTimeDefined = 1;
       }
    }

// get requested type

   bool inclBackLevel = false;
   if (!userSliceWhich.isEmpty())
   {
      if (userSliceWhich == "bl") inclBackLevel = true;
   }

   int paComp = -1;   
   QString paLegend;

   std::vector<int> paComps;
   QStringList paLegends;

   if (!userSliceType.isEmpty())
   {
      if (userSliceType == "para")
      {
         paComp = 0;
	 paLegend = " (0 deg PA)";
         paComps.push_back(0);
         paLegends << " (0 deg PA)";
      }
      else if (userSliceType == "perp")
      {
         paComp = 6;
	 paLegend = " (90 deg PA)";
         paComps.push_back(6);
         paLegends << " (90 deg PA)";
      }
      else if (userSliceType == "antipara")
      {
         paComp = 11;
	 paLegend = " (180 deg PA)";
         paComps.push_back(11);
         paLegends << " (180 deg PA)";
      }
      else if (userSliceType == "ppa")
      {
         paComps.push_back(0);
         paComps.push_back(6);
         paComps.push_back(11);
         paLegends << " (0 deg PA)";
         paLegends << " (90 deg PA)";
         paLegends << " (180 deg PA)";
      }
   }

// declarations

    int numSlices = 0;
    int wFile[MAX_SIZE] = {0};      // which file contains the data
    int wVarNum[MAX_SIZE] = {0};    // variable containing data
    int wVarPolar[MAX_SIZE] = {0};  // number of polar bins
    int wVarEnergy[MAX_SIZE] = {0}; // number of energy bins
    int wVarNumBL[MAX_SIZE] = {0};  // variable containing background level data
   
    int sizeE = 0, sizeP = 0;       // size of energy & polar arrays
    QString DEPEND_P[MAX_SIZE];     // name of variable containing polar angles
    QString DEPEND_E[MAX_SIZE];     // name of variable containing energy values
    QString wheelClabels[MAX_SIZE]; // units of each wheel plot

// decide what to plot

   for (int j = 0; j < numFiles; j++) // <-- loop over all CDF/CEF files specified by user
   {      
      QString OBSERVATORY = caaRead[j].getGlobalAttr("OBSERVATORY");
      QString EXPERIMENT = caaRead[j].getGlobalAttr("EXPERIMENT");
      for (int i=0; i<caaRead[j].getNumVariables(); i++) // <-- loop over variables in each CDF/CEF file
      {                 
         QString NAME = caaRead[j].getVariableAttr(i, "NAME");
         QString PROPERTY = caaRead[j].getVariableAttr(i, "PROPERTY");
         QString PARAMETER_TYPE = caaRead[j].getVariableAttr(i, "PARAMETER_TYPE");
         QString FIELDNAM = caaRead[j].getVariableAttr(i, "FIELDNAM");
         
         if (PARAMETER_TYPE.startsWith("DATA", Qt::CaseInsensitive) &&
             NAME.startsWith("Data__") && NAME.contains("PITCH_SPIN"))
         {
            wFile[numSlices] = j;
            wVarNum[numSlices] = i;
            wVarNumBL[numSlices] = i+1;

	    // get polar bin info

            DEPEND_P[numSlices] = caaRead[j].getVariableAttr(i, "DEPEND_1");

            // get energy bin info

            DEPEND_E[numSlices] = caaRead[j].getVariableAttr(i, "DEPEND_2");   

            // colour bar label

            QString units = caaRead[j].getVariableAttr(i, "UNITS");   
            QString lablaxis = caaRead[j].getVariableAttr(i, "LABLAXIS");
            wheelClabels[numSlices] = lablaxis + " (" + units + ")";

            numSlices++;
         }
      }
    }

// find variables corresponding to polar & energy info

   for (int i=0; i < numSlices; i++)
   {
      for (int j=0; j < caaRead[wFile[i]].getNumVariables(); j++)
      {
	 QString NAME = caaRead[wFile[i]].getVariableAttr(j, "NAME");
         if (NAME == DEPEND_P[i])
	 {
	    wVarPolar[i] = j;
	 }
	 if (NAME == DEPEND_E[i])
	 {
	    wVarEnergy[i] = j;          
	 }
      }
   }

// read in data; make plot when we get to the right record

   for (int i=0; i< numSlices; i++)
   {
      bool endNow = false;
      int iC = 0;
      while (caaRead[wFile[i]].readRecord() != _READ_FAILED && !endNow)
      {

  	 // get timing info

	 QString time_stamp, time_stampBegin, time_stampEnd;
         caaRead[wFile[i]].getDataEpoch(0, &time_stamp);
         float timeDeltaLo, timeDeltaHi;
         caaRead[wFile[i]].getDataValueDelta(0, _MINUS, &timeDeltaLo);
         caaRead[wFile[i]].getDataValueDelta(0, _PLUS, &timeDeltaHi);
         time_stampBegin = timeBounds(time_stamp, -timeDeltaLo);
         time_stampEnd = timeBounds(time_stamp, timeDeltaHi);
         QDateTime dateTimeIn = QDateTime::fromString(time_stamp, "yyyy-MM-ddTHH:mm:ss.zzzZ");
         QTime timeOnlyIn = QTime::fromString(time_stamp.mid(11, 12), "HH:mm:ss.zzz");
         dateTimeIn = dateTimeIn.toUTC();

         if ((dateTimeIn >= dateTimeUser && userTimeDefined == 0) || userTimeDefined == -1 ||
             (timeOnlyIn >= timeOnlyUser && userTimeDefined == 1))
	 {

         // allocate memory

         float binsElo[60], binsEhi[60], binsPlo[12], binsPhi[12];
         double wYmin, wYmax, wCmin, wCmax;

         sizeP = caaRead[wFile[i]].getVariableSize(wVarPolar[i], 0);
         sizeE = caaRead[wFile[i]].getVariableSize(wVarEnergy[i], 0);

         float *data = new float[sizeP*sizeE];
         float *dataBL = new float[sizeP*sizeE];

         float *dataP = new float[sizeP];
         float *dataPp = new float[sizeP];
         float *dataPm = new float[sizeP];

         float *dataE = new float[sizeE];
         float *dataEp = new float[sizeE];
         float *dataEm = new float[sizeE];

         // get the pitch angle bin data

         caaRead[wFile[i]].getDataValue(wVarPolar[i], &dataP[0]);
         caaRead[wFile[i]].getDataValueDelta(wVarPolar[i], _PLUS, &dataPp[0]);
	 caaRead[wFile[i]].getDataValueDelta(wVarPolar[i], _MINUS, &dataPm[0]);

         for (int k=0; k<sizeP; k++)
	 {
	    binsPlo[k] = dataP[k] - dataPm[k] + 90.0;
	    binsPhi[k] = dataP[k] + dataPp[k] + 90.0;
	 }

         // get the energy bin data & determine lower & upper limits

         caaRead[wFile[i]].getDataValue(wVarEnergy[i], &dataE[0]);
         caaRead[wFile[i]].getDataValueDelta(wVarEnergy[i], _PLUS, &dataEp[0]);
	 caaRead[wFile[i]].getDataValueDelta(wVarEnergy[i], _MINUS, &dataEm[0]);
         double energyFill = caaRead[wFile[i]].getVariableAttr(wVarEnergy[i], "FILLVAL").toDouble();

         wYmin = 1.0e+06;
         wYmax = -1.0e+06;
         for (int k=0; k<sizeE; k++)
	 {
	    binsElo[k] = dataE[k] - dataEm[k];
	    binsEhi[k] = dataE[k] + dataEp[k];
            if (binsElo[k] < wYmin && dataE[k] != energyFill) wYmin = binsElo[k];
            if (binsEhi[k] > wYmax) wYmax = binsEhi[k];
	 }

	 if (wYmin < 1.0) wYmin = 1.0;

         // get the data

         caaRead[wFile[i]].getDataValue(wVarNum[i], &data[0]);
         if (inclBackLevel) caaRead[wFile[i]].getDataValue(wVarNumBL[i], &dataBL[0]);

         // get the default colour bar scale from the header

         wCmin = caaRead[wFile[i]].getVariableAttr(wVarNum[i], "SCALEMIN").toFloat();
         wCmax = caaRead[wFile[i]].getVariableAttr(wVarNum[i], "SCALEMAX").toFloat();

	 // make the slice plot
	 
	 
         wYmin = 1.0e+06;
         wYmax = -1.0e+06;	 
	 double xx[60], yy[60], xx_bl[60], yy_bl[60];
	 
	 if (inclBackLevel)
	 {
	    for (int k=0; k<sizeE; k++)
	    {	
	       xx_bl[k] = dataE[k];
	       yy_bl[k] = dataBL[k + sizeE*paComp];
               if (dataE[k] < wYmin && dataE[k] != energyFill) wYmin = dataE[k];
               if (dataE[k] > wYmax) wYmax = dataE[k];		    
	    } 
	 }
	 
	 int sizeEi = 0;
	 for (int k=0; k<sizeE; k++)
	 {
	    if (data[k + sizeE*paComp] > 0)
	    {
	       yy[sizeEi] = data[k + sizeE*paComp];
	       xx[sizeEi] = dataE[k];
	       sizeEi++;	       
	    }	   
	 }
	 
	 QString subtitle = time_stampBegin + " - " + time_stampEnd;
	 
         myPlot.setLayout(1, 1, myTitle, subtitle);
         myPlot.setLegendLocation(TopLeft);
         myPlot.setPlotType(0, 0, Scalar);
	 myPlot.showValuesX(0, 0, pTrue);
	 myPlot.setScaleTypeX(0, 0, Log);
	 myPlot.setScaleTypeY(0, 0, Log);	
	 myPlot.setPlotAxisLimits(0, 0, 
	                          (yMinAll != MY_FILL) ? yMinAll : wYmin,
				  (yMaxAll != MY_FILL) ? yMaxAll : wYmax,
	                          (cMinAll != MY_FILL) ? cMinAll : wCmin, 
	                          (cMaxAll != MY_FILL) ? cMaxAll : wCmax);
	 myPlot.drawFrame(0, 0, false);
	 
	 myPlot.drawCurve(0, 0, QColor("black"), sizeEi, xx, yy, "data" + paLegend, PLOT_LINE);
	 if (inclBackLevel)
	 {
	    myPlot.showLegend(0, 0, 0, PLOT_LINE);
	    myPlot.drawCurve(0, 0, QColor("blue"), sizeE, xx_bl, yy_bl, "background level" + paLegend, PLOT_LINE);
	    myPlot.showLegend(0, 0, 1, PLOT_LINE);
	 }
	 
	 myPlot.drawFrame(0, 0);
	 myPlot.setNumTicksX(0, 0, 0, 0);
	 myPlot.setNumTicksY(0, 0, 0, 0);
	 
	 myPlot.setLabelX(0, 0, "Energy (eV)");
	 myPlot.setLabelY(0, 0, wheelClabels[0]);

         // de-allocate memory

         delete [] data;
         delete [] dataBL;
         delete [] dataP;
         delete [] dataPp;
         delete [] dataPm;
         delete [] dataE;
         delete [] dataEp;
         delete [] dataEm;

         endNow = true;

         }
         iC++;

      }
   }

   } // <-- end make slice plot   
   else
   {

    /*******************************************************************************
    **
    ** make time series plots & spectrograms
    **
    *******************************************************************************/

//-------------------------
// continue setting up plot 
//-------------------------

// set the initial date

    int beginYr = beginDateUsed.mid(0, 4).toInt();
    int beginMn = beginDateUsed.mid(5, 2).toInt();
    int beginDy = beginDateUsed.mid(8, 2).toInt();
    
    QDate fileInitialDate(beginYr, beginMn, beginDy);
    myPlot.initialDate = fileInitialDate;  

// set the number of panels and title

    int plotRows = numPlotPanels, plotCols = 1;
    if (showingAllPolarBins > 0)
    {
    
// if the user is making a multiple-panel spectrogram plot showing all polar or
// PA bins of a product, use a matrix plot format rather than a single column    
    
       numColsIndex = -1;
       if (showingAllPolarBins == 12)
       {
          plotRows = 6;
	  plotCols = 2;
       }
       else if (showingAllPolarBins == 6)
       {
          plotRows = 3;
	  plotCols = 2;
       }
    }

    myPlot.setLayout(plotRows, plotCols, myTitle);
    myPlot.setLegendLocation(TopLeft);

// set colours

   if (usersColListSize > 0)
   {
      for (int i=0; i<numPlotPanels; i++)
      {
         for (int j=0; j<plotPanels[i].numTimeSeries; j++)
	 {
	    if (!usersColList[i][j].isEmpty())
	    {
	       plotPanels[i].colour[j] = usersColList[i][j];
	    }
	 }
      }
   }

//----------------------------------
// set plot panel sizes if necessary
//----------------------------------

   if (userHeights)
   {

      // Users specify percentage reduction in specified plots, whereas QsiPlot requires a full
      // list of fractions for each row. We therefore need to convert what the user has specified
      // into what QsiPlot requires.

      std::vector<double> userHeightVals;
      userHeightVals.resize(plotRows);

      // Initialize
      for (int i=0; i<plotRows; ++i) userHeightVals.at(i) = 1.0;

      // Default plot height fraction of page
      double defaultHeightFraction = 1.0/((double)plotRows);

      for (int i=0; i<userHeightsList.size(); ++i)
      { 
         userHeightVals.at(userHeightsList.at(i).plot) = userHeightsList.at(i).height;
      }

      // Rescale
      for (int i=0; i<plotRows; ++i) userHeightVals.at(i) = userHeightVals.at(i)*defaultHeightFraction;

      // Ensure sum of all height fractions is 1, so the plots fill the page
      double ssum = 0.0;
      for (int i=0; i<plotRows; ++i) 
      {
         ssum += userHeightVals.at(i);
      }
      ssum = (1.0 - ssum)/((double)plotRows);
      for (int i=0; i<plotRows; ++i) userHeightVals.at(i) = userHeightVals.at(i) + ssum;

      // Set plot row heights

      for (int i=0; i<plotRows; ++i)
      {
         for (int j=0; j<plotCols; ++j)
         {
            myPlot.setSize(i, j, userHeightVals.at(i));
         }
      }
   }

//------------------------------------------------------------------
// determine the size of arrays needed to store the time series data
//------------------------------------------------------------------

// data for plotting

   int tsArraySize = 0;
   for (int i=0; i<numPlotPanels; i++)
   {
      plotPanels[i].tsFromDiff = 0;
      int giv = -1, gif = -1;
      for (int j=0; j<plotPanels[i].numTimeSeries; j++)
      {
         tsArraySize++;
	 if (j == 0)
	 {
	    giv = plotPanels[i].tsVar[j];
	    gif = plotPanels[i].tsFile[j];
	 }
	 else
	 {
	    if (plotPanels[i].tsVar[j] == giv && plotPanels[i].tsFile[j] == gif)
	    {
	    }
	    else
	    {
               plotPanels[i].tsFromDiff = 1;			   
	    }
	 }
      }
   }

// additional data for "-scp-corr" option
// - this is "hidden" data, i.e. time-series data that we require but don't want
//   to plot

   if (scpCorr)
   {
      for (int i=0; i<MAX_DISTINCT_SPACECRAFT; i++)
      {
         if (scpAssignCorr[i] == 1)
         {
            tsArraySize++;
         }
      }
   }

//------------------------------------------------
// determine DEPEND variables for each spectrogram
//------------------------------------------------

   int readFirst[MAX_FILES] = {0}; 

   for (int i=0; i<numPlotPanels; i++)
   {
      for (int j=0; j<plotPanels[i].numSpectrograms; j++)
      {

// initialize 
	  
         for (int k=0; k<4; k++)
	 {
	    plotPanels[i].spVars[j][k] = -1;
	    plotPanels[i].spDepends[j][k] = -1;
	 }

// read in all the DEPEND_i

         QString DEPEND[4];
         DEPEND[0] = caaRead[plotPanels[i].spFile[j]].getVariableAttr(plotPanels[i].spVar[j], "DEPEND_0");
         DEPEND[1] = caaRead[plotPanels[i].spFile[j]].getVariableAttr(plotPanels[i].spVar[j], "DEPEND_1");
         DEPEND[2] = caaRead[plotPanels[i].spFile[j]].getVariableAttr(plotPanels[i].spVar[j], "DEPEND_2");
         DEPEND[3] = caaRead[plotPanels[i].spFile[j]].getVariableAttr(plotPanels[i].spVar[j], "DEPEND_3");       

// for CEF files, we need to make sure we read in one record
		 
         if (readFirst[plotPanels[i].spFile[j]] == 0)
         {
            caaRead[plotPanels[i].spFile[j]].readRecord(false);
            readFirst[plotPanels[i].spFile[j]] = 1;
         }		   

// find the variable associated with each DEPEND_i

	 plotPanels[i].spVars[j][0] = -1;
	 plotPanels[i].spDepends[j][0] = -1;		 
	 plotPanels[i].spVars[j][1] = -1;
	 plotPanels[i].spDepends[j][1] = -1;		 
	 plotPanels[i].spVars[j][2] = -1;
	 plotPanels[i].spDepends[j][2] = -1;	

         for (int jj=0; jj<4; jj++)
	 {
	     
	    for (int k=0; k<numVariables[plotPanels[i].spFile[j]]; k++)
	    {
	       QString varName = caaRead[plotPanels[i].spFile[j]].getVariableAttr(k, "NAME");
               if (varName == DEPEND[jj])
	       {
	          QString qField = caaRead[plotPanels[i].spFile[j]].getVariableAttr(k, "FIELDNAM");
	          QString qUnits = caaRead[plotPanels[i].spFile[j]].getVariableAttr(k, "UNITS");
		  QString qLabel = caaRead[plotPanels[i].spFile[j]].getVariableAttr(k, "LABLAXIS");
				  
		  if (qField.contains("Energy", Qt::CaseInsensitive) || 
		      qLabel.contains("Energy", Qt::CaseInsensitive) || 
	              qUnits.contains("eV", Qt::CaseInsensitive) ||
                      qField.contains("Spectral Frequencies"))
		  {
		     plotPanels[i].spVars[j][2] = k;
		     plotPanels[i].spDepends[j][2] = jj-1;
		  }
		  else if ((qLabel.contains("Theta", Qt::CaseInsensitive) && !qField.contains("Azimuth", Qt::CaseInsensitive)) || 
			   qLabel.contains("Polar", Qt::CaseInsensitive) || 
			   qField.contains("Polar", Qt::CaseInsensitive) ||
                           qLabel.contains("Pitch", Qt::CaseInsensitive) ||
                           qField.contains("Pitch", Qt::CaseInsensitive))
		  {
		     plotPanels[i].spVars[j][1] = k;
		     plotPanels[i].spDepends[j][1] = jj-1;
		  }
		  else if (qLabel.contains("Azimuth", Qt::CaseInsensitive) || 
			   qField.contains("Azimuth", Qt::CaseInsensitive))
		  {
		     plotPanels[i].spVars[j][0] = k;
	             plotPanels[i].spDepends[j][0] = jj-1;
		  }				  
	       }
	    } // <-- end loop over all variables
	 } // <-- end loop over DEPEND_i

// read in the energy table to determine the lower & upper limits - we'll use these for the
// y-axis limits

         if (plotPanels[i].spVars[j][2] != -1)
		 {
            int sizeA = caaRead[plotPanels[i].spFile[j]].getVariableSize(plotPanels[i].spVars[j][2], 0);
             float fillValueV = caaRead[plotPanels[i].spFile[j]].getVariableAttr(plotPanels[i].spVars[j][2], "FILLVAL").toFloat();
			float *arc = new float[sizeA];
			float *arl = new float[sizeA];
			float *arh = new float[sizeA];
			caaRead[plotPanels[i].spFile[j]].getDataValue(plotPanels[i].spVars[j][2], arc);
			caaRead[plotPanels[i].spFile[j]].getDataValueDelta(plotPanels[i].spVars[j][2], _PLUS, arh);
			caaRead[plotPanels[i].spFile[j]].getDataValueDelta(plotPanels[i].spVars[j][2], _MINUS, arl);
                        double myMin = 1.0e+30, myMax = -1.0e+30;
                        int myMinIn = -1, myMaxIn = -1;
			for (int h=0; h<sizeA; h++)
			  {
			    if (arc[h] > myMax)
                            {
			      myMaxIn = h;
                              myMax = arc[h];
			    }
                            if (arc[h] < myMin && arc[h] != fillValueV)
			      {
				myMinIn = h;
                                myMin = arc[h];
                              }
			  }
			plotPanels[i].yAxisMin = (int)(arc[myMinIn] - arl[myMinIn]);
			plotPanels[i].yAxisMax = (ceil)(arc[myMaxIn] + arh[myMaxIn]);
			//                        if (plotPanels[i].yAxisMin == 0) plotPanels[i].yAxisMin = 1.0;
			delete [] arc; 
			delete [] arl;
			delete [] arh;		 
		 }	  
	  }
   }

//--------------------------------------------------
// allocate storage for scalar and time series plots
//--------------------------------------------------

   double **xx;
   xx = new double*[numFiles];
   for (register int i=0; i<numFiles; i++) xx[i] = new double [MAX_ARRAY_SIZE];
   
   double **aa;
   aa = new double*[tsArraySize];
   for (register int i=0; i<tsArraySize; i++) aa[i] = new double [MAX_ARRAY_SIZE];

// initialize lower/upper y-axis limits for each plot panel

   double vMinStore[MAX_SIZE], vMaxStore[MAX_SIZE];

   for (int i=0; i<numPlotPanels; i++)
   {
      vMinStore[i] = 3.0e+30;
      vMaxStore[i] = -3.0e+30;
   }  
   
//------------------------------   
// prepare list of required data
//------------------------------

// we generate a list of unique variables required from each file

   int tsWhat[MAX_FILES][MAX_SIZE] = {0};
   int tsWhatRef[MAX_FILES][MAX_SIZE][2] = {0};
   int tsWhatSize[MAX_FILES] = {0};
   int tsWhatType[MAX_FILES] = {0};

   for (int i=0; i<numPlotPanels; i++)
   {
      for (int j=0; j<plotPanels[i].numTimeSeries; j++)
      {
         if (tsWhatSize[plotPanels[i].tsFile[j]] == 0)
         {
            tsWhat[plotPanels[i].tsFile[j]][tsWhatSize[plotPanels[i].tsFile[j]]] = plotPanels[i].tsVar[j];
	    tsWhatRef[plotPanels[i].tsFile[j]][tsWhatSize[plotPanels[i].tsFile[j]]][0] = i;
            tsWhatRef[plotPanels[i].tsFile[j]][tsWhatSize[plotPanels[i].tsFile[j]]][1] = j;
	    tsWhatSize[plotPanels[i].tsFile[j]]++;
         }
         else
         {
            bool found = false;	
            for (int k=0; k<tsWhatSize[plotPanels[i].tsFile[j]]; k++)
            {
               if (tsWhat[plotPanels[i].tsFile[j]][k] == plotPanels[i].tsVar[j])
               {
                  found = true;				  
               }
            }
            if (!found)
            {
               tsWhat[plotPanels[i].tsFile[j]][tsWhatSize[plotPanels[i].tsFile[j]]] = plotPanels[i].tsVar[j];
               tsWhatRef[plotPanels[i].tsFile[j]][tsWhatSize[plotPanels[i].tsFile[j]]][0] = i;
               tsWhatRef[plotPanels[i].tsFile[j]][tsWhatSize[plotPanels[i].tsFile[j]]][1] = j;			   
               tsWhatSize[plotPanels[i].tsFile[j]]++;			   
            }
         }
      }
   }   

// now add hidden data required for PEACE spacecraft potential correction to spectrograms

   if (scpCorr)
   {
      for (int i=0; i<MAX_DISTINCT_SPACECRAFT; i++)
      {
         if (scpAssignCorr[i] == 1)
         {
            tsWhatSize[scpFileCorr[i]] = 1;
            tsWhat[scpFileCorr[i]][0] = scpVarCorr[i];
            tsWhatType[scpFileCorr[i]] = 1;
            tsWhatRef[scpFileCorr[i]][0][0] = i;
         }
      }
   }

/*
   cout << "num files = " << numFiles << endl;
   for (int i=0; i<numFiles; i++)
   {
      cout << "[" << i << "] " << tsWhatSize[i] << endl;
      for (int j=0; j<tsWhatSize[i]; j++)
      {
         cout << "REQDATA (time-series) = " << i << " : " << j << " :: " << tsWhat[i][j] << endl;
      }
   }
*/

//-----------------------------------   
// read in data for time-series plots
//-----------------------------------

   cout << "[plotcf] reading time-series data if necessary..." << endl;

   int ii[MAX_FILES] = {0};
   int jjF[MAX_FILES] = {0};

   QDate currentDate(2050, 1, 1), previousDate(2050, 1, 1); 

   for (int mF=0; mF<numFiles; mF++)
   {
      bool doneOnce = false;
      if (tsWhatSize[mF] > 0)
      {
         while (caaRead[mF].readRecord() != _READ_FAILED)
         {
            if (ii[mF] < MAX_ARRAY_SIZE)
            {
            QString timeData;
            float dataF[12], dataIn;
            int dataI[12];
            double dataD[12];
            short int dataI2[12];
	 
            caaRead[mF].getDataEpoch(0, &timeData);
            xx[mF][ii[mF]] = time_to_tis((char *)(timeData.mid(11,12).toStdString()).c_str());

            if (xx[mF][ii[mF]] > -1.)
            {
           	//cout << "reading " << timeData.toStdString() << endl;
         
            currentDate = toQDate(timeData.mid(0, 10));
            if (currentDate > fileInitialDate) xx[mF][ii[mF]] = xx[mF][ii[mF]] + 86400.0;        

            int jj = 0;
            if (mF > 0) for (int n=0; n< mF; n++) jj = jj + jjF[n];
			
            if (tsWhatType[mF] == 0) // visible variables, i.e. actually shown in plots
            {
               for (int i=0; i<tsWhatSize[mF]; i++) // loop over all required variables from the current file
               {		
                  for (int ia=0; ia<numPlotPanels; ia++)
                  {
                     for (int ib=0; ib<plotPanels[ia].numTimeSeries; ib++)
                     {
//cout << "IN0" << endl;
                        if (plotPanels[ia].tsFile[ib] == mF && plotPanels[ia].tsVar[ib] == tsWhat[mF][i])
                        {
//cout << "IN" << endl;
                           if (plotPanels[ia].tsType[ib] == INT) caaRead[mF].getDataValue(plotPanels[ia].tsVar[ib], dataI);
                           else if (plotPanels[ia].tsType[ib] == INT2) caaRead[mF].getDataValue(plotPanels[ia].tsVar[ib], dataI2);
                           else if (plotPanels[ia].tsType[ib] == DOUBLE) caaRead[mF].getDataValue(plotPanels[ia].tsVar[ib], dataD);
                           else caaRead[mF].getDataValue(plotPanels[ia].tsVar[ib], dataF);		

                           if (plotPanels[ia].tsType[ib] == INT) dataIn = dataI[plotPanels[ia].tsVarComp[ib]];
                           else if (plotPanels[ia].tsType[ib] == INT2) dataIn = dataI2[plotPanels[ia].tsVarComp[ib]];
                           else if (plotPanels[ia].tsType[ib] == DOUBLE) dataIn = dataD[plotPanels[ia].tsVarComp[ib]];
                           else dataIn = dataF[plotPanels[ia].tsVarComp[ib]];
                           //dataIn = dataIn*plotPanels[ia].multFactor[ib];
//cout <<"DATA = " << dataIn << " from " << ia << " mult = " << plotPanels[ia].multFactor[ib] << endl;
     
                           aa[jj][ii[mF]] = (fl_compare(dataIn, plotPanels[ia].tsFillValue[ib]) || fabs(dataIn) > MaxAllowedValue) ? FillValue : dataIn;
//cout << "data _ADL1_ = " << aa[jj][ii[mF]] << " type = " << plotPanels[ia].tsType[ib] << " fill = " << plotPanels[ia].tsFillValue[ib] << std::endl;;
                           aa[jj][ii[mF]] *= plotPanels[ia].multFactor[ib];
                           if (aa[jj][ii[mF]] < vMinStore[ia] && fabs(aa[jj][ii[mF]]) < MaxAllowedValue) vMinStore[ia] = aa[jj][ii[mF]];
			   if (aa[jj][ii[mF]] > vMaxStore[ia] && fabs(aa[jj][ii[mF]]) < MaxAllowedValue) vMaxStore[ia] = aa[jj][ii[mF]];
                           plotPanels[ia].tsDataRef[ib] = jj;					    
                           if (!doneOnce) jjF[mF] = jjF[mF] + 1;
                           jj++;
                        }
                     } // <-- loop over each time-series data in a plot panel
	          } // <-- loop over plot panels
               } // <-- loop over required variables from current file
            }
            else // now consider hidden variables, i.e. not shown in any plots
            {
               for (int i=0; i<tsWhatSize[mF]; i++)
               {
                  if (scpTypeCorr[tsWhatRef[mF][i][0]] == INT) caaRead[mF].getDataValue(scpVarCorr[tsWhatRef[mF][i][0]], dataI);
                  else if (scpTypeCorr[tsWhatRef[mF][i][0]] == INT2) caaRead[mF].getDataValue(scpVarCorr[tsWhatRef[mF][i][0]], dataI2);
                  else caaRead[mF].getDataValue(scpVarCorr[tsWhatRef[mF][i][0]], dataF);

                  if (scpTypeCorr[tsWhatRef[mF][i][0]] == INT) dataIn = dataI[0];
                  else if (scpTypeCorr[tsWhatRef[mF][i][0]] == INT2) dataIn = dataI2[0];
                  else dataIn = dataF[0];

                  aa[jj][ii[mF]] = (dataIn == scpFillCorr[tsWhatRef[mF][i][0]] || fabs(dataIn) > MaxAllowedValue) ? FillValue : dataIn;

                  scpDataRefCorr[tsWhatRef[mF][i][0]] = jj;

                  if (!doneOnce) jjF[mF] = jjF[mF] + 1;
                  jj++;
               }
            }

            ii[mF] = ii[mF] + 1;
            doneOnce = true;
            previousDate = currentDate;

            } // <-- if loop (tis > -1)

         } // <-- if to ensure we don't go over the max array size
         } // <-- while loop over date
      } // <-- if - check for spectrograms in file
   } // <-- loop over files


//------------   
// axis limits
//------------
   
   for (int i=0; i<numPlotPanels; i++)
   {   

// specify time series plots

      myPlot.setPlotType(i, numColsIndex, TimeSeries);

// only show time/date values on bottom plot(s)

      int rem = (i+1) % plotRows;
      if ((plotCols == 1 && i < numPlotPanels-1) ||
          (plotCols > 1 && rem != 0))
      {
         myPlot.showValuesX(i, numColsIndex, pFalse);
      }	 
      else
      {
         myPlot.showValuesX(i, numColsIndex, pTrue);
      }
   }

   int nMinX;
   double tMin, tMax, durMajX;

   int pListMinY[MAX_SIZE], pListMajY[MAX_SIZE];

   if (userDefinedStartTime == -1) userDefinedStartTime = beginDateTimeUsed;
   if (userDefinedStopTime == -1) userDefinedStopTime = endDateTimeUsed;

   // adjust format of time values on x-axis depending on duration of plot

   double totalDuration = userDefinedStopTime - userDefinedStartTime;
   if (totalDuration > 3600.0)
   {
      myPlot.setTimeStringFormat("hh:mm");
   }
   
   myPlot.getAutoScaleX(&xx[0][0], -1, userDefinedStartTime, userDefinedStopTime, &durMajX, &nMinX, &tMin, &tMax);   

   int nMaj, nMin, currentScaleType, iPsEnd = -1, iMaster = -1;
   double vMin, vMax, uMin, uMax;    
   
   for (int i=0; i<numPlotPanels; i++)
   {                
      if (plotPanels[i].spMaster > -1 || plotPanels[i].spMaster == -2)
      {	        	
         if (plotPanels[i].numSpectrograms > 0)
	 {
	    uMin = plotPanels[i].yAxisMin;
	    uMax = plotPanels[i].yAxisMax;
	  } 
	  else
	  {
             uMin = vMinStore[i];
      	     uMax = vMaxStore[i];	  
	  }

          if (userYminA[i] > MY_FILL) uMin = userYminA[i];
          if (userYmaxA[i] > MY_FILL) uMax = userYmaxA[i];

          if (yMinAll > MY_FILL) uMin = yMinAll;
          if (yMaxAll > MY_FILL) uMax = yMaxAll;
	  
	  if (plotPanels[i].yAxisScaleType == 1)
	  currentScaleType = Log;
	  else currentScaleType = Linear;             

	  myPlot.getAutoScaleY(uMin, uMax, &nMaj, &nMin, &vMin, &vMax, plotPanels[i].yAxisScaleType);	   

          // fix possible log plot problems
          if (currentScaleType == Log && vMin == 0) vMin = 1;
	     
	  for (int iii=i; iii<numPlotPanels; iii++)
	  {
	     if (plotPanels[iii].spSlave == 1)
	     {
	        iPsEnd = iii;
	     }
	  }     
      }      

      myPlot.setScaleTypeZ(i, numColsIndex, Log);
      myPlot.setScaleTypeY(i, numColsIndex, currentScaleType);
      myPlot.setScaleTypeX(i, numColsIndex, Linear);
      myPlot.setPlotAxisLimits(i, numColsIndex, tMin, tMax, vMin, vMax);   
      myPlot.drawFrame(i, numColsIndex, false);	
  
      pListMinY[i] = nMin;
      pListMajY[i] = nMaj;
   }
   
//--------------------
// draw colour box(es)
//--------------------
   
// for multiple polar bin spectrograms, use a single colour box   
   
   int endPrev[50] = {0};
   for (int i=0; i<numPlotPanels; i++)
   {

      // set the colour bar limits if user has specified the same limits for all 

      if (cMinAll > MY_FILL) plotPanels[i].cBarMin = cMinAll;
      if (cMaxAll > MY_FILL) plotPanels[i].cBarMax = cMaxAll;
     
      QString tmpQStrY;
      for (int j=0; j<plotPanels[i].numSpectrograms; j++)
      {
         QString yLabel = caaRead[plotPanels[i].spFile[j]].getVariableAttr(plotPanels[i].spVars[j][2], "LABLAXIS");
         QString unitsLabelY = caaRead[plotPanels[i].spFile[j]].getVariableAttr(plotPanels[i].spVars[j][2], "UNITS");
          
         if (yLabel == "NULL" && unitsLabelY == "eV") yLabel = "energy"; // cope with CIS
                    
         tmpQStrY = yLabel + " (" + unitsLabelY + ")";
         if (!userListLabelsY[i].isEmpty()) tmpQStrY = userListLabelsY[i];      
      }	
   	
      if (plotPanels[i].spMaster != -1 || plotPanels[i].spMaster == -2)
      {
      	 iMaster = i;
	 for (int iii=i; iii<numPlotPanels; iii++)
	 {
	    if (plotPanels[iii].spSlave == 1)
	    {
	       iPsEnd = iii;
	    }
	 }  	 
         if (plotPanels[i].yLabel.isEmpty())
         {
            plotPanels[i].yLabel = tmpQStrY;
         }
         myPlot.setLabelY(i, numColsIndex, plotPanels[i].yLabel);     
      }
      else
      {
         if (plotPanels[i].spSlave == 1)
         {
            myPlot.setLabelY(i, numColsIndex, plotPanels[iMaster].yLabel);
         }
         else 
         {  
            if (!plotPanels[i].yLabel.isEmpty()) myPlot.setLabelY(i, 0, tmpQStrY);
            else myPlot.setLabelY(i, numColsIndex, plotPanels[i].yLabel);
         }
      }

      if (plotPanels[i].numSpectrograms > 0 && plotPanels[i].spMaster != -1)
      { 

         // should we plot the box?

         bool plotTheBox = true;
         if (i > 0)
         {
            for (int ii=0; ii<i; ++ii)
            {
               if (iPsEnd == endPrev[ii] && iPsEnd != -1) plotTheBox = false;
            }
           
            if (specialModeFlag == 0) plotTheBox = false;
         }
         else
         {
            if (specialModeFlag == 0) iPsEnd = 1;
         }
         
	 if (plotTheBox) myPlot.drawColourBox(i, numColsIndex, plotPanels[i].cBarMin, plotPanels[i].cBarMax, plotPanels[i].cLabel, iPsEnd);
         endPrev[i] = iPsEnd;
      }      
      
   }

//-------------------------------------
// smooth time-series data if necessary
//-------------------------------------
   
   if (smoothTimeSeriesData)
   {
      cout << "[plotcf] smoothing data..." << endl;

// make sure the smoothing time is sensible

      if (smoothTimeSeriesSecs <=0.) smoothTimeSeriesSecs = 4.;
      else if (smoothTimeSeriesSecs > 40000.) smoothTimeSeriesSecs = 40000.;
  
// do the smoothing

      for (int i=0; i<numPlotPanels; i++)
      {
         for (int j=0; j<plotPanels[i].numTimeSeries; j++)
         {

// allocate memory

            float *tmparr = new float[ii[plotPanels[i].tsFile[j]]];

// loop over all data

            for (int k=0; k<ii[plotPanels[i].tsFile[j]]; k++)
            {
               int indexBeg, indexEnd;
            
// find data before the current timestamp within the box

               indexBeg = k;
               bool exitNow = false;
               while (!exitNow)
               {
                  if (indexBeg == 0 ||
                     xx[plotPanels[i].tsFile[j]][indexBeg] < xx[plotPanels[i].tsFile[j]][k] - smoothTimeSeriesSecs)
                  exitNow = true;
                  indexBeg--;
               }

// find data after the current timestamp within the box

               indexEnd = k;
               exitNow = false;
               while (!exitNow)
               {
                  if (indexEnd == ii[plotPanels[i].tsFile[j]]-1 ||
                      xx[plotPanels[i].tsFile[j]][indexEnd] > xx[plotPanels[i].tsFile[j]][k] + smoothTimeSeriesSecs)
                  exitNow = true;
                  indexEnd++;
               }

// average the data

               int avCount = 0;
               double avValue = 0;
               for (int l=indexBeg; l<indexEnd; l++)
               {
                  if (!fl_compare(aa[plotPanels[i].tsDataRef[j]][l], plotPanels[i].tsFillValue[j]))
                  {
                     avValue += aa[plotPanels[i].tsDataRef[j]][l];
                     avCount++;
                  }
               }
               avValue = (avCount > 0) ? avValue/((double)avCount) : avValue;
               tmparr[k] = avValue;
            }

// replace the old array with the smoothed data

            for (int k=0; k<ii[plotPanels[i].tsFile[j]]; k++)
            {
               aa[plotPanels[i].tsDataRef[j]][k] = tmparr[k];
            }

// de-allocate memory

            delete [] tmparr;
         }
      }
   }
      
//-----------------------------------
// draw horizontal lines if necessary
//-----------------------------------

   if (allLinesY != MY_FILL)
   {
      for (int i=0; i<numPlotPanels; i++)
      {
         myPlot.drawLineY(i, numColsIndex, QColor("black"), PLOT_LINE_DOTTED, allLinesY);
      }
   }
   else
   {
      for (int i=0; i<numPlotPanels; i++)
      {
         if (userLinesY[i].size() > 0)
         myPlot.drawLineY(i, numColsIndex, QColor("black"), PLOT_LINE_DOTTED, userLinesY[i].toFloat());
      }
   }

//------------   
// colour list
//------------

   QString colorList[10];
   colorList[0]="black";
   colorList[1]="red";
   colorList[2]="lime";
   colorList[3]="blue";
   colorList[4]="magenta";
   colorList[5]="cyan";
   colorList[6]="navy";
   colorList[7]="darkorange";
   colorList[8]="blueviolet";   

//-----------
// reset data
//-----------

// return to the beginning of the data sections of each file; just in case there is both time series
// and spectrogram data in a single file. We already have all necessary time series data.

   for (int j=0; j<numFiles; j++)
   {
      caaRead[j].rewindData();
   }
   
//----------   
// plot data
//----------

   cout << "[plotcf] making plot..." << endl;

   for (int im=0; im<numPlotPanels; im++)
   {
      for (int jm=0; jm<plotPanels[im].numSpectrograms; jm++)
      {    
      	  	 	 
         int iCounterPot = 0;
         while (caaRead[plotPanels[im].spFile[jm]].readRecord() != _READ_FAILED)
         {

	    // how many spectrograms from the same file do we need to plot? (there can be multiple spectrogrammable
            // data in a single file, e.g. PEACE PITCH_FULL, PEACE PITCH_3D)

	    int numSpectrogramsToPlot = 0;
            int whichSpectrogramFromFile[10][2] = {0};
            for (int iit=0; iit<numPlotPanels; iit++)
            {
               for (int jj=0; jj<plotPanels[iit].numSpectrograms; jj++)
	       {
		  if (plotPanels[im].spFile[jm] == plotPanels[iit].spFile[jj])
		  {
		     whichSpectrogramFromFile[numSpectrogramsToPlot][0] = iit;
		     whichSpectrogramFromFile[numSpectrogramsToPlot][1] = jj;
		     numSpectrogramsToPlot++;
		  }
	       }
	    }

            for (int ial=0; ial<numSpectrogramsToPlot; ial++)
	    {

	    int i = whichSpectrogramFromFile[ial][0];
            int j = whichSpectrogramFromFile[ial][1]; 

	    // initial setup

            int sSize[3] = {0}, varDim = 0;
	    varDim = caaRead[plotPanels[i].spFile[j]].getVariableDimension(plotPanels[i].spVar[j]);	 
	 
	    if (plotPanels[i].spVars[j][0] > -1)
	       sSize[0] = caaRead[plotPanels[i].spFile[j]].getVariableSize(plotPanels[i].spVar[j], plotPanels[i].spDepends[j][0]);
	    else sSize[0] = 1;
	 
	    if (plotPanels[i].spVars[j][1] > -1)
	       sSize[1] = caaRead[plotPanels[i].spFile[j]].getVariableSize(plotPanels[i].spVar[j], plotPanels[i].spDepends[j][1]); 
  	    else sSize[1] = 1;
	 
	    if (plotPanels[i].spVars[j][2] > -1)
	       sSize[2] = caaRead[plotPanels[i].spFile[j]].getVariableSize(plotPanels[i].spVar[j], plotPanels[i].spDepends[j][2]); 
	    else sSize[2] = 1;	 	 
	 
            int maxSize = maxValueOf(sSize, 3);

            // allocate memory

            float enCorrFac = 0.0; // energy correction factor for S/C potental correction

	    float *data = new float[sSize[0]*sSize[1]*sSize[2]];

            float **dataDepL;
            dataDepL = new float*[3];
            for (register int iy=0; iy<3; iy++) dataDepL[iy] = new float [maxSize];

            float **dataDepH;
            dataDepH = new float*[3];
            for (register int iy=0; iy<3; iy++) dataDepH[iy] = new float [maxSize];           

            float **dataDepC;
            dataDepC = new float*[3];
            for (register int iy=0; iy<3; iy++) dataDepC[iy] = new float [maxSize];  			

            // get the timing info

            QString time_stamp;
            caaRead[plotPanels[i].spFile[j]].getDataEpoch(0, &time_stamp);           
            float tisSpinCentre = time_to_tis((char *)(time_stamp.mid(11,12).toStdString()).c_str());
            float timeDeltaLo, timeDeltaHi;
            caaRead[plotPanels[i].spFile[j]].getDataValueDelta(0, _MINUS, &timeDeltaLo);
            caaRead[plotPanels[i].spFile[j]].getDataValueDelta(0, _PLUS, &timeDeltaHi);

           // get the depend data (azimuth, polar, energy)

           for (int id=0; id<3; id++)
           {
	      if (plotPanels[i].spVars[j][id] > -1)
	      {
                 caaRead[plotPanels[i].spFile[j]].getDataValue(plotPanels[i].spVars[j][id], &dataDepC[id][0]);
                 caaRead[plotPanels[i].spFile[j]].getDataValueDelta(plotPanels[i].spVars[j][id], _PLUS, &dataDepH[id][0]);
                 caaRead[plotPanels[i].spFile[j]].getDataValueDelta(plotPanels[i].spVars[j][id], _MINUS, &dataDepL[id][0]);
              }
	   }   

           // get the data
            
            caaRead[plotPanels[i].spFile[j]].getDataValue(plotPanels[i].spVar[j], data);

            // draw spectrogram slice

            float xBegin = tisSpinCentre - timeDeltaLo;
            float xEnd = tisSpinCentre + timeDeltaHi;

            // how many real azimuths are there?

            int numRealAzi = 0;
            for (int jIl=0; jIl<sSize[0]; jIl++)
            {
               if (dataDepC[0][jIl] > -1) numRealAzi++;
            }
            if (numRealAzi < 1) numRealAzi = 1;
			
            int jI[3];
            for (jI[0]=0; jI[0]<sSize[0]; jI[0]++) // <-- loop over azimuths
            {
               float axBegin = xBegin + (double)jI[0]*(xEnd - xBegin)/(double)numRealAzi;
               float axEnd = xBegin + (double)(jI[0]+1)*(xEnd - xBegin)/(double)numRealAzi;

               if (plotPanels[i].spPotentialCorrect[j] == 1)
               {
                  getPotentialValue(axBegin, axEnd, &iCounterPot, ii[scpFileCorr[plotPanels[i].scN[j]-1]],
                                    &xx[scpFileCorr[plotPanels[i].scN[j]-1]][0],
                                    &aa[scpDataRefCorr[plotPanels[i].scN[j]-1]][0], &enCorrFac);
               }
               else enCorrFac = 0.0;

               float *dcT = new float[sSize[2]];
               int *dcTc = new int[sSize[2]];
               for (int gg=0; gg<sSize[2]; gg++) {dcT[gg] = 0.0; dcTc[gg] = 0;}
               for (jI[1]=0; jI[1]<sSize[1]; jI[1]++) // <-- loop over polars
               {
                  for (jI[2]=0; jI[2]<sSize[2]; jI[2]++) // <-- loop over energies
                  {
                     float thisEnergyLower = dataDepC[2][jI[2]] - dataDepL[2][jI[2]] - enCorrFac - scpOffset;
                     float thisEnergyUpper = dataDepC[2][jI[2]] + dataDepH[2][jI[2]] - enCorrFac - scpOffset;

                     if (userMultFactorsListSpY[i][j] != 0.0) // energy scaling factors
                     {
                        thisEnergyLower = thisEnergyLower * userMultFactorsListSpY[i][j];
                        thisEnergyUpper = thisEnergyUpper * userMultFactorsListSpY[i][j];
                     }
		     
                     if (dataDepC[2][jI[2]] >= 0.0) // make sure we only attempt to plot data with sensible energies
		     {
		        float dc = data[jI[2] + jI[1]*sSize[2] + jI[0]*sSize[1]*sSize[2]]*plotPanels[i].multFactorSp[j];
					    
			if (plotPanels[i].spMaster == -2)
			{
			   if (dc != plotPanels[i].spFillValue[j] && dc > 0.)
                           {
                              dcT[jI[2]] = dcT[jI[2]] + dc;  		    	
                              dcTc[jI[2]]++;
                           }
			   if (jI[1] == sSize[1]-1)
			   {
                              dcT[jI[2]] = (dcTc[jI[2]] > 0) ? dcT[jI[2]]/((float)dcTc[jI[2]]) : 0.0;
                 	      myPlot.drawSpectrogramSlice(i, numColsIndex, dcT[jI[2]], axBegin, axEnd, thisEnergyLower, thisEnergyUpper,
                                                          plotPanels[i].cBarMin, plotPanels[i].cBarMax);	                               
			    }
			 }
                         else
		         {
		            for (int jh=0; jh < plotPanels[i].spMaster; jh++)
		            {
			       if (jI[1] == plotPanels[jh].spPolarComp || plotPanels[jh].spPolarComp == -1)
			       {
                 	          myPlot.drawSpectrogramSlice(i+jh, numColsIndex, dc, axBegin, axEnd, thisEnergyLower, thisEnergyUpper,
                                                              plotPanels[i].cBarMin, plotPanels[i].cBarMax);	
                               }
			    }					    	
			 }				
		      }			 
		  }
               }
               delete [] dcT;
               delete [] dcTc;
            }
			
            // de-allocate memory used by spectrograms
         	
            delete [] data;
            for (register int iy=2; iy>=0; iy--) delete [] dataDepL[iy]; delete [] dataDepL;
            for (register int iy=2; iy>=0; iy--) delete [] dataDepH[iy]; delete [] dataDepH;
            for (register int iy=2; iy>=0; iy--) delete [] dataDepC[iy]; delete [] dataDepC;			
			

	    } // <-- loop over spectrograms from the same file
         } // <-- while loop
      } // <-- loop over spectrograms

      int i = im;

// draw time series plots

      if (plotPanels[i].tsFromDiff == 0)
      {
         myPlot.setLegendLocation(Outside);	  
      }
      else
      {
         myPlot.setLegendLocation(TopLeft);	  
      }
	  
      for (int j=0; j<plotPanels[i].numTimeSeries; j++)
      {
         QString legendTmp;
         if (plotPanels[i].tsFromDiff == 0)
	 {
	    QStringList label1List, label2List;
            QString label1 = caaRead[plotPanels[i].tsFile[j]].getVariableAttr(plotPanels[i].tsVar[j], "LABEL_1");
            QString label2 = caaRead[plotPanels[i].tsFile[j]].getVariableAttr(plotPanels[i].tsVar[j], "LABEL_2");

            // sometimes REPRESENTATION is used instead of LABEL
            if (label1.isEmpty())
            {
               label1 = caaRead[plotPanels[i].tsFile[j]].getVariableAttr(plotPanels[i].tsVar[j], "REPRESENTATION_1");
            }
            if (label2.isEmpty())
            {
               label2 = caaRead[plotPanels[i].tsFile[j]].getVariableAttr(plotPanels[i].tsVar[j], "REPRESENTATION_2");
            }

            label1List = label1.split(",", QString::SkipEmptyParts);	
            label2List = label2.split(",", QString::SkipEmptyParts);
            if (j < label1List.size())
            {
	       legendTmp = label1List.at(j);
	    }
            else if (j-label1List.size() < label2List.size())
            {
               legendTmp = label2List.at(j-label1List.size());
            }
	    else
	    {
	       legendTmp = "";
	    }
	 }
	 else
	 {
	    legendTmp = plotPanels[i].legLabel[j];
	 }
		 
	 if (!pListLegends[i][j].isEmpty()) legendTmp = pListLegends[i][j];

         // set the line/symbol colour
	  
         QColor tmpp;
	 tmpp.setNamedColor(colorList[j]);
         if (!plotPanels[i].colour[j].isEmpty()) tmpp.setNamedColor(plotPanels[i].colour[j]);

         // draw data point
	  
         myPlot.drawCurve(i, numColsIndex, tmpp, ii[plotPanels[i].tsFile[j]], &xx[plotPanels[i].tsFile[j]][0], 
	                  &aa[plotPanels[i].tsDataRef[j]][0], legendTmp, pListSymbolPlots[i][j]);

         // show legend if more than 1 dataset in the panel

         if (plotPanels[i].numTimeSeries > 1)
         {
	    if ((!autoPlotV && !autoPlotNV && !autoPlotNVT && !autoPlotNT) || ((autoPlotV || autoPlotNV || autoPlotNVT) && i == vLoc) ||
                ((autoPlotN || autoPlotNV || autoPlotNVT || autoPlotNT) && i == nLoc) ||
                ((autoPlotNT || autoPlotNVT) && i == tLoc) || (iFoundB > 0 && i == bLoc))
            myPlot.showLegend(i, numColsIndex, j, pListSymbolPlots[i][j]);
	 }
      }

      // finally draw the panel's frame as well as x- & y-axis ticks
	  
      myPlot.drawFrame(i, numColsIndex);
      myPlot.setNumTicksY(i, numColsIndex, pListMajY[i], pListMinY[i]);
      myPlot.setNumTicksX(i, numColsIndex, durMajX, nMinX);
      if (!userListPlotTitles[i].isEmpty()) myPlot.drawPlotTitle(i, numColsIndex, userListPlotTitles[i], 0);	  
      if (!userListPlotTitlesRHS[i].isEmpty()) myPlot.drawPlotTitle(i, numColsIndex, userListPlotTitlesRHS[i], 1);
   }
   
//------------------
// deallocate memory
//------------------

   for (register int i=tsArraySize-1; i>=0; i--) delete [] aa[i]; delete [] aa;
   for (register int i=numFiles-1; i>=0; i--) delete [] xx[i]; delete [] xx;  

   } // <-- end if wheel plot
    
//----------------------
// close CDF/CEF file(s)
//----------------------

   for (int i=0; i<numFiles; i++)
   {
      caaRead[i].close();
   }

//----------------------------------------
// show plot - either wheel or time-series
//----------------------------------------

   if (!makeWheel)
   {
      // save to file if requested

      myPlot.drawLabel(plotLabel);
      if (!outputFileName.isEmpty())
      {
         myPlot.saveToFile(outputFileName, plotMetaData);
      }
      else if (noDisplay && numFiles == 1)
      {

         // Automatically generate a PNG image filename if the user specified not to display
         // the image & an output filename wasn't specified

         outputFileName = QString(argv[fileLocArg[0]]);
         outputFileName = outputFileName.mid(0, outputFileName.length() - 3) + "png";
         myPlot.saveToFile(outputFileName, plotMetaData);
      }

      // display the plot on the screen
      
      if (!noDisplay)
      {
	 PlotWindow myPlotWindow(plotWidth, plotHeight);
         myPlotWindow.plot = myPlot.plot;
         myPlotWindow.show();
         myPlotWindow.setMaximumSize(myPlotWindow.minimumWidth(),
                                     myPlotWindow.minimumHeight());
         myPlotWindow.update();
         myPlotWindow.raise();
         myPlotWindow.activateWindow();
         cout << "[plotcf] Done." << endl;
         return app.exec();
      }
   }
   else
   {
      // save to file if requested

      myWheelPlot.drawLabel(plotLabel);
      if (!outputFileName.isEmpty())
      {
          myWheelPlot.saveToFile(outputFileName, plotMetaData);
      }
      else if (noDisplay && numFiles == 1)
      {

         // Automatically generate a PNG image filename if the user specified not to display
         // the image & an output filename wasn't specified

         outputFileName = QString(argv[fileLocArg[0]]);
         outputFileName = outputFileName.mid(0, outputFileName.length() - 3) + "png";
         myWheelPlot.saveToFile(outputFileName, plotMetaData);
      }

      // display the plot on the screen
     
      if (!noDisplay)
      {
         PlotWindow myPlotWindow(plotWidth, plotHeight);
         myPlotWindow.plot = myWheelPlot.plot;
         myPlotWindow.show();
         myPlotWindow.setMaximumSize(myPlotWindow.minimumWidth(),
                                     myPlotWindow.minimumHeight());
         myPlotWindow.update();
         myPlotWindow.raise();
         myPlotWindow.activateWindow();
         cout << "[plotcf] Done." << endl;
         return app.exec();
      }
   }

// end

   cout << "[plotcf] Done." << endl;

}

